#!/usr/bin/env python
# -*- coding: latin-1; -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import print_function
from builtins import range
import numpy as np
import os
import shutil 

import Metafor.Mparams.TaskManager as tm
import Metafor.Mparams.mparams as mp


def main():
    if tm.rank == 0:
        #--------------------------------------------------------
        #                       Model parameters
        #--------------------------------------------------------
        model_parameters = {}
        isUnix = lambda: os.name == 'posix'

        if isUnix(): 
            model_parameters['Metafor_exe']  = '/home/kliegeois/Codes/Metafor/oo_metaB/bin/Metafor'
            model_parameters['Metafor_model'] = '/home/kliegeois/Codes/waves/Metafor/models/tube01'      
        else:     
            model_parameters['Metafor_exe']  = 'C:/Users/kimli_000/dev/oo_metaB/bin/Release/Metafor'
            model_parameters['Metafor_model'] = 'C:/Users/kimli_000/dev/waves/Metafor/models/tube01'
        #--------------------------------------------------------
        #                    Surrogate parameters
        #--------------------------------------------------------
        surrogate_parameters = {}
        surrogate_parameters['display'] = True  
        if not isUnix():
            surrogate_parameters['display'] = False 
        surrogate_parameters['normalisation'] = True  
        #--------------------------------------------------------    
        surrogate_parameters['fname'] = 'mparams.log'
        d = 2
        surrogate_parameters['d'] = d
        surrogate_parameters['P_name'] = np.array([ 'R1', 'R2'])

        M = np.zeros((d,), dtype =np.int)
        for j in range(0,d):
            M[j] = 8    

        surrogate_parameters['M'] = M
            
        surrogate_parameters['Dom_min'] = np.array([ 3., 4.5])
        surrogate_parameters['Dom_max'] = np.array([ 4., 5.5])   

        N = np.zeros((d,), dtype =np.int)
        for j in range(0,d):
            N[j] = 4   

        surrogate_parameters['N'] = N
        #--------------------------------------------------------   
        #-------------------------------------------------------- 
        surrogate_parameters['function_type'] = 'monomial'
        surrogate_parameters['training_points'] = 'gauss'
        surrogate_parameters['solver'] = 'QR'
        #--------------------------------------------------------
        surrogate_parameters['m1'] = 40
        surrogate_parameters['m2'] = 40
        #--------------------------------------------------------
        #                    Images parameters
        #--------------------------------------------------------
        images_parameters = {}
        images_parameters['xlabel'] = 'R_1 $[$mm$]$'
        images_parameters['ylabel'] = 'R_2 $[$mm$]$'
        images_parameters['zlabel'] = 'F_x $[$N$]$'

    if mp.AvailableComputedResults():
        if tm.rank == 0:
            print("the master/sol.json file will be loaded")
            print("the master/Z.json file will be loaded")
            print("the master/W_sr.json file will be loaded")
            sol, Z, W_sr = mp.LoadComputedResults()

            from Metafor.Mparams.sequential import scatterplot
            Image = scatterplot(sol, surrogate_parameters, images_parameters)
            from Metafor.Mparams.surrogate import surrogate_creation
            s_hat = surrogate_creation(sol, Z, W_sr, surrogate_parameters, Image, images_parameters)
    else:
        if tm.rank == 0:
            ##

            sol, Z, W_sr, s_hat = mp.master(model_parameters,surrogate_parameters,images_parameters)

            ##

            import codecs, json 

            tmp = sol.tolist() 
            file_path = "sol.json" 
            json.dump(tmp, codecs.open(file_path, 'w', encoding='utf-8'), separators=(',', ':'), sort_keys=True, indent=4) 

            tmp = Z.tolist() 
            file_path = "Z.json" 
            json.dump(tmp, codecs.open(file_path, 'w', encoding='utf-8'), separators=(',', ':'), sort_keys=True, indent=4) 

            tmp = W_sr.tolist() 
            file_path = "W_sr.json" 
            json.dump(tmp, codecs.open(file_path, 'w', encoding='utf-8'), separators=(',', ':'), sort_keys=True, indent=4) 

            tmp = s_hat.tolist() 
            file_path = "s_hat.json" 
            json.dump(tmp, codecs.open(file_path, 'w', encoding='utf-8'), separators=(',', ':'), sort_keys=True, indent=4) 


        else:
            mp.worker()

if __name__ == "__main__":
    main()
