/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wKutta.h"
#include "wGroup.h"
#include "wTag.h"
#include <algorithm>
using namespace tbox;
using namespace flow;

Kutta::Kutta(std::shared_ptr<MshData> _msh, std::vector<int> const &nos)
    : Groups(_msh, nos)
{
    // Sanity check
    if (nos.size() != 4)
    {
        std::stringstream err;
        err << "flow::Kutta should be built with 4 physical groups but " << nos.size() << " were given!\n";
        throw std::runtime_error(err.str());
    }
    // Build data structure
    connectNodes();
    createElements();
}
Kutta::Kutta(std::shared_ptr<MshData> _msh, std::vector<std::string> const &names)
    : Groups(_msh, names)
{
    // Sanity check
    if (names.size() != 4)
    {
        std::stringstream err;
        err << "flow::Kutta should be built with 4 physical groups but " << names.size() << " were given!\n";
        throw std::runtime_error(err.str());
    }
    // Build data structure
    connectNodes();
    createElements();
}
Kutta::~Kutta()
{
    for (size_t i = 0; i < kEle.size(); ++i)
    {
        delete kEle[i];
    }
    std::cout << "~Kutta()\n";
}

/**
 * @brief Connect matching nodes on wing trailing edge
 * @authors Adrien Crovato
 */
void Kutta::connectNodes()
{
    // Temporary container for nodes
    std::vector<Node *> nodesUp, nodesLw;
    // Get unique node
    for (auto e : groups[0]->tag->elems)
        for (auto n : e->nodes)
            nodesUp.push_back(n);
    for (auto e : groups[1]->tag->elems)
        for (auto n : e->nodes)
            nodesLw.push_back(n);
    std::sort(nodesUp.begin(), nodesUp.end());
    auto it = std::unique(nodesUp.begin(), nodesUp.end());
    nodesUp.resize(std::distance(nodesUp.begin(), it));
    std::sort(nodesLw.begin(), nodesLw.end());
    it = std::unique(nodesLw.begin(), nodesLw.end());
    nodesLw.resize(std::distance(nodesLw.begin(), it));

    // Create the node map
    for (auto nUp : nodesUp)
    {
        //std::cout << "MASTER: " << nUp->pos.x[0] << ',' << nUp->pos.x[1] << ',' << nUp->pos.x[2] << std::endl;
        for (auto nLw : nodesLw)
        {
            //std::cout << "    SLAVE: " << nLw->pos.x[0] << ',' << nLw->pos.x[1] << ',' << nLw->pos.x[2] << std::endl;
            if (nUp->pos.compare(nLw->pos, 1e-14))
            {
                nodMap.push_back(std::pair<Node *, Node *>(nUp, nLw));
                break;
            }
        }
    }
    // Check if node map is not empty
    if (nodMap.size() != nodesUp.size())
    {
        std::stringstream err;
        err << "flow::Kutta: some nodes from " << *groups[0]->tag << " and " << *groups[1]->tag << " could not be mapped (number of mapped nodes: " << nodMap.size() << ")!\n";
        throw std::runtime_error(err.str());
    }
}

/**
 * @brief Create Kutta elements
 * @authors Adrien Crovato
 */
void Kutta::createElements()
{
    // temporary
    std::vector<std::pair<Element *, Element *>> uPair, lPair;
    std::vector<std::vector<std::pair<size_t, Node *>>> uNodes, lNodes;
    // find and connect upper and lower surface-volume elements
    connectElements(uPair, uNodes, true);
    connectElements(lPair, lNodes, false);

    // Create the Kutta elements
    size_t i = 0;
    size_t k = 0;
    for (auto it = uPair.begin(); it != uPair.end(); it++, i++)
    {
        KuttaElement *ke = new KuttaElement(i + 1, (*it).first, (*it).second, uNodes[k], 1);
        kEle.push_back(ke);
        k++;
    }
    k = 0;
    for (auto it = lPair.begin(); it != lPair.end(); it++, i++)
    {
        KuttaElement *ke = new KuttaElement(i + 1, (*it).first, (*it).second, lNodes[k], -1);
        kEle.push_back(ke);
        k++;
    }
}

/**
 * @brief Detect surface elements touching TE and connect with their volume element
 * @authors Adrien Crovato
 */
void Kutta::connectElements(std::vector<std::pair<Element *, Element *>> &svPair, std::vector<std::vector<std::pair<size_t, Node *>>> &surN, bool uFlag)
{
    // find elements with (at least) one node on TE and store in vector by comparing the element nodes to the trailing edge nodes
    std::vector<Element *> surE;
    for (auto sE : groups[2]->tag->elems)
    {
        std::vector<std::pair<size_t, Node *>> tmp;
        for (size_t i = 0; i < sE->nodes.size(); ++i)
        {
            for (auto teP : nodMap)
            {
                // check if the node is on the upper or lower TE
                Node *teN;
                if (uFlag)
                    teN = teP.first;
                else
                    teN = teP.second;
                if (sE->nodes[i] == teN)
                {
                    tmp.push_back(std::pair<size_t, Node *>(i, teP.first));
                    break;
                }
            }
        }
        if (!tmp.empty())
        {
            surN.push_back(tmp);
            surE.push_back(sE);
        }
    }
    // find matching volume elements for each surface element by comparing each node of surface element to volume element
    svPair.resize(surE.size());
    int i = 0;
    for (auto sE : surE)
    {
        for (auto vE : groups[3]->tag->elems)
        {
            size_t cnt = 0;
            for (auto sN : sE->nodes)
            {
                for (auto vN : vE->nodes)
                {
                    if (sN == vN)
                    {
                        cnt++;
                        break;
                    }
                }
                if (cnt == 0)
                    break;
            }
            // if all surface nodes are also volume nodes
            if (cnt == sE->nodes.size())
            {
                svPair[i] = std::pair<Element *, Element *>(sE, vE);
                break;
            }
        }
        i++;
    }
}

void Kutta::write(std::ostream &out) const
{
    out << "flow::Kutta on " << *groups[0]->tag << " and " << *groups[1]->tag << "\n";
    for (auto p : nodMap)
        out << "\t" << p.first->no << " linked with " << p.second->no << "\n";
    for (auto ke : kEle)
        ke->write(out);
}