/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wSolver.h"
#include "wProblem.h"
#include "wMedium.h"
#include "wAssign.h"
#include "wBoundary.h"

#include "wMshData.h"
#include "wNode.h"
#include "wElement.h"
#include "wMem.h"
#include "wTag.h"
#include "wResults.h"
#include "wMshExport.h"

#include <tbb/task_scheduler_init.h>
#include <tbb/tbb.h>
#include <tbb/parallel_do.h>
#include <tbb/spin_mutex.h>

using namespace tbox;
using namespace flow;

#define ANSI_COLOR_YELLOW "\x1b[1;33m"
#define ANSI_COLOR_RESET "\x1b[0m"

/**
 * @brief Initialize the solver and perform sanity checks
 * @authors Adrien Crovato
 */
Solver::Solver(std::shared_ptr<Problem> _pbl) : pbl(_pbl)
{
    // Say hi
    std::cout << std::endl;
    std::cout << "*******************************************************************************" << std::endl;
    std::cout << "**                                    \\_/                                    **" << std::endl;
    std::cout << "**                           \\_______O(_)O_______/                           **" << std::endl;
    std::cout << "**                                                                           **" << std::endl;
    std::cout << "**                           ___   _                                         **" << std::endl;
    std::cout << "**                          | __| | |  __  __   __                           **" << std::endl;
    std::cout << "**                          | __| | | /  \\ \\ \\_/ /                           **" << std::endl;
    std::cout << "**                          |_|   |_| \\__/  \\/ \\/                            **" << std::endl;
    std::cout << "*******************************************************************************" << std::endl;
    std::cout << "** Hi! My name is Flow v1.6-2003                                             **" << std::endl;
    std::cout << "** Adrien Crovato & Romain Boman                                             **" << std::endl;
    std::cout << "** ULiege 2018-2020                                                          **" << std::endl;
    std::cout << "*******************************************************************************" << std::endl
              << std::endl;

    // Default parameters
    nthreads = 1;
    verbose = false;
    maxIt = 50;
    relTol = 1e-6;
    absTol = 1e-8;

    // Check the problem and update element memory
    pbl->check();
    pbl->updateMem();

    // Assign a row (unknown number) for each node
    for (auto n : pbl->msh->nodes)
        n->row = n->no - 1;

    // Setup variables (unkwown vector)
    phi.resize(pbl->msh->nodes.size(), 0.);
    rPhi.resize(pbl->msh->nodes.size(), 0.);
    vPhi.resize(pbl->msh->nodes.size(), 0.);
    U.resize(pbl->msh->nodes.size(), Pt(0., 0., 0.));
    rho.resize(pbl->msh->nodes.size(), 0.);
    M.resize(pbl->msh->nodes.size(), 0.);
    Cp.resize(pbl->msh->nodes.size(), 0.);

    // Apply Initial and Dirichlet boundary conditions
    for (auto iIC : pbl->iICs)
        iIC->apply(phi);
    for (auto dBC : pbl->dBCs)
        dBC->apply(phi);
}
Solver::~Solver()
{
    // Say bye
    std::cout << std::endl;
    std::cout << "*******************************************************************************" << std::endl;
    std::cout << "**                                    \\_/                                    **" << std::endl;
    std::cout << "**                           \\_______O(_)O_______/                           **" << std::endl;
    std::cout << "**                                  !     !                                  **" << std::endl;
    std::cout << "*******************************************************************************" << std::endl;
    std::cout << std::endl;
}

/**
 * @brief Dummy run
 * @authors Adrien Crovato
 */
bool Solver::run()
{
    throw std::runtime_error("flow::Solver::run not implemented!\n");
}

/**
 * @brief Write the results
 * @authors Adrien Crovato
 */
void Solver::save(int n, std::shared_ptr<MshExport> mshWriter)
{
    // Write files
    std::cout << "Saving files... " << std::endl;
    // setup results
    Results results;
    results.scalars_at_nodes["phi"] = &phi;
    results.scalars_at_nodes["varPhi"] = &vPhi;
    results.scalars_at_nodes["resPhi"] = &rPhi;
    results.vectors_at_nodes["U"] = &U;
    results.scalars_at_nodes["rho"] = &rho;
    results.scalars_at_nodes["Mach"] = &M;
    results.scalars_at_nodes["Cp"] = &Cp;
    // save (all mesh and boundary surface)
    if (n > 0)
    {
        mshWriter->save(pbl->msh->name + "_" + std::to_string(n), results);
        for (auto sur : pbl->bnd)
            sur->save(sur->groups[0]->tag->name + "_" + std::to_string(n), results);
    }
    else
    {
        mshWriter->save(pbl->msh->name, results);
        for (auto sur : pbl->bnd)
            sur->save(sur->groups[0]->tag->name, results);
    }
}

/**
 * @brief Compute total aerodynamic load coefficients on boundaries
 * @authors Adrien Crovato
 */
void Solver::computeLoad()
{
    Cl = 0;
    Cd = 0;
    Cm = 0;
    for (auto sur : pbl->bnd)
    {
        // Map each element to its value
        std::map<Element *, double> euMap;
        for (auto e : sur->groups[0]->tag->elems)
            euMap[e] = sur->integrate(e, phi, pbl->medium->cP);

        // Compute normalized load (i.e. load / pressure)
        for (size_t i = 0; i < sur->nodes.size(); ++i)
        {
            sur->cLoadX[i] = 0;
            sur->cLoadY[i] = 0;
            sur->cLoadZ[i] = 0;
            for (auto e : sur->neMap.at(sur->nodes[i]))
            {
                sur->cLoadX[i] += euMap.at(e) * e->normal().x[0] / e->nodes.size();
                sur->cLoadY[i] += euMap.at(e) * e->normal().x[1] / e->nodes.size();
                sur->cLoadZ[i] += euMap.at(e) * e->normal().x[2] / e->nodes.size();
            }
        }

        // Compute aerodynamic load coefficients (i.e. Load / pressure / surface)
        // compute coefficients along x and normal (y in 2D, z in 3D) directions and pitching moment (along -z in 2D, y in 3D)
        sur->Cm = 0;
        double Cx = 0., Cn = 0.;
        for (auto p : euMap)
        {
            double lx = p.first->getVMem().getCg().x[0] - pbl->x_ref.x[0];
            double ln = p.first->getVMem().getCg().x[pbl->nDim - 1] - pbl->x_ref.x[pbl->nDim - 1];
            double Cfx = p.second * p.first->normal().x[0];
            double Cfn = p.second * p.first->normal().x[pbl->nDim - 1];
            Cx -= Cfx;
            Cn -= Cfn;
            sur->Cm -= (ln * Cfx - lx * Cfn) / (pbl->S_ref * pbl->c_ref); // moment is positive along y (3D) and negative along z (2D) => positive nose-up
        }
        // rotate to flow direction
        sur->Cl = (Cn * cos(pbl->alpha) - Cx * sin(pbl->alpha)) / pbl->S_ref;
        sur->Cd = (Cn * sin(pbl->alpha) + Cx * cos(pbl->alpha)) / pbl->S_ref;
        // compute total
        Cl += sur->Cl;
        Cd += sur->Cd;
        Cm += sur->Cm;
    }
}

/**
 * @brief Compute variables in the flow field
 * @authors Adrien Crovato
 */
void Solver::computeFlow()
{
    // Compute results in volume
    auto fluid = pbl->medium;
    tbb::parallel_do(fluid->neMap.begin(), fluid->neMap.end(), [&](std::pair<Node *, std::vector<Element *>> p) {
        Node *n = p.first;
        // Perturbation potential
        vPhi[n->row] = phi[n->row] - fluid->phiInf.eval(n->pos);
        // Velocity (averaged at nodes)
        U[n->row] = Pt(0., 0., 0.);
        for (auto e : p.second)
        {
            Eigen::VectorXd grad;
            e->computeGrad(phi, grad, 0);
            for (size_t i = 0; i < grad.size(); ++i)
                U[n->row].x[i] += grad[i];
        }
        U[n->row] /= p.second.size();
        // Density (averaged at nodes)
        rho[n->row] = 0.;
        for (auto e : p.second)
            rho[n->row] += fluid->rho.eval(*e, phi, 0);
        rho[n->row] /= p.second.size();
        // Mach number (averaged at nodes)
        M[n->row] = 0.;
        for (auto e : p.second)
            M[n->row] += fluid->mach.eval(*e, phi, 0);
        M[n->row] /= p.second.size();
        // Pressure coefficient (averaged at nodes)
        Cp[n->row] = 0.;
        for (auto e : p.second)
            Cp[n->row] += fluid->cP.eval(*e, phi, 0);
        Cp[n->row] /= p.second.size();
    });

    // Check maximum Mach number
    if (*std::max_element(M.begin(), M.end()) >= 1.25)
        std::cout << ANSI_COLOR_YELLOW << "Max. Mach greater than 1.25!" << ANSI_COLOR_RESET << std::endl;
}
