#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# @brief Compute the flow on a deformed mesh
# @authors Adrien Crovato
# Test the mesh deformation process
#
# CAUTION
# This test is provided to ensure that the solver works properly.
# Mesh refinement may have to be performed to obtain physical results.

from __future__ import print_function
import numpy as np
import flow as flo
import flow.utils as floU
import flow.default as floD
import tbox
import tbox.utils as tboxU
import tbox.gmsh as gmsh
import fwk
from fwk.testing import *
from fwk.coloring import ccolors 

def main():
    # timer
    tms = fwk.Timers()
    tms['total'].start()

    # define flow variables
    alpha0 = 0 # must be zero for this testfile
    U_inf = [np.cos(alpha0), np.sin(alpha0)] # norm should be 1
    M_inf = 0.3
    M_crit = 5 # Squared critical Mach number (above which density is modified)
    c_ref = 1
    dim = len(U_inf)
    # parameters for mesh deformation
    alfa = 10*np.pi/180
    xc = 0.25
    dx = 0.2
    dy = 0.5

    # mesh the geometry
    print(ccolors.ANSI_BLUE + 'PyMeshing...' + ccolors.ANSI_RESET)
    tms['msh'].start()
    pars = {'xLgt' : 5, 'yLgt' : 5, 'msF' : 1., 'msTe' : 0.01, 'msLe' : 0.01}
    msh = gmsh.MeshLoader("../models/n0012.geo",__file__).execute(**pars)
    # crack the mesh
    mshCrck = tbox.MshCrack(msh, dim)
    mshCrck.setCrack('wake')
    mshCrck.addBoundaries(['field', 'airfoil', 'downstream'])
    mshCrck.run()
    del mshCrck
    gmshWriter = tbox.GmshExport(msh)
    gmshWriter.save(msh.name)
    # create mesh deformation handler
    mshDef = tbox.MshDeform(msh, dim)
    mshDef.setField("field")
    mshDef.addFixed(["upstream", "side", "downstream"])
    mshDef.addMoving(["airfoil"])
    mshDef.addInternal(["wake", "wake_"])
    mshDef.initialize()
    # mesh the reference geometry
    pars = {'xLgt' : 5, 'yLgt' : 5, 'msF' : 1., 'msTe' : 0.01, 'msLe' : 0.01, 'angle' : alfa, 'xRot' : xc}
    msh_ref = gmsh.MeshLoader("../models/n0012.geo",__file__).execute(**pars)
    # crack the mesh
    mshCrck_ref = tbox.MshCrack(msh_ref, dim)
    mshCrck_ref.setCrack('wake')
    mshCrck_ref.addBoundaries(['field', 'airfoil', 'downstream'])
    mshCrck_ref.run()
    del mshCrck_ref
    gmshWriter_ref = tbox.GmshExport(msh_ref)
    gmshWriter_ref.save(msh_ref.name)
    tms['msh'].stop()

    # set the problem and add medium, initial/boundary conditions
    tms['pre'].start()
    pbl = flo.Problem(msh, dim, alpha0, M_inf, c_ref, c_ref, xc, 0., 0.)
    if M_inf == 0:
        pbl.set(flo.Medium(msh, 'field', flo.F0ElRhoL(), flo.F0ElMachL(), flo.F0ElCpL(), flo.F0PsPhiInf(dim, alpha0)))
    else:
        pbl.set(flo.Medium(msh, 'field', flo.F0ElRho(M_inf, M_crit), flo.F0ElMach(M_inf), flo.F0ElCp(M_inf), flo.F0PsPhiInf(dim, alpha0)))
    pbl.add(flo.Initial(msh, 'field', flo.F0PsPhiInf(dim, alpha0)))
    pbl.add(flo.Dirichlet(msh, 'upstream', flo.F0PsPhiInf(dim, alpha0)))
    pbl.add(flo.Freestream(msh, 'side', tbox.Fct1C(-U_inf[0], -U_inf[1], 0.)))
    pbl.add(flo.Freestream(msh, 'downstream', tbox.Fct1C(-U_inf[0], -U_inf[1], 0.)))
    pbl.add(flo.Wake(msh, ['wake', 'wake_', 'field']))
    pbl.add(flo.Kutta(msh, ['te', 'wake_', 'airfoil', 'field']))
    bnd = flo.Boundary(msh, ['airfoil', 'field'])
    pbl.add(bnd)
    # set the refrence problem and add medium, initial/boundary conditions
    pbl_ref = flo.Problem(msh_ref, dim, alpha0, M_inf, c_ref, c_ref, xc, 0., 0.)
    if M_inf == 0:
        pbl_ref.set(flo.Medium(msh_ref, 'field', flo.F0ElRhoL(), flo.F0ElMachL(), flo.F0ElCpL(), flo.F0PsPhiInf(dim, alpha0)))
    else:
        pbl_ref.set(flo.Medium(msh_ref, 'field', flo.F0ElRho(M_inf, M_crit), flo.F0ElMach(M_inf), flo.F0ElCp(M_inf), flo.F0PsPhiInf(dim, alpha0)))
    pbl_ref.add(flo.Initial(msh_ref, 'field', flo.F0PsPhiInf(dim, alpha0)))
    pbl_ref.add(flo.Dirichlet(msh_ref, 'upstream', flo.F0PsPhiInf(dim, alpha0)))
    pbl_ref.add(flo.Freestream(msh_ref, 'side', tbox.Fct1C(-U_inf[0], -U_inf[1], 0.)))
    pbl_ref.add(flo.Freestream(msh_ref, 'downstream', tbox.Fct1C(-U_inf[0], -U_inf[1], 0.)))
    pbl_ref.add(flo.Wake(msh_ref, ['wake', 'wake_', 'field']))
    pbl_ref.add(flo.Kutta(msh_ref, ['te', 'wake_', 'airfoil', 'field']))
    bnd_ref = flo.Boundary(msh_ref, ['airfoil', 'field'])
    pbl_ref.add(bnd_ref)
    tms['pre'].stop()

    # solver problem for 0°AOA
    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['solver0'].start()
    solver = flo.Newton(pbl)
    floD.initNewton(solver)
    solver.run()
    tms['solver0'].stop()

    # deform the mesh (dummy rotation + translation)
    print(ccolors.ANSI_BLUE + 'PyDeforming...' + ccolors.ANSI_RESET)
    tms['deform'].start()
    floD.initMeshDeformer(mshDef)
    mshDef.savePos()
    rMat = np.matrix([[np.cos(alfa), np.sin(alfa)], [-np.sin(alfa), np.cos(alfa)]])
    vNods = {}
    for e in bnd.groups[0].tag.elems:
        for n in e.nodes:
            if n.no not in vNods:
                vNods[n.no] = n
    for no in vNods:
        nod = np.matrix([[vNods[no].pos.x[0]-xc], [vNods[no].pos.x[1]]])
        new = rMat * nod
        vNods[no].pos.x[0] = new[0,0] + dx + xc
        vNods[no].pos.x[1] = new[1,0] + dy
    # smooth the mesh
    mshDef.deform()
    gmshWriter.save("n0012_def")
    tms['deform'].stop()

    # solve problem for alfa AOA
    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['solver1'].start()
    solver.run()
    solver.save(0, gmshWriter)
    tms['solver1'].stop()

    # compute reference solution
    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['solver_ref'].start()
    solver_ref = flo.Newton(pbl_ref)
    floD.initNewton(solver_ref)
    solver_ref.run()
    solver_ref.save(0, gmshWriter_ref)
    tms['solver_ref'].stop()

    # extract Cp
    Cp = floU.extract(bnd.groups[0].tag.elems, solver.Cp)
    tboxU.write(Cp, 'Cp_airfoil.dat', '%1.5e', ', ', 'x, y, z, Cp', '')
    Cp_ref = floU.extract(bnd_ref.groups[0].tag.elems, solver_ref.Cp)
    tboxU.write(Cp, 'Cp_airfoil_ref.dat', '%1.5e', ', ', 'x, y, z, Cp', '')
    # display results
    print(ccolors.ANSI_BLUE + 'PyRes...' + ccolors.ANSI_RESET)
    print('  case        M    alpha       Cl       Cd       Cm')
    print('  true {0:8.2f} {1:8.1f} {2:8.4f} {3:8.4f} {4:8.4f}'.format(M_inf, alfa*180/np.pi, solver.Cl, solver.Cd, solver.Cm))
    print('   ref {0:8.2f} {1:8.1f} {2:8.4f} {3:8.4f} {4:8.4f}'.format(M_inf, alfa*180/np.pi, solver_ref.Cl, solver_ref.Cd, solver_ref.Cm))
    print('Cm will differ unless dx and dy are set to 0...')

    # display timers
    tms["total"].stop()
    print(ccolors.ANSI_BLUE + 'PyTiming...' + ccolors.ANSI_RESET)
    print("CPU statistics")
    print(tms)

    # visualize solution and plot results
    tboxU.plot(Cp[:,0], Cp[:,3], 'x', 'Cp', 'Cl = {0:.{3}f}, Cd = {1:.{3}f}, Cm = {2:.{3}f}'.format(solver.Cl, solver.Cd, solver.Cm, 4), True)
    tboxU.plot(Cp_ref[:,0], Cp_ref[:,3], 'x', 'Cp', 'Cl = {0:.{3}f}, Cd = {1:.{3}f}, Cm = {2:.{3}f}'.format(solver_ref.Cl, solver_ref.Cd, solver_ref.Cm, 4), True)

    # check results
    print(ccolors.ANSI_BLUE + 'PyTesting...' + ccolors.ANSI_RESET)
    tests = CTests()
    tests.add(CTest('solver: iteration count', solver.nIt, 5, 3, forceabs=True)) # todo: tolerance should be reset to 1 as soon as gmsh4 is used
    tests.add(CTest('solver_ref: iteration count', solver_ref.nIt, 5, 1, forceabs=True))
    tests.add(CTest('min(Cp)', min(Cp[:,3]), min(Cp_ref[:,3]), 5e-2))
    tests.add(CTest('Cl', solver.Cl, solver_ref.Cl, 5e-2))
    tests.run()

    # eof
    print('')

if __name__ == "__main__":
    main()
