#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# calculate the conductivity matrix of a particular problem
# from the linear combination of 2 calculations (2 perpendicular gradients)
# then, checks if the resulting flux is the same as a direct calculation

from __future__ import print_function
from __future__ import division
from past.utils import old_div
import heat as h
import heat.utils as u
import tbox
import tbox.gmsh as gmsh
import fwk
from fwk.testing import *

class MyMaterial(tbox.Fct2XYZ):
    """
    the considered conductivity is a function of the position (x,y)
    """
    def __init__(self, k1, k2):
        tbox.Fct2XYZ.__init__(self)
        self.k1 = k1
        self.k2 = k2
    def eval(self, pt, k, fake):
        x = pt.x[0]
        y = pt.x[1]
        L = 1.
        import math
        v = old_div((self.k1+self.k2),2)+ abs(self.k2-self.k1)/2*math.sin(2*math.pi/L*4*(x+old_div(y,2)))
        k.resize(2, 2)
        k[0,0] = k[1,1] = v
        k[0,1] = k[1,0] = 0.
        #k[0,0] = self.k1  # constant
        #k[1,1] = self.k2


def solve(TM, dTxM, dTyM):
    """
    solve one FE problem (for given mean values of T and gradT)
    """
    pars={ 'Lx' : 1 , 'Ly' : 1, 'Nx' : 20, 'Ny' : 20}
    msh = gmsh.MeshLoader("../../models/rect_stru.geo",__file__).execute(**pars)
    gmshWriter = tbox.GmshExport(msh)
    pbl = h.Problem(msh)    

    myk = MyMaterial(1.,1000.)
    pbl.add(h.Medium(msh, "S1", myk))

    p = h.Periodic(pbl, TM, dTxM, dTyM)
    pbl.add(p)
    
    # solve problem
    solver = u.solve(pbl, gmshWriter)

    q1i = p.qM.x[0]
    q2i = p.qM.x[1]

    return (q1i, q2i)


def main():

    chrono = fwk.Timer(); chrono.start()

    # ------------- main problem ----------------------
    # -------------------------------------------------

    TM0   = 0.0
    dTxM0 = 0.3
    dTyM0 = 0.2
 
    q1i,q2i = solve(TM0, dTxM0, dTyM0)
    print("flux along x dir = ", q1i)
    print("flux along y dir = ", q2i)

        
    # ------------- linear combination ----------------
    # -------------------------------------------------

    TM   = TM0
    dTxM = 1.0   # <= gradX =1
    dTyM = 0.0
    
    K11,K21 = solve(TM, dTxM, dTyM)
    print("flux along x dir = ", K11)
    print("flux along y dir = ", K21)    
    
    # -------------------------------------------------

    TM   = TM0
    dTxM = 0.0
    dTyM = 1.0   # <= gradY =1

    K12,K22 = solve(TM, dTxM, dTyM)
    print("flux along x dir = ", K12)
    print("flux along y dir = ", K22)

    # -------------------------------------------------
    # try to calculate the first flux 
    # with the computed sensitivity matrix
    # -------------------------------------------------

    q1 = K11*dTxM0 + K12*dTyM0
    q2 = K21*dTxM0 + K22*dTyM0

    print()
    print("K=[[", K11, ",", K12,"]")
    print("   [", K21, ",", K22,"]]\n")
    
    print("q  (linear combination) =",q1,q2)
    print("qi (direct calculation) =",q1i,q2i)
    print(chrono)
    
    tests = CTests()
    tests.add(CTest('q-qi (x)', abs(q1-q1i), 0.0, 1e-6))
    tests.add(CTest('q-qi (y)', abs(q2-q2i), 0.0, 1e-6))
    tests.run()




if __name__ == "__main__":
    main()
