#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function
from builtins import str
from builtins import range
import katropton.CMAME.model as model

import katropton as m
import tbox
import tbox.gmsh as gmsh
import shlex
import subprocess
import os
import numpy as np

from fwk.wutils import parseargs

from PyTrilinos import Teuchos

import katropton.utilities as u


def evaluate_beam_test(msh, comm, test_case):

    directory = str(test_case)
    if not os.path.exists(directory):
        os.makedirs(directory)
    os.chdir(directory)

    work_dir = os.getcwd()
    args = parseargs()

    pbl = m.Problem(msh, comm)

    ensemble_size = 1

    # Units are in mm
    E = 70000 * np.ones(ensemble_size)  # 70e9 Pa = 70000 MPa
    nu = 0.35 * np.ones(ensemble_size)
    k = 0.
    d = 0.
    m.Medium(pbl, "body", "test", E, nu, k, d)

    m.Dirichlet(pbl, "clamped 1", "Clamped", 1, 0., 1, 0., 1, 0., 0, 0.,
                ensemble_size)
    m.Dirichlet(pbl, "clamped 2", "Clamped", 1, 0., 1, 0., 1, 0., 0, 0.,
                ensemble_size)

    f = 70 * np.ones(ensemble_size)  # 700e6 Pa = 700 MPa
    zero = np.zeros(ensemble_size)
    one = np.ones(ensemble_size)
    m.Neumann(pbl, 'load 1', 'load', 0, zero, 1, -f, 0, zero, 0, one,
              ensemble_size)
    m.Neumann(pbl, 'load 2', 'load', 0, zero, 1, -f, 0, zero, 0, one,
              ensemble_size)

    solverList = Teuchos.ParameterList()
    solverList['Ensemble Convergence Tolerance'] = 1e-8
    solverList['Maximum Iterations'] = 200
    solverList['Verbosity'] = 33
    solverList['Flexible Gmres'] = False
    solverList['Output Style'] = 1
    solverList['Output Frequency'] = 1

    if test_case == 1:
        mueluParams = Teuchos.ParameterList()
        mueluParams['number of equations'] = 3
        mueluParams['max levels'] = 1
        mueluParams['coarse: type'] = "RELAXATION"

        mueluParamsSub2 = Teuchos.ParameterList()
        mueluParamsSub2['relaxation: type'] = "Gauss-Seidel"
        mueluParamsSub2['relaxation: sweeps'] = 3
        mueluParamsSub2['relaxation: damping factor'] = 0.75

        mueluParams['verbosity'] = "high"

        mueluParams['coarse: params'] = mueluParamsSub2

        solverList['mueluParams'] = mueluParams

    elif test_case == 2:
        file_dir = os.path.dirname(__file__)
        muelu_xml_in = file_dir + '/../../preconditioners/gs_lvl.xml.in'

        old_strings = ['$BLOCK_00_STRIDING_INFO', '$BLOCK_11_STRIDING_INFO',
                       '$BLOCK_00_GS_sweeps', '$BLOCK_00_GS_damping',
                       '$SCHUR_omega', '$SimpleSmoother_damping', "$N_LVLS", "$N_SIZES", "$EXPORT_DATA"]
        new_strings = [
            '3',
            '3',
            '3',
            '0.75',
            '1',
            '1',
            '2',
            '18',
            "true"
        ]
        u.replace_strings(muelu_xml_in, work_dir + '/gs.xml',
                          old_strings, new_strings)
        solverList['use xml file'] = True
        solverList['MueLu xml file name'] = work_dir + '/gs.xml'

    elif test_case == 2.5:
        mueluParams = Teuchos.ParameterList()
        mueluParams['number of equations'] = 3
        mueluParams['max levels'] = 2
        mueluParams['smoother: type'] = "RELAXATION"

        mueluParamsSub2 = Teuchos.ParameterList()
        mueluParamsSub2['relaxation: type'] = "MT Gauss-Seidel"
        mueluParamsSub2['relaxation: type'] = "Gauss-Seidel"
        mueluParamsSub2['relaxation: sweeps'] = 3
        mueluParamsSub2['relaxation: damping factor'] = 0.75

        mueluParams['verbosity'] = "high"
        mueluParams['smoother: params'] = mueluParamsSub2
        mueluParams['coarse: type'] = "Klu"
        mueluParams['coarse: max size'] = 18
        #mueluParams['aggregation: min agg size'] = 5
        #mueluParams['aggregation: max selected neighbors'] = 1
        mueluParams['aggregation: export visualization data'] = True
        exportdata = Teuchos.ParameterList()
        exportdata['A'] = '{0,1}'
        exportdata['P'] = '{0}'
        exportdata['R'] = '{0}'
        exportdata['Aggregates'] = '{0}'
        mueluParams['export data'] = exportdata
        solverList['mueluParams'] = mueluParams

    else:
        norm = tbox.Pt(0, 1, 0)
        cont = m.Contact(pbl, "conctact 1", "contact", norm)
        cont.setInitialyOpen()
        cont.setNoUpdate()
        if test_case == 4 or test_case == 6:
            cont.setSticking()
        cont.setMaster(pbl, "conctact 2", norm)

        file_dir = os.path.dirname(__file__)

        if test_case == 3 or test_case == 4:
            muelu_xml_in = file_dir + '/../../preconditioners/SIMPLE_gs_direct.xml.in'

            old_strings = ['$BLOCK_00_STRIDING_INFO', '$BLOCK_11_STRIDING_INFO',
                           '$BLOCK_00_GS_sweeps', '$BLOCK_00_GS_damping',
                           '$SCHUR_omega', '$SimpleSmoother_damping']
            if test_case == 4:
                new_strings = [
                    '3',
                    '3',
                    '3',
                    '0.75',
                    '1',
                    '1'
                ]
            else:
                new_strings = [
                    '3',
                    '1',
                    '3',
                    '0.75',
                    '1',
                    '1'
                ]
        else:
            muelu_xml_in = file_dir + '/../../preconditioners/SIMPLE_gs_direct_lvl.xml.in'

            old_strings = ['$BLOCK_00_STRIDING_INFO', '$BLOCK_11_STRIDING_INFO',
                           '$BLOCK_00_GS_sweeps', '$BLOCK_00_GS_damping',
                           '$SCHUR_omega', '$SimpleSmoother_damping', "$N_LVLS", "$N_SIZES", "$EXPORT_DATA"]
            if test_case == 6:
                new_strings = [
                    '3',
                    '3',
                    '3',
                    '0.75',
                    '1',
                    '1',
                    '2',
                    '18',
                    "true"
                ]
            else:
                new_strings = [
                    '3',
                    '1',
                    '3',
                    '0.75',
                    '1',
                    '1',
                    '2',
                    '18',
                    "true"
                ]
        u.replace_strings(muelu_xml_in, work_dir + '/SIMPLE_gs_direct.xml',
                          old_strings, new_strings)
        solverList['MueLu xml file name'] = work_dir + '/SIMPLE_gs_direct.xml'

    solverList['convert MueLu xml file'] = True
    solverList['Create Preconditioned Matrix'] = True

    solverList['Print Teuchos timers'] = True

    solverList['Use preconditioner'] = True

    solverList['type'] = 'BlockGmres'
    solverList['Maximum active set iteration'] = 1

    solverList['Write matrix and vectors'] = True
    solverList['Write txt files'] = True

    slv = m.IterativeSolver(pbl, args.k, solverList, 3, ensemble_size)
    slv.start()
    f = open('timers.txt', 'w')
    print(slv.getTimers(), file=f)
    f.closed

    os.chdir('..')


def main():
    comm, rank, size = m.utilities.import_MPI()

    args = parseargs()

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    # Units in the meshes are in mm
    msh = gmsh.MeshLoader('beam.geo', file_dir).execute()

    for i in range(1, 7):
        evaluate_beam_test(msh, comm, i)


if __name__ == "__main__":
    main()
