from __future__ import print_function
from __future__ import division
from builtins import str
from builtins import range
from past.utils import old_div
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import axes3d
import matplotlib.tri as mtri
from matplotlib import cm

import os

from katropton.Juelich.UQ.kde import *
import katropton.utilities as u

import scipy as sc


def build_exponents_L2(order):
    exponents = np.zeros(((order + 1)**2, 2), dtype=int)
    exponents[:, 0], exponents[:, 1] = np.divmod(np.arange((order + 1)**2),
                                                 order + 1)
    return exponents


def build_exponents_L1(order):
    exponents = build_exponents_L2(order)
    tmp = exponents[np.where((exponents.T[0] + exponents.T[1]) <= order), :]
    return tmp[0, :, :]


def build_PCE_system(x,
                     y,
                     weights,
                     exponents,
                     p0=np.array([0, 0]),
                     scale=np.array([1, 1]),
                     monomial=True):
    n = len(x)
    m = exponents.shape[0]
    Z = np.zeros((n, m))
    W = np.zeros((n, n))

    if not monomial:
        u = 1. - (x - 30.) / 76.
        v = -1. + (y - 30.) / 76.

    for i in range(0, m):
        if monomial:
            Z[:, i] = (old_div((x - p0[0]), scale[0]))**exponents[i, 0] * (
                old_div((y - p0[1]), scale[1]))**exponents[i, 1]
        else:
            p_m_x = np.ones(u.shape)
            index = np.argwhere(v < 1)
            p_m_x[index] = old_div(
                (2 * u[index] + v[index] + 1), (1 - v[index]))
            p_m = sc.special.eval_legendre(exponents[i, 0], p_m_x)
            p_n = sc.special.eval_jacobi(exponents[i, 1],
                                         2 * exponents[i, 0] + 1, 0, v)
            Z[:, i] = p_m * (old_div((1 - v), 2))**exponents[i, 0] * p_n
    for i in range(0, n):
        W[i, i] = weights[i]

    ZTW = Z.T.dot(W)
    ZTWZ = ZTW.dot(Z)

    ZTWZinv = np.linalg.pinv(ZTWZ)
    Ainv = ZTWZinv.dot(ZTW)
    return Ainv


def evaluate_PCE(x,
                 y,
                 coefficients,
                 exponents,
                 p0=np.array([0, 0]),
                 scale=np.array([1, 1]),
                 monomial=True):
    n = len(x)
    m = exponents.shape[0]
    z = np.zeros((n, ))

    if not monomial:
        u = 1. - (x - 30.) / 76.
        v = -1. + (y - 30.) / 76.

    for i in range(0, m):
        if monomial:
            z = z + coefficients[i] * (
                old_div((x - p0[0]), scale[0]))**exponents[i, 0] * (
                    old_div((y - p0[1]), scale[1]))**exponents[i, 1]
        else:
            p_m_x = np.ones(u.shape)
            index = np.argwhere(v < 1)
            p_m_x[index] = old_div(
                (2 * u[index] + v[index] + 1), (1 - v[index]))
            p_m = sc.special.eval_legendre(exponents[i, 0], p_m_x)
            p_n = sc.special.eval_jacobi(exponents[i, 1],
                                         2 * exponents[i, 0] + 1, 0, v)
            z = z + coefficients[i] * p_m * (
                old_div((1 - v), 2))**exponents[i, 0] * p_n
    return z


def triangular_mesh(n_x, n_y, x_min=30, x_max=180, y_min=30, y_max=180):
    x = np.linspace(x_min, x_max, n_x, endpoint=True)
    y = np.linspace(y_min, y_max, n_y, endpoint=True)
    x, y = np.meshgrid(x, y)
    x = x.flatten()
    y = y.flatten()

    x_TRI = x[np.where(x >= y)]
    y_TRI = y[np.where(x >= y)]

    tri = mtri.Triangulation(x_TRI, y_TRI)

    return x_TRI, y_TRI, tri


def compute_coeff(Ainv, z):
    coefficients = Ainv.dot(z)
    return coefficients


def write_patch(tri, z, filename):
    f = open(filename + "_data.txt", "w+")
    print("x y c", file=f)
    for i in range(0, tri.triangles.shape[0]):
        for j in range(0, 3):
            index = tri.triangles[i, j]
            print(str(tri.x[index]) + " " + str(
                tri.y[index]) + " " + str(z[index]), file=f)
        print(" ", file=f)

    f.close()
    f = open(filename + "_range.txt", "w+")
    print(np.amin(z), file=f)
    print(np.amax(z), file=f)
    f.close()


def compute_qoi(base_dir, ensemble_size, ensemble_index, l):
    OAX, OAY = 0.0595, 0.017
    npzfile = np.load(base_dir + str(ensemble_size) + '/' +
                      str(ensemble_index) + '/mirror_data_all.npz')
    x = npzfile['x']
    y = npzfile['y']
    z = 0. * x
    dz = npzfile['dz']
    T = npzfile['T']
    tri = npzfile['tri']
    optical_coefficients, dz_residual = u.compute_RBM_Curvature_Irreg(
        x, y, z, tri, dz[:, l], "m", np.array([OAX, OAY]))
    print(optical_coefficients)
    return [
        np.max(T[:, l]),
        np.min(T[:, l]),
        np.mean(T[:, l]), 1000 * np.mean(dz[:, l]), optical_coefficients[0],
        optical_coefficients[1], optical_coefficients[2],
        optical_coefficients[3], optical_coefficients[4]
    ]


def read_data(order, base_dir, ax, display=False):

    result_file = 'qois_' + str(order) + '.npz'
    if os.path.isfile(result_file):
        loaded = np.load(result_file)
        qois = loaded['qois']
        samples = loaded['samples']
        weights = loaded['weights']
        if display:
            #fig = plt.figure()
            #ax = fig.gca(projection='3d')
            ax.scatter(samples[:, 0], samples[:, 1], qois[:, -2])

    else:
        file_dir = os.path.dirname(__file__)
        loaded = np.load(file_dir + '/../samples_files/Xiao_' + str(order) +
                         '.npz')
        samples = loaded['samples']
        weights = loaded['weights']
        N = samples.shape[0]

        order_base_dir = base_dir + str(order) + '/1/'

        n_qois = 9

        qois = np.zeros((N, n_qois))

        for i in range(0, N):

            loaded = np.load(order_base_dir + str(i) + '/mirror_data_all.npz')
            T = loaded['T']
            loaded = np.load(order_base_dir + str(i) +
                             '/mirror_optical_results_0.npz')
            optical_coefficients = loaded['optical_coefficients']

            #qois[i, 0] = np.min(T)
            #qois[i, 1] = np.mean(T)
            #qois[i, 2] = np.max(T)
            #qois[i, 3:8] = optical_coefficients

            qois[i, :] = compute_qoi(base_dir + str(order) + '/', 1, i, 0)

        np.savez('qois_' + str(order),
                 qois=qois,
                 samples=samples,
                 weights=weights)

    return qois, samples, weights


def generate_random_samples(N):
    tmp = np.random.rand(3*N, 2)
    index = np.argwhere(tmp[:, 0] >= tmp[:, 1])

    tmp_2 = tmp[index[0:N], :]

    samples = 30. + 150*tmp_2

    samples = np.reshape(samples, (N, 2))

    return samples


def main():
    base_dir = "/Users/kimliegeois/Desktop/Juelich/order_2_30/order_"

    orders = np.arange(2, 32, 2)
    n_orders = len(orders)

    x_0 = [30, 30]
    scale = [150, 150]

    monomial = False

    x_PCE = np.array([180., 30.])
    y_PCE = np.array([180., 30.])

    #value = np.array([0.0109, -0.04723])
    #which_qoi = -2

    value = np.array([0.00688, 0.00742])
    which_qoi = -1

    #value = np.array([105.13469634968574, 172.45597099271995])
    #which_qoi = 1

    #value = np.array([120.26643788294894, 189.0950734335578])
    #which_qoi = 0

    error = np.zeros((n_orders, 2))

    cauchy_error = np.zeros((n_orders-1, ))

    N_vec = np.zeros((n_orders, 1))
    N_vec_2 = np.zeros((n_orders, 1))

    fig = plt.figure()
    ax = fig.gca(projection='3d')

    n_real = 100000

    rand_samples = generate_random_samples(n_real)

    rand_z = np.zeros((n_real, len(orders)))

    for i in range(0, n_orders):
        order = orders[i]
        qois, samples, weights = read_data(order, base_dir, ax, True)

        N = len(samples[:, 0])

        N_vec[i] = N

        d = old_div(order, 2)

        x = samples[0:N, 0]
        y = samples[0:N, 1]

        exponents = build_exponents_L1(d)
        Ainv = build_PCE_system(x,
                                y,
                                weights,
                                exponents,
                                x_0,
                                scale,
                                monomial=monomial)

        coefficients = compute_coeff(Ainv, qois[:, which_qoi])

        N_vec_2[i] = len(coefficients)

        z_PCE = evaluate_PCE(x_PCE,
                             y_PCE,
                             coefficients,
                             exponents,
                             x_0,
                             scale,
                             monomial=monomial)

        rand_z[:, i] = evaluate_PCE(rand_samples[:, 0],
                                    rand_samples[:, 1],
                                    coefficients,
                                    exponents,
                                    x_0,
                                    scale,
                                    monomial=monomial)

        error[i, 0] = np.abs(z_PCE[0] - value[0])
        error[i, 1] = np.abs(z_PCE[1] - value[1])

        if i != 0:
            cauchy_error[i -
                         1] = np.mean(np.power(rand_z[:, i]-rand_z[:, i-1], 2))

    print(old_div(error[:, 0], value[0]))
    print(old_div(error[:, 1], value[1]))

    tmp = np.zeros((len(orders), 2))
    tmp[:, 0] = old_div(orders, 2)

    tmp[:, 1] = np.abs(old_div(error[:, 0], value[0]))
    np.savetxt('birr.txt', tmp)

    tmp[:, 1] = np.abs(old_div(error[:, 1], value[1]))
    np.savetxt('airr.txt', tmp)

    plt.figure()
    plt.plot(old_div(orders, 2), np.abs(old_div(error[:, 0], value[0])))
    plt.plot(old_div(orders, 2), np.abs(old_div(error[:, 1], value[1])))

    plt.figure()
    plt.plot(old_div(orders, 2), N_vec)
    plt.plot(old_div(orders, 2), N_vec_2)

    print(cauchy_error)
    plt.figure()
    plt.plot(cauchy_error)

    plt.figure()
    plt.semilogy(cauchy_error)

    tmp = np.zeros((len(orders)-1, 2))
    tmp[:, 0] = np.arange(2, 16, 1)
    tmp[:, 1] = cauchy_error
    np.savetxt('cauchy_error.txt', tmp)

    plt.figure()
    plt.semilogy(old_div(orders, 2), np.abs(old_div(error[:, 0], value[0])))
    plt.semilogy(old_div(orders, 2), np.abs(old_div(error[:, 1], value[1])))
    plt.show()


if __name__ == "__main__":
    main()
