#include "ContactsList.h"

#include "NodesList.h"

#include "wProblem.h"
#include "wTag.h"
#include "wMshData.h"
#include "wElement.h"
#include "wNode.h"
#include "wMedium.h"
#include "wDirichlet.h"
#include "wContact.h"

using namespace katropton;
using namespace tbox;

/**
 * @brief ContactsList constructor
 * 
 * This constructor allocates 11 Kokkos::View which store all the required information related to each contact interface, including:
 * <ul>
 *  <li> all the slaves and master nodes and elements IDs,
 *  <li> all the normal of the surfaces,
 *  <li> type of Lagrange multiplier shape functions,
 *  <li> type of the interface (contact, Signorini, sticking contact, mesh tying),
 *  <li> initial status of the Lagrange multipliers and whether they should be updated.
 * </ul>
 * 
 * To do so, we need the information regarding the contact interfaces (stored into the katropton::Problem object), the information on the 
 * distribution of the nodes (stored into the katropton::Map object), and the information on the nodes and elements (stored into the 
 * katropton::NodesList and katropton::ElementList objects).
 * 
 * Arguments:
 * <ul>
 *  <li> pbl: a Problem object,
 *  <li> map: an RCP to a Map object,
 *  <li> elementsList: an RCP to an ElementList object,
 *  <li> nodesList: an RCP to a NodesList object.
 * </ul>
 */
ContactsList::ContactsList(Problem &pbl, Teuchos::RCP<Map> map, Teuchos::RCP<ElementsList> elementsList, Teuchos::RCP<NodesList> nodesList)
{
    typedef Map::local_ordinal_type local_ordinal_type;
    typedef Map::global_ordinal_type global_ordinal_type;

    contacts_number = pbl.Contacts.size();
    local_ordinal_type maxNumSlaveNodesPerContactFace = 0;
    local_ordinal_type maxNumMasterNodesPerContactFace = 0;
    local_ordinal_type maxNumSlaveElemsPerContactFace = 0;
    local_ordinal_type maxNumMasterElemsPerContactFace = 0;
    local_ordinal_type maxNumSlaveDirichletNodesPerContactFace = 0;

    for (auto i = 0; i < contacts_number; ++i)
    {
        std::vector<global_ordinal_type> myContactSlaveNodes;
        std::vector<global_ordinal_type> myContactSlaveElems;
        for (auto j = 0; j < pbl.Contacts[i]->slave_elems.size(); ++j)
        {
            local_ordinal_type e = map->mapElems->getLocalElement(pbl.Contacts[i]->slave_elems[j]->no - 1);
            if (elementsList->getElementType(e) == static_cast<int>(ELTYPE::QUAD4))
                ;
            else if (elementsList->getElementType(e) == static_cast<int>(ELTYPE::TRI3))
                ;
            else
                continue;
            myContactSlaveElems.push_back(e);
            for (auto n1 = 0; n1 < elementsList->getElementSize(e); ++n1)
            {
                global_ordinal_type i_global = map->mapNodesWO->getGlobalElement(elementsList->getElementNode(e, n1));

                if (std::find(myContactSlaveNodes.begin(), myContactSlaveNodes.end(), i_global) == myContactSlaveNodes.end())
                    myContactSlaveNodes.push_back(i_global);
            }
        }

        int myNumContactSlaveNodes = myContactSlaveNodes.size();
        int myNumContactSlaveElems = myContactSlaveElems.size();

        int myNumContactSlaveDirichletNodes = pbl.Contacts[i]->slave_dirichlet_nodes.size();

        std::vector<global_ordinal_type> myContactMasterNodes;
        std::vector<global_ordinal_type> myContactMasterElems;
        for (auto j = 0; j < pbl.Contacts[i]->master_elems.size(); ++j)
        {
            local_ordinal_type e = map->mapElems->getLocalElement(pbl.Contacts[i]->master_elems[j]->no - 1);
            if (elementsList->getElementType(e) == static_cast<int>(ELTYPE::QUAD4))
                ;
            else if (elementsList->getElementType(e) == static_cast<int>(ELTYPE::TRI3))
                ;
            else if (elementsList->getElementType(e) == static_cast<int>(ELTYPE::LINE2))
                ;
            else
                continue;
            myContactMasterElems.push_back(e);
            for (auto n1 = 0; n1 < elementsList->getElementSize(e); ++n1)
            {
                global_ordinal_type i_global = map->mapNodesWO->getGlobalElement(elementsList->getElementNode(e, n1));
                if (std::find(myContactMasterNodes.begin(), myContactMasterNodes.end(), i_global) == myContactMasterNodes.end())
                    myContactMasterNodes.push_back(i_global);
            }
        }

        int myNumContactMasterNodes = myContactMasterNodes.size();
        int myNumContactMasterElems = myContactMasterElems.size();

        if (maxNumSlaveNodesPerContactFace < myNumContactSlaveNodes)
            maxNumSlaveNodesPerContactFace = myNumContactSlaveNodes;
        if (maxNumSlaveElemsPerContactFace < myNumContactSlaveElems)
            maxNumSlaveElemsPerContactFace = myNumContactSlaveElems;

        if (maxNumSlaveDirichletNodesPerContactFace < myNumContactSlaveDirichletNodes)
            maxNumSlaveDirichletNodesPerContactFace = myNumContactSlaveDirichletNodes;

        if (maxNumMasterNodesPerContactFace < myNumContactMasterNodes)
            maxNumMasterNodesPerContactFace = myNumContactMasterNodes;
        if (maxNumMasterElemsPerContactFace < myNumContactMasterElems)
            maxNumMasterElemsPerContactFace = myNumContactMasterElems;
    }

    contacts_slave_nodes_size = 0;
    contacts_slave_elems_size = 1;
    contacts_master_nodes_size = 2;
    contacts_master_elems_size = 3;
    contacts_slave_normal = 4;
    contacts_master_normal = 7;
    contacts_slave_nodes = 10;
    contacts_slave_elems = contacts_slave_nodes + maxNumSlaveNodesPerContactFace;
    contacts_master_nodes = contacts_slave_elems + maxNumSlaveElemsPerContactFace;
    contacts_master_elems = contacts_master_nodes + maxNumMasterNodesPerContactFace;

    size_t numMaxDirichletSlaveNode = 1;
    for (auto i = 0; i < contacts_number; ++i)
        if (numMaxDirichletSlaveNode < pbl.Contacts[i]->slave_dirichlet_nodes.size())
            numMaxDirichletSlaveNode = pbl.Contacts[i]->slave_dirichlet_nodes.size();

    size_t numMaxClosedNode = 1;
    for (auto i = 0; i < contacts_number; ++i)
        if (numMaxClosedNode < pbl.Contacts[i]->initially_closed_nodes.size())
            numMaxClosedNode = pbl.Contacts[i]->initially_closed_nodes.size();

    Kokkos::resize(contacts, contacts_number, contacts_master_elems + maxNumMasterElemsPerContactFace);
    Kokkos::resize(contacts_normal, contacts_number, 6);
    Kokkos::resize(shape_function, contacts_number);
    Kokkos::resize(signorini, contacts_number);
    Kokkos::resize(initially_open, contacts_number);
    Kokkos::resize(no_update, contacts_number);
    Kokkos::resize(sticking, contacts_number);
    Kokkos::resize(mesh_tying, contacts_number);
    Kokkos::resize(contacts_scale, contacts_number);
    Kokkos::resize(contacts_dirichlet, contacts_number, numMaxDirichletSlaveNode + 1);
    Kokkos::resize(initially_closed_nodes, contacts_number, numMaxClosedNode + 1);

    for (auto i = 0; i < contacts_number; ++i)
    {
        contacts_dirichlet(i, 0) = pbl.Contacts[i]->slave_dirichlet_nodes.size();
        for (auto j = 0; j < contacts_dirichlet(i, 0); ++j)
            contacts_dirichlet(i, j + 1) = pbl.Contacts[i]->slave_dirichlet_nodes[j];
    }

    for (auto i = 0; i < contacts_number; ++i)
    {
        initially_closed_nodes(i, 0) = pbl.Contacts[i]->initially_closed_nodes.size();
        for (auto j = 0; j < initially_closed_nodes(i, 0); ++j)
            initially_closed_nodes(i, j + 1) = pbl.Contacts[i]->initially_closed_nodes[j];
    }

    for (auto i = 0; i < contacts_number; ++i)
    {
        int myContactSlaveNodes_id = 0;
        int myContactSlaveElems_id = 0;

        std::vector<global_ordinal_type> myContactSlaveNodes;

        for (auto j = 0; j < pbl.Contacts[i]->slave_elems.size(); ++j)
        {
            local_ordinal_type e = map->mapElems->getLocalElement(pbl.Contacts[i]->slave_elems[j]->no - 1);
            if (elementsList->getElementType(e) == static_cast<int>(ELTYPE::QUAD4))
                ;
            else if (elementsList->getElementType(e) == static_cast<int>(ELTYPE::TRI3))
                ;
            else
                continue;
            contacts(i, contacts_slave_elems + myContactSlaveElems_id) = e;
            ++myContactSlaveElems_id;
            for (auto n1 = 0; n1 < elementsList->getElementSize(e); ++n1)
            {
                global_ordinal_type i_global = map->mapNodesWO->getGlobalElement(elementsList->getElementNode(e, n1));
                if (std::find(myContactSlaveNodes.begin(), myContactSlaveNodes.end(), i_global) == myContactSlaveNodes.end())
                {
                    myContactSlaveNodes.push_back(i_global);
                    ++myContactSlaveNodes_id;
                }
            }
        }
        for (auto j = 0; j < myContactSlaveNodes.size(); ++j)
            contacts(i, contacts_slave_nodes + j) = myContactSlaveNodes[j];

        contacts(i, contacts_slave_nodes_size) = myContactSlaveNodes_id;
        contacts(i, contacts_slave_elems_size) = myContactSlaveElems_id;

        int myContactMasterNodes_id = 0;
        int myContactMasterElems_id = 0;

        std::vector<global_ordinal_type> myContactMasterNodes;

        for (auto j = 0; j < pbl.Contacts[i]->master_elems.size(); ++j)
        {
            local_ordinal_type e = map->mapElems->getLocalElement(pbl.Contacts[i]->master_elems[j]->no - 1);
            if (elementsList->getElementType(e) == static_cast<int>(ELTYPE::QUAD4))
                ;
            else if (elementsList->getElementType(e) == static_cast<int>(ELTYPE::TRI3))
                ;
            else if (elementsList->getElementType(e) == static_cast<int>(ELTYPE::LINE2))
                ;
            else
                continue;
            contacts(i, contacts_master_elems + myContactMasterElems_id) = e;
            ++myContactMasterElems_id;
            for (auto n1 = 0; n1 < elementsList->getElementSize(e); ++n1)
            {
                global_ordinal_type i_global = map->mapNodesWO->getGlobalElement(elementsList->getElementNode(e, n1));
                if (std::find(myContactMasterNodes.begin(), myContactMasterNodes.end(), i_global) == myContactMasterNodes.end())
                {
                    myContactMasterNodes.push_back(i_global);
                    ++myContactMasterNodes_id;
                }
            }
        }

        for (auto j = 0; j < myContactMasterNodes.size(); ++j)
            contacts(i, contacts_master_nodes + j) = myContactMasterNodes[j];

        contacts(i, contacts_master_nodes_size) = myContactMasterNodes_id;
        contacts(i, contacts_master_elems_size) = myContactMasterElems_id;
    }

    for (auto i = 0; i < contacts_number; ++i)
    {

        for (auto j = 0; j < 3; ++j)
        {
            contacts(i, contacts_slave_normal + j) = pbl.Contacts[i]->slave_normal.x[j];
            contacts(i, contacts_master_normal + j) = pbl.Contacts[i]->master_normal.x[j];

            contacts_normal(i, j) = pbl.Contacts[i]->slave_normal.x[j];
            contacts_normal(i, j + 3) = pbl.Contacts[i]->master_normal.x[j];
        }

        if (pbl.Contacts[i]->standard_shape_functions)
            shape_function(i) = 1;
        else
            shape_function(i) = 0;

        if (pbl.Contacts[i]->is_Signorini)
            signorini(i) = 1;
        else
            signorini(i) = 0;

        if (pbl.Contacts[i]->mesh_tying)
            mesh_tying(i) = 1;
        else
            mesh_tying(i) = 0;

        if (pbl.Contacts[i]->no_Update)
            no_update(i) = 1;
        else
            no_update(i) = 0;

        if (pbl.Contacts[i]->sticking)
        {
            sticking(i) = true;
            at_least_one_sticking = true;
        }
        else
            sticking(i) = false;

        if (pbl.Contacts[i]->initial_closed)
            initially_open(i) = 0;
        else
            initially_open(i) = 1;

        contacts_scale(i) = pbl.Contacts[i]->scale;
    }
}
