namespace katropton
{

/**
 * @brief DirichletList constructor
 *
 * This constructor allocates 2 Kokkos::View which store all the required information related to each Dirichlet BC, including:
 * <ul>
 *  <li> all the Dirichlet nodes,
 *  <li> all the values of the Dirichlet BC.
 * </ul>
 *
 * Arguments:
 * <ul>
 *  <li> pbl: a Problem object,
 *  <li> map: an RCP to a Map object,
 *  <li> elementsList: an RCP to an ElementList object.
 * </ul>
 */
template <typename scalar>
DirichletList<scalar>::DirichletList(Problem &pbl, Teuchos::RCP<Map> map, Teuchos::RCP<ElementsList> elementsList)
{
    typedef Map::local_ordinal_type local_ordinal_type;
    typedef Map::global_ordinal_type global_ordinal_type;

    typedef EnsembleTraits<scalar> ET;
    const int ensemble_size = ET::size;

    local_ordinal_type numDPN = map->numPrimalDPN;
    dirichlet_number = pbl.duBCs.size();
    local_ordinal_type maxNumNodesPerBC = 0;

    std::vector<std::vector<global_ordinal_type>> bcNodes = {};

    for (auto i = 0; i < dirichlet_number; ++i)
    {
        std::vector<global_ordinal_type> tmpMyBcNodes = {};
        std::vector<global_ordinal_type> myBcNodes = {};

        for (auto j = 0; j < pbl.duBCs[i]->nodes.size(); ++j)
            tmpMyBcNodes.push_back(pbl.duBCs[i]->nodes[j]->row);

        std::sort(tmpMyBcNodes.begin(), tmpMyBcNodes.end());

        if (tmpMyBcNodes.size() >= 1)
            myBcNodes.push_back(tmpMyBcNodes[0]);

        for (auto ii = 1; ii < tmpMyBcNodes.size(); ++ii)
            if (tmpMyBcNodes[ii] != tmpMyBcNodes[ii - 1])
                myBcNodes.push_back(tmpMyBcNodes[ii]);

        bcNodes.push_back(myBcNodes);
        if (maxNumNodesPerBC < myBcNodes.size())
            maxNumNodesPerBC = myBcNodes.size();
    }

    dirichlet_size = 0;
    dirichlet_nodes = 1;
    dirichlet_nodesList = Kokkos::View<int **, Kokkos::LayoutRight>("R", dirichlet_number, dirichlet_nodes + maxNumNodesPerBC);

    for (auto i = 0; i < dirichlet_number; ++i)
    {
        dirichlet_nodesList(i, dirichlet_size) = bcNodes[i].size();
        for (auto j = 0; j < bcNodes[i].size(); ++j)
        {
            dirichlet_nodesList(i, dirichlet_nodes + j) = bcNodes[i][j];
        }
    }

    dirichlet_time = 0;
    dirichlet_dofs = 1;
    dirichlet_values = 1 + numDPN;
    dirichlet_values_list = Kokkos::View<scalar ***, Kokkos::LayoutRight>("R", dirichlet_number, 1, 1 + 2 * numDPN);
    for (auto i = 0; i < dirichlet_number; ++i)
    {
        dirichlet_values_list(i, 0, dirichlet_time) = 0.; //to change latter
        for (auto j = 0; j < numDPN; ++j)
            if (pbl.duBCs[i]->which_dof[j])
                dirichlet_values_list(i, 0, dirichlet_dofs + j) = 1;
            else
                dirichlet_values_list(i, 0, dirichlet_dofs + j) = 0;

        for (int j = 0; j < ensemble_size; ++j)
        {
            ET::coeff(dirichlet_values_list(i, 0, dirichlet_values + 0), j) = pbl.duBCs[i]->x_values[j];
            ET::coeff(dirichlet_values_list(i, 0, dirichlet_values + 1), j) = pbl.duBCs[i]->y_values[j];
            ET::coeff(dirichlet_values_list(i, 0, dirichlet_values + 2), j) = pbl.duBCs[i]->z_values[j];
            if (numDPN == 4)
                ET::coeff(dirichlet_values_list(i, 0, dirichlet_values + 3), j) = pbl.duBCs[i]->T_values[j];
        }
    }
}
}; // namespace katropton
