#include "Graph.h"

#include "wProblem.h"
#include "wMshData.h"
#include "wElement.h"
#include "wNode.h"

#include <utility>

#include <Kokkos_UnorderedMap.hpp>
#include <Kokkos_Sort.hpp>

#include "Tpetra_CrsMatrix.hpp"
#include <MatrixMarket_Tpetra.hpp>

#include "LinearSolver.h"

using namespace katropton;

int min(int a, int b)
{
    return (a <= b) ? a : b;
}

int max(int a, int b)
{
    return (a >= b) ? a : b;
}

/**
 * @brief Graph constructor
 *
 * This constructor computes the sparsity pattern of the primal matrix.
 * 
 * Arguments:
 * <ul>
 *  <li> map: an RCP to a Map object,
 *  <li> elementsList: an RCP to a ElementsList object.
 * </ul>
 */
Graph::Graph(Teuchos::RCP<Map> map, Teuchos::RCP<ElementsList> elementsList)
{
    using std::pair;
    using Teuchos::RCP;
    using Teuchos::rcp;
    using Teuchos::tuple;
    typedef Map::local_ordinal_type local_ordinal_type;
    typedef local_ordinal_type LO;
    typedef Map::global_ordinal_type global_ordinal_type;
    typedef Tpetra::CrsGraph<local_ordinal_type, global_ordinal_type> crs_graph_type;

    auto numMyElems = elementsList->getElementNumber();

    auto numPrimalDPN = map->numPrimalDPN;

    int numLclNodes = map->mapNodesWO->getNodeNumElements();
    int numLclRows = numPrimalDPN * numLclNodes;

    Kokkos::View<size_t *> rowCounts("row counts", numLclRows);

    const int maxConnections = 30;
    Kokkos::UnorderedMap<pair<int, int>, bool> nodenode(numLclNodes * maxConnections);

    size_t numLclEntries = 0;
    Kokkos::parallel_reduce(
        numMyElems,
        KOKKOS_LAMBDA(const LO e, size_t &curNumLclEntries) {
            const int element_size = elementsList->getElementSize(e);
            for (auto n1 = 0; n1 < element_size; ++n1)
                for (auto n2 = n1; n2 < element_size; ++n2)
                {
                    const LO lclNode1 = elementsList->getElementNode(e, n1);
                    const LO lclNode2 = elementsList->getElementNode(e, n2);

                    const pair<int, int> key(min(lclNode1, lclNode2), max(lclNode1, lclNode2));

                    auto result = nodenode.insert(key);

                    if (result.success())
                        for (auto i = 0; i < numPrimalDPN; ++i)
                            for (auto j = 0; j < numPrimalDPN; ++j)
                            {
                                const LO lclRow = numPrimalDPN * lclNode1 + i;
                                const LO lclCol = numPrimalDPN * lclNode2 + j;

                                if (lclRow < numLclRows)
                                {
                                    if (!(i == 3 && j != 3))
                                    {
                                        Kokkos::atomic_fetch_add(&rowCounts(lclRow), 1);
                                        curNumLclEntries++;
                                    }
                                }
                                if (lclNode1 != lclNode2 && lclCol < numLclRows)
                                {
                                    if (!(j == 3 && i != 3))
                                    {
                                        Kokkos::atomic_fetch_add(&rowCounts(lclCol), 1);
                                        curNumLclEntries++;
                                    }
                                }
                            }
                }
        },
        numLclEntries);

    Kokkos::View<size_t *> rowOffsets("row offsets", numLclRows + 1);
    Kokkos::parallel_scan(
        numLclRows,
        KOKKOS_LAMBDA(int irow, int &update, bool final) {
            // parallel scan is a multi-pass parallel pattern
            // In the ‘final’ pass ‘update’ has the prefix value
            if (final)
                rowOffsets(irow) = update;
            update += rowCounts(irow);
            if (final && numLclRows == irow + 1)
                rowOffsets(irow + 1) = update; // total non-zeros
        });

    Kokkos::deep_copy(rowCounts, static_cast<size_t>(0));

    Kokkos::View<LO *> colIndices("column indices", rowOffsets(numLclRows));

    Kokkos::parallel_for(
        nodenode.capacity(),
        KOKKOS_LAMBDA(int ientry) {
            if (nodenode.valid_at(ientry))
            {
                const pair<int, int> key = nodenode.key_at(ientry);
                const int lclNode1 = key.first;
                const int lclNode2 = key.second;

                for (auto i = 0; i < numPrimalDPN; ++i)
                    for (auto j = 0; j < numPrimalDPN; ++j)
                    {
                        const LO lclRow = numPrimalDPN * lclNode1 + i;
                        const LO lclCol = numPrimalDPN * lclNode2 + j;

                        if (lclRow < numLclRows)
                        {
                            if (!(i == 3 && j != 3))
                            {
                                const size_t count = Kokkos::atomic_fetch_add(&rowCounts(lclRow), 1);
                                colIndices(rowOffsets(lclRow) + count) = lclCol;
                            }
                        }

                        if (lclNode1 != lclNode2 && lclCol < numLclRows)
                        {
                            if (!(j == 3 && i != 3))
                            {
                                const size_t count = Kokkos::atomic_fetch_add(&rowCounts(lclCol), 1);
                                colIndices(rowOffsets(lclCol) + count) = lclRow;
                            }
                        }
                    }
            }
        });

    LinearSolver::getTimers()["graph: sort"].start();

    Kokkos::parallel_for(
        numLclRows,
        KOKKOS_LAMBDA(int lclRow) {
            auto currentColIndices = subview(colIndices, Kokkos::make_pair(rowOffsets(lclRow), rowOffsets(lclRow + 1)));
            LO *const lclColIndsRaw = currentColIndices.data();
            std::sort(lclColIndsRaw, lclColIndsRaw + rowOffsets(lclRow + 1) - rowOffsets(lclRow));
        });
    LinearSolver::getTimers()["graph: sort"].stop();

    AWO = rcp(new crs_graph_type(map->mapDofsWO, map->mapDofsWO, rowOffsets, colIndices));
    AWO->fillComplete();

    // Create a Tpetra vector for the right hand side (with overlaps).

    const Tpetra::Export<> exportAWO(map->mapDofsWO, map->mapDofs);

    A = rcp(new crs_graph_type(map->mapDofs, 0));
    A->doExport(*AWO, exportAWO, Tpetra::INSERT);
    A->fillComplete();
}
