#ifndef KATROPTON_LINEARSOLVER_H
#define KATROPTON_LINEARSOLVER_H

#include "katropton.h"
#include "wObject.h"
#include "wTimer.h"
#include "wTimers.h"
#include "wTag.h"
#include "wMedium.h"
#include "wNode.h"
#include "wMem.h"
#include "Teuchos_ParameterList.hpp"

#include "wProblem.h"
#include "wContact.h"
#include "wMshData.h"
#include "wDisplayHook.h"
#include "wElement.h"

// Tpetra
#include "Stokhos_Tpetra_MP_Vector.hpp"
#include "Stokhos_Tpetra_Utilities_MP_Vector.hpp"
#include "Tpetra_ConfigDefs.hpp"
#include "Tpetra_Map.hpp"
#include "Tpetra_MultiVector.hpp"
#include "Tpetra_Vector.hpp"
#include "Tpetra_CrsGraph.hpp"
#include "Tpetra_CrsMatrix.hpp"
#include "Stokhos_Tpetra_CG.hpp"

#include <Teuchos_Array.hpp>
#include <Teuchos_GlobalMPISession.hpp>
#include <Teuchos_oblackholestream.hpp>
#include <Teuchos_ScalarTraits.hpp>
#include <Teuchos_RCP.hpp>
#include "Teuchos_ParameterList.hpp"

#include "Stokhos_Amesos2_MP_Vector.hpp"
#include "Amesos2_Factory.hpp"

#include "Stokhos_Ifpack2_MP_Vector.hpp"
#include "Ifpack2_Factory.hpp"

#include "Belos_Tpetra_MP_Vector.hpp"
#include "BelosLinearProblem.hpp"
#include "BelosPseudoBlockGmresSolMgr.hpp"
#include "BelosPseudoBlockCGSolMgr.hpp"
#include "BelosBlockGmresSolMgr.hpp"

#include "Kokkos_ViewFactory.hpp"
#include "Kokkos_Vector.hpp"

#include <Epetra_Map.h>
#include "Epetra_MpiComm.h"

#include "BelosConfigDefs.hpp"

#include "wRandomField.h"

#include "wResults.h"
#include "wLazy.h"
#include "Map.h"

#include <map>

#include <limits>

#include "EnsembleTraits.h"

#include "ResultsDecl.hpp"

namespace katropton
{

/**
 * @brief Solver class which includes utilities such as timers and VTK writers.
 */
class KATROPTON_API LinearSolver
{
public:
#ifndef SWIG
    Problem &pbl;
#endif
    Teuchos::RCP<Teuchos::ParameterList> solverList;
    const int numPrimalDPN;
    const size_t minimum_ensemble_size;
    size_t active_set_iteration;
#ifndef SWIG
    DisplayHook *dhook;

#endif
public:
    /**
     * @brief LinearSolver constructor
     *
     * Arguments:
     * <ul>
     *  <li> _pbl: A Problem object,
     *  <li> _solverList: an RCP to a ParameterList object which describes the solver,
     *  <li> _numPrimalDPN: the number of degrees of freedom per node (without taking into account the Lagrange multipliers),
     *  <li> _minimum_ensemble_size: the used ensemble size (1, 8, 16, 24, or 32),
     * </ul>
     */
    LinearSolver(Problem &_pbl, Teuchos::RCP<Teuchos::ParameterList> _solverList,
                 int _numPrimalDPN,
                 size_t _minimum_ensemble_size) : pbl(_pbl),
                                                  solverList(_solverList),
                                                  numPrimalDPN(_numPrimalDPN),
                                                  minimum_ensemble_size(_minimum_ensemble_size)
    {
        dhook = NULL;
    }

    virtual ~LinearSolver() {}

    /**
    * @brief Manage timers.
    */
    static fwk::Timers &getTimers()
    {
        static fwk::Timers tms;
        return tms;
    }

    /**
     * @brief Set the GUI.
     */
    void setGUI(DisplayHook &hook)
    {
        dhook = &hook;
    }

protected:
    /**
     * @brief Write results on disk in VTK format.
     */
    template <typename scalar>
    void writeResults(
        Teuchos::RCP<Tpetra::Vector<scalar, Map::local_ordinal_type, Map::global_ordinal_type>> xPost,
        size_t numNodes,
        std::string name,
        size_t ast,
        bool final_result,
        bool write_txt,
        Teuchos::RCP<Map> map,
        Teuchos::RCP<RandomField<scalar, Kokkos::DefaultExecutionSpace>> random_field)
    {
        LinearSolver::getTimers()["write results"].start();

        typedef Map::local_ordinal_type LocalOrdinal;
        typedef Map::global_ordinal_type GlobalOrdinal;

        writeResultsVTK<scalar, LocalOrdinal, GlobalOrdinal>(pbl, xPost, map,
                                                             random_field,
                                                             name, ast, final_result, write_txt);

        LinearSolver::getTimers()["write results"].stop();
    }
};
} // namespace katropton

#endif //KATROPTON_LINEARSOLVER_H
