#ifndef KATROPTON_MATERIALSLIST_H
#define KATROPTON_MATERIALSLIST_H

#include "Stokhos_Sacado_Kokkos_MP_Vector.hpp"

#include "katropton.h"

#include "Map.h"
#include "wProblem.h"
#include "wTag.h"
#include "wMshData.h"
#include "wElement.h"
#include "wNode.h"
#include "wMedium.h"

#include "Material.h"

#include <Kokkos_Core.hpp>
//#include "Sacado.hpp"

#include <Tpetra_Map.hpp>
#include <Tpetra_Vector.hpp>
#include <Teuchos_RCP.hpp>
#include <Kokkos_ViewFactory.hpp>
#include <vector>

#include "EnsembleTraits.h"

namespace katropton
{

/**
 * @brief Class which stores the list of the used materials.
 */
template <typename scalar>
class MaterialsList
{
private:
    Kokkos::View<scalar ***, Kokkos::LayoutRight> materials;

    Map::local_ordinal_type materials_number;

public:
    Map::local_ordinal_type materials_T;
    Map::local_ordinal_type materials_E;
    Map::local_ordinal_type materials_nu;
    Map::local_ordinal_type materials_k;
    Map::local_ordinal_type materials_beta;

    MaterialsList(Problem &pbl, Teuchos::RCP<Map> map);

    Material<scalar> getMaterial(int current_material, int i_thread);
};

/**
 * @brief MaterialsList constructor
 * 
 * This constructor allocates 1 Kokkos::View which stores all the constitutive parameters of each material.
 * 
 * Arguments:
 * <ul>
 *  <li> pbl: A Problem object,
 *  <li> map: an RCP to a Map object..
 * </ul>
 */
template <typename scalar>
MaterialsList<scalar>::MaterialsList(Problem &pbl, Teuchos::RCP<Map> map)
{
    materials_number = pbl.media.size();

    materials_T = 0;
    materials_E = 1;
    materials_nu = 2;
    materials_k = 3;
    materials_beta = 4;

    typedef EnsembleTraits<scalar> ET;
    const int ensemble_size = ET::size;

    materials = Kokkos::View<scalar ***, Kokkos::LayoutRight>("R", materials_number, 1, 5);
    for (auto i = 0; i < materials_number; ++i)
    {
        materials(i, 0, materials_T) = (scalar)0; // !!! needed to be changed latter on !!!
        for (int j = 0; j < ensemble_size; ++j)
        {
            ET::coeff(materials(i, 0, materials_E), j) = (double)pbl.media[i]->E_vector[j];
            ET::coeff(materials(i, 0, materials_nu), j) = (double)pbl.media[i]->nu_vector[j];
            ET::coeff(materials(i, 0, materials_k), j) = (double)pbl.media[i]->k_vector[j];
            ET::coeff(materials(i, 0, materials_beta), j) = (double)pbl.media[i]->beta_vector[j];
        }
        materials(i, 0, materials_k) = (scalar)pbl.media[i]->k;
        materials(i, 0, materials_beta) = (scalar)pbl.media[i]->beta;
    }
}

/**
 * @brief Return a Material
 * 
 * Arguments:
 * <ul>
 *  <li> current_material: current material ID,
 *  <li> i_thread: thread ID.
 * </ul>
 */
template <typename scalar>
Material<scalar> MaterialsList<scalar>::getMaterial(int current_material, int i_thread)
{
    scalar E;
    scalar nu;
    scalar k;
    scalar beta;

    size_t materials_T = 0;

    typedef EnsembleTraits<scalar> ET;
    const int ensemble_size = ET::size;

    for (int j = 0; j < ensemble_size; ++j)
    {
        typename ET::value_type Tj = 0.;

        size_t index_0 = 0;
        size_t index_1 = materials.extent(1) - 1;

        if (Tj <= ET::coeff(materials(current_material, index_0, materials_T), j))
        {
            ET::coeff(E, j) = ET::coeff(materials(current_material, index_0, 1), j);
            ET::coeff(nu, j) = ET::coeff(materials(current_material, index_0, 2), j);
            ET::coeff(k, j) = ET::coeff(materials(current_material, index_0, 3), j);
            ET::coeff(beta, j) = ET::coeff(materials(current_material, index_0, 4), j);
        }
        else if (Tj > ET::coeff(materials(current_material, index_1, materials_T), j))
        {
            ET::coeff(E, j) = ET::coeff(materials(current_material, index_1, 1), j);
            ET::coeff(nu, j) = ET::coeff(materials(current_material, index_1, 2), j);
            ET::coeff(k, j) = ET::coeff(materials(current_material, index_1, 3), j);
            ET::coeff(beta, j) = ET::coeff(materials(current_material, index_1, 4), j);
        }
        else
        {
            while (index_1 != index_0 + 1)
            {
                size_t index_c = index_0 + (index_1 - index_0) / 2;

                if (Tj == ET::coeff(materials(current_material, index_c, materials_T), j))
                {
                    index_1 = index_c;
                    index_0 = index_1 - 1;
                }
                else if (Tj < ET::coeff(materials(current_material, index_c, materials_T), j))
                    index_1 = index_c;
                else
                    index_0 = index_c;
            }

            double coeff = (Tj - ET::coeff(materials(current_material, index_0, materials_T), j)) / (ET::coeff(materials(current_material, index_1, materials_T), j) - ET::coeff(materials(current_material, index_0, materials_T), j));

            ET::coeff(E, j) = ET::coeff(materials(current_material, index_0, 1), j) + coeff * (ET::coeff(materials(current_material, index_1, 1), j) - ET::coeff(materials(current_material, index_0, 1), j));
            ET::coeff(nu, j) = ET::coeff(materials(current_material, index_0, 2), j) + coeff * (ET::coeff(materials(current_material, index_1, 2), j) - ET::coeff(materials(current_material, index_0, 2), j));
            ET::coeff(k, j) = ET::coeff(materials(current_material, index_0, 3), j) + coeff * (ET::coeff(materials(current_material, index_1, 3), j) - ET::coeff(materials(current_material, index_0, 3), j));
            ET::coeff(beta, j) = ET::coeff(materials(current_material, index_0, 4), j) + coeff * (ET::coeff(materials(current_material, index_1, 4), j) - ET::coeff(materials(current_material, index_0, 4), j));
        }
    }

    Material<scalar> material(E, nu, k, beta);
    return material;
}
}; // namespace katropton

#endif //KATROPTON_MATERIALSLIST_H
