#ifndef KATROPTON_STRESSCOMPUTATION_H
#define KATROPTON_STRESSCOMPUTATION_H

#include "wHex8.h"
#include "wSfHex8.h"
#include "wGaussHex8.h"
#include "wCacheHex8.h"
#include "wMemHex8.h"

#include "wTetra4.h"
#include "wSfTetra4.h"
#include "wGaussTetra4.h"
#include "wCacheTetra4.h"
#include "wMemTetra4.h"

/**
 * @brief Those functions compute the stress and strain at nodes and in average per element.
 * They are currently implemented using Eigen data structure but should be refactored using Kokkos:view
 * such that it will be possible to use parallel for in the future.
 */

void strain_at_gp(size_t n_n, tbox::Mem &mem, tbox::Cache &cache, size_t a, const Eigen::MatrixXd &u_e, Eigen::MatrixXd &Strain)
{
    Eigen::MatrixXd const &J = mem.getJinv(a);
    Eigen::Matrix3d du_d_Phi = Eigen::Matrix3d::Zero();
    for (size_t i = 0; i < n_n; ++i)
    {
        std::vector<double> &dffi = cache.dff[a][i];
        for (auto k = 0; k < 3; ++k)
            for (auto j = 0; j < 3; ++j)
                du_d_Phi(j, k) += dffi[j] * u_e(k, i);
    }
    Strain += (J * du_d_Phi).transpose() * (J * du_d_Phi) * 0.5;
}

void compute_stress(size_t n_n, const Eigen::MatrixXd &Strain, Eigen::MatrixXd &Stress, const Eigen::MatrixXd &H, const Eigen::MatrixXd &D, const std::vector<double> &T, bool thermal)
{
    Stress = Eigen::Matrix3d::Zero();
    for (size_t i = 0; i < 3; ++i)
        for (size_t j = 0; j < 3; ++j)
            for (size_t k = 0; k < 3; ++k)
                for (size_t l = 0; l < 3; ++l)
                    Stress(i, j) += H(i * 3 + j, k * 3 + l) * Strain(k, l);

    if (thermal)
        for (size_t i = 0; i < 3; ++i)
            for (size_t j = 0; j < 3; ++j)
            {
                double tmp = 0.;
                for (size_t k = 0; k < n_n; ++k)
                    tmp += D(i, j) * T[k];
                tmp /= n_n;
                Stress(i, j) += tmp;
            }
}

void strain_stress_x_y_z(tbox::Element *elem, Eigen::MatrixXd &Strain, Eigen::MatrixXd &Stress, const Eigen::MatrixXd &H, const Eigen::MatrixXd &D, const std::vector<double> &x, const std::vector<double> &y, const std::vector<double> &z, const std::vector<double> &T, bool thermal = false)
{
    tbox::Cache &cache = elem->getVCache();
    tbox::Mem &mem = elem->getVMem();

    size_t n_gp = cache.dff.size();
    size_t n_n = elem->nodes.size();

    Eigen::MatrixXd u_e(3, n_n);

    for (size_t i = 0; i < n_n; ++i)
    {
        u_e(0, i) = x[i];
        u_e(1, i) = y[i];
        u_e(2, i) = z[i];
    }

    Strain = Eigen::Matrix3d::Zero();
    for (size_t a = 0; a < n_gp; ++a)
    {
        strain_at_gp(n_n, mem, cache, a, u_e, Strain);
    }
    Strain /= n_n;

    compute_stress(n_n, Strain, Stress, H, D, T, thermal);
}

void strain_stress_x_y_z_at_node(tbox::Element *elem, std::vector<Eigen::MatrixXd> &Strain_at_node, std::vector<Eigen::MatrixXd> &Stress_at_node, const Eigen::MatrixXd &H, const Eigen::MatrixXd &D, const std::vector<double> &x, const std::vector<double> &y, const std::vector<double> &z, const std::vector<double> &T, bool thermal = false)
{
    tbox::Cache &cache = elem->getVCache();
    tbox::Mem &mem = elem->getVMem();

    size_t n_gp = cache.dff.size();
    size_t n_n = elem->nodes.size();

    std::vector<Eigen::MatrixXd> Strain_at_gp(n_gp);
    std::vector<Eigen::MatrixXd> Stress_at_gp(n_gp);
    Eigen::MatrixXd u_e(3, n_n);
    Eigen::MatrixXd Strain(3, 3);

    for (auto &k : Strain_at_gp)
        k.resize(3, 3);
    for (auto &k : Stress_at_gp)
        k.resize(3, 3);

    for (size_t i = 0; i < n_n; ++i)
    {
        u_e(0, i) = x[i];
        u_e(1, i) = y[i];
        u_e(2, i) = z[i];
    }

    for (size_t a = 0; a < n_gp; ++a)
    {
        Strain = Eigen::Matrix3d::Zero();
        strain_at_gp(n_n, mem, cache, a, u_e, Strain);
        Strain_at_gp[a] = Strain;
        compute_stress(n_n, Strain, Stress_at_gp[a], H, D, T, thermal);
    }

    // Extrapolation at node points
    Eigen::MatrixXd values_at_gp(n_gp, 18);
    Eigen::MatrixXd values_at_node;

    for (size_t i = 0; i < 3; ++i)
        for (size_t j = 0; j < 3; ++j)
            for (size_t a = 0; a < n_gp; ++a)
            {
                values_at_gp(a, i * 3 + j) = Strain_at_gp[a](i, j);
                values_at_gp(a, 9 + i * 3 + j) = Stress_at_gp[a](i, j);
            }

    elem->gp2node(values_at_gp, values_at_node);

    for (size_t k = 0; k < n_n; ++k)
        for (size_t i = 0; i < 3; ++i)
            for (size_t j = 0; j < 3; ++j)
            {
                Strain_at_node[k](i, j) = values_at_node(k, i * 3 + j);
                Stress_at_node[k](i, j) = values_at_node(k, 9 + i * 3 + j);
            }
}

#endif //KATROPTON_STRESSCOMPUTATION_H