#include "wNeumann.h"
#include "wProblem.h"
#include "wTag.h"
#include "wElement.h"
#include "wNode.h"
#include <algorithm>
using namespace katropton;
using namespace tbox;

/**
 * @brief Neumann constructor
 * 
 * Arguments:
 * <ul>
 *  <li> pbl: a Problem object,
 *  <li> name: the name of the loaded surface of the msh,
 *  <li> _BC_name: the name of new Neumann BC (does not have to be a nametag of the msh),
 *  <li> _which_dof: a vector that specify which degree(s) of freedom are loaded,
 *  <li> _values: a vector that specify the values of the loads.
 * </ul>
 */
Neumann::Neumann(katropton::Problem &pbl,
                 std::string const &name,
                 std::string const &_BC_name,
                 std::vector<int> _which_dof,
                 std::vector<double> _values) : Group(pbl.msh, name), BC_name(_BC_name), which_dof(_which_dof), values(_values)
{
    for (auto e : tag->elems)
        for (auto n : e->nodes)
            nodes.push_back(n);

    pbl.nBCs.push_back(this);
}

/**
 * @brief Neumann constructor
 * 
 * Arguments:
 * <ul>
 *  <li> pbl: a Problem object,
 *  <li> name: the name of the loaded surface of the msh,
 *  <li> _BC_name: the name of new Neumann BC (does not have to be a nametag of the msh),
 *  <li> x: a int which specify whether the x displacement component is loaded,
 *  <li> v_x: value of the load,
 *  <li> y: a int which specify whether the y displacement component is loaded,
 *  <li> v_y: value of the load,
 *  <li> z: a int which specify whether the z displacement component is loaded,
 *  <li> v_z: value of the load,
 *  <li> T: a int which specify whether the temperature is loaded,
 *  <li> v_T: value of the load,
 *  <li> ensemble_size: the ensemble size.
 * </ul>
 */
Neumann::Neumann(Problem &pbl,
                 std::string const &name,
                 std::string const &_BC_name,
                 int x,
                 std::vector<double> v_x,
                 int y,
                 std::vector<double> v_y,
                 int z,
                 std::vector<double> v_z,
                 int T,
                 std::vector<double> v_T,
                 size_t ensemble_size) : Group(pbl.msh, name), BC_name(_BC_name)
{
    for (auto e : tag->elems)
        for (auto n : e->nodes)
            nodes.push_back(n);

    which_dof = {};
    which_dof.push_back(x);
    which_dof.push_back(y);
    which_dof.push_back(z);
    which_dof.push_back(T);

    values = {};
    for (auto i = 0; i < ensemble_size; ++i)
        values.push_back(v_x[i]);
    for (auto i = 0; i < ensemble_size; ++i)
        values.push_back(v_y[i]);
    for (auto i = 0; i < ensemble_size; ++i)
        values.push_back(v_z[i]);
    for (auto i = 0; i < ensemble_size; ++i)
        values.push_back(v_T[i]);

    pbl.nBCs.push_back(this);
}

/**
 * @brief Display information.
 */
void Neumann::write(std::ostream &out) const
{
    out << "Neumann BC on " << *tag << "\n";
    out << "With applied load of (";
    for (auto i = 0; i < (which_dof.size() - 1); ++i)
    {
        if (which_dof[i] == 1)
            out << values[i] << ", ";
        else
            out << "0, ";
    }

    if (which_dof[which_dof.size() - 1] == 1)
        out << values[which_dof.size() - 1] << ")\n";
    else
        out << "0)\n";
}
