#!/usr/bin/env python
# -*- coding: latin-1; -*-

from __future__ import print_function
from builtins import str
from builtins import range
import katropton as m
import tbox
import tbox.gmsh as gmsh
import katropton.utilities as u
from PyTrilinos import Teuchos
import shlex
import subprocess
import os
import numpy as np
import tarfile
import glob
from fwk.wutils import parseargs
from future.utils import native_str


def model(msh, E, p, ensemble_size, case_name, max_iter, use_Prec=False):
    comm, rank, size = m.utilities.import_MPI()
    file_dir = os.path.dirname(__file__)

    pbl = m.Problem(msh, comm)

    nu = 0.2 * np.ones(ensemble_size)
    k = 0.114
    d = 0.

    zero = np.zeros(ensemble_size)
    m.Medium(pbl, "Body 1", "test", E, nu, k, d)

    m.Dirichlet(pbl, "Surf 1 1", "Clamped", 1, 0., 1, 0., 1, 0., 1, 0.,
                ensemble_size)
    m.Neumann(pbl, "Surf 1 2", "heat fluxes", 0, zero, 0, zero, 1, p, 0, zero,
              ensemble_size)

    solverList = Teuchos.ParameterList()
    solverList['Ensemble Convergence Tolerance'] = 0.  # 10**(-8)
    solverList['Maximum Iterations'] = int(max_iter)
    solverList['Verbosity'] = 33
    solverList['Flexible Gmres'] = False
    solverList['Output Style'] = 1
    solverList['Output Frequency'] = 1

    solverList['Use preconditioner'] = use_Prec

    solverList['type'] = "BlockGmres"
    solverList['Maximum active set iteration'] = 1

    solverList['Print Teuchos timers'] = True
    solverList['Write vtk files'] = False

    if use_Prec:
        solverList['convert MueLu xml file'] = True
        solverList["use xml file"] = True
        solverList["MueLu xml file name"] = file_dir + '/prec_1.xml'

    slv = m.IterativeSolver(pbl, solverList, 3, ensemble_size)

    u.mkdir_MPI(case_name, comm, rank, size)
    os.chdir(case_name)

    if rank == 0:
        f = open('msh_size.txt', 'w')
        print(len(msh.nodes), file=f)
        f.closed

        np.savetxt('E.txt', E)
        np.savetxt('p.txt', p)

    slv.start()

    if rank == 0:
        f = open('timers.txt', 'w')
        print(slv.getTimers(), file=f)
        f.closed

    os.chdir('..')


def read_msh(i=1):
    comm, rank, size = m.utilities.import_MPI()

    mesh_name = native_str('1_cube_' + str(i) + '.msh')

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    if size != 1:
        msh = gmsh.MeshLoader(mesh_name, work_dir).execute(myrank=rank)
    else:
        msh = gmsh.MeshLoader(mesh_name, work_dir).execute()

    return msh


def main():
    comm, rank, size = m.utilities.import_MPI()

    args = parseargs()

    os.chdir('MPI_' + str(size))
    work_dir = os.getcwd()

    E_mean = 1.
    p_mean = 0.1

    ensemble_sizes = [1, 8, 16, 32]

    nCx = 9
    nCy = 9

    N = 100
    tmp = np.logspace(1, 5, N, endpoint=True, dtype=int)
    nCzs = tmp - 1

    i = np.loadtxt(work_dir + '/i.txt', dtype=int)
    max_iter = np.loadtxt(work_dir + '/m.txt', dtype=int)
    nCz = nCzs[i]
    msh = read_msh(i)
    nThreads = u.Initialize_Kokkos()
    for ensemble_size in ensemble_sizes:
        E = E_mean * np.ones(ensemble_size)
        p = p_mean * np.ones(ensemble_size)

        directory_case_id = 'test_' + str(i) + '_s_' + str(
            ensemble_size) + '_no_prec_m_' + str(max_iter)
        model(msh,
              E,
              p,
              ensemble_size,
              directory_case_id,
              max_iter,
              use_Prec=False)
        directory_case_id = 'test_' + str(i) + '_s_' + str(
            ensemble_size) + '_prec_m_' + str(max_iter)
        model(msh,
              E,
              p,
              ensemble_size,
              directory_case_id,
              max_iter,
              use_Prec=True)

    u.Finalize_Kokkos()


if __name__ == "__main__":
    main()
