#!/usr/bin/env python
# -*- coding: latin-1; -*-

from __future__ import print_function
from __future__ import division
from builtins import str
from builtins import range
from past.utils import old_div
from future.utils import native_str
import katropton as m
import tbox
import tbox.gmsh as gmsh
import shlex
import subprocess
import os
import numpy as np
from katropton.eigenvalues import *


def import_MPI():
    try:
        import mpi4py.MPI as mpi
        comm = mpi.COMM_WORLD
        rank = comm.rank
        size = comm.size
        name = mpi.Get_processor_name()
        status = mpi.Status()
        print('info: MPI found')
    except:
        comm = None
        rank = 0
        size = 1
        name = "noname"
        print('info: MPI not found => MPI disabled')
    return comm, rank, size


def Initialize_Kokkos():
    from fwk.wutils import parseargs
    args = parseargs()

    m.Kokkos_Initialize(args.k)
    return args.k


def Finalize_Kokkos():
    m.Kokkos_Finalize()


def mkdir_MPI(directory, comm, rank, size):
    # Create a new directory if rank == 0
    if rank == 0:
        if not os.path.exists(directory):
            os.makedirs(directory)

    comm.Barrier()


def read_mesh_MPI(geo_name, file_dir, work_dir, comm, rank, size, params=None):
    base_name = geo_name.partition(".")[0]
    geo_type = geo_name.partition(".")[2]

    mesh_name = base_name+'.msh'
    partitioned_mesh_name = native_str(base_name+'_'+str(size)+'.msh')

    if geo_type == "geo" and rank == 0:
        if params is not None:
            gmsh.MeshLoader(geo_name, file_dir).execute(**params)
        else:
            gmsh.MeshLoader(geo_name, file_dir).execute()

    if size > 1:
        comm.Barrier()

    msh_dir = work_dir
    if geo_type == "msh":
        msh_dir = file_dir

    if size == 1:
        msh = gmsh.MeshLoader(mesh_name, msh_dir).execute()
    else:
        if rank == 0:
            command_line = 'gmsh -3 ' + msh_dir + '/' + mesh_name + ' -o ' + \
                work_dir + '/' + partitioned_mesh_name + \
                ' -part ' + str(size) + ' -format msh22'
            print(command_line)
            tmp = shlex.split(command_line)
            fileout = open('gmsh.log', 'w')
            p = subprocess.Popen(tmp, stdin=subprocess.PIPE, stdout=fileout,
                                 stderr=fileout, env=os.environ, shell=False, close_fds=True)
            retcode = p.wait()

        comm.Barrier()

        msh = gmsh.MeshLoader(partitioned_mesh_name,
                              work_dir).execute(myrank=rank)

    return msh, mesh_name, partitioned_mesh_name


def read_mesh_to_fuse(filename, filename_output, file_dir, work_dir, comm, rank, size, index_nodes=0, index_elements=0, index_tags=0):
    msh, mesh_name, partitioned_mesh_name = read_mesh_MPI(
        filename, file_dir, work_dir, comm, rank, size)

    gmshWriter = tbox.GmshExport(msh)
    n_nodes = index_nodes + len(msh.nodes)
    if index_nodes != 0:
        for n in msh.nodes:
            n.no = n.no + index_nodes
    n_elems = index_elements + len(msh.elems)
    if index_elements != 0:
        for e in msh.elems:
            e.no = e.no + index_elements
    n_ptags = index_tags + len(msh.ptags)
    if index_tags != 0:
        for p in msh.ptags:
            if msh.ptags[p].no > 1:
                msh.ptags[p].no = msh.ptags[p].no+index_tags
    gmshWriter.save(filename_output)

    return n_nodes, n_elems, n_ptags


def fuse_meshes(filename1, filename12, filename2, filename22, filename_output, file_dir, work_dir, comm, rank, size):

    if rank == 0:
        n_nodes, n_elems, n_ptags = read_mesh_to_fuse(
            filename1, filename12, file_dir, work_dir, None, 0, 1)
        n_nodes, n_elems, n_ptags = read_mesh_to_fuse(
            filename2, filename22, file_dir, work_dir, None, 0, 1, n_nodes, n_elems, n_ptags)

        command_line = 'gmsh -3 ' + work_dir + '/' + filename12 + '.msh' + ' ' + work_dir + \
            '/' + filename22 + '.msh' + ' -o ' + work_dir + '/' + \
            filename_output + '.msh' + ' -format msh22'
        tmp = shlex.split(command_line)
        fileout = open('gmsh.log', 'w')
        p = subprocess.Popen(tmp, stdin=subprocess.PIPE, stdout=fileout,
                             stderr=fileout, env=os.environ, shell=False, close_fds=True)
        retcode = p.wait()

    if size > 1:
        partitioned_mesh_name = native_str(
            filename_output + '_' + str(size) + '.msh')
        if rank == 0:
            command_line = 'gmsh -3 ' + work_dir + '/' + filename_output + '.msh' + ' -o ' + work_dir + '/' + \
                partitioned_mesh_name + ' -part ' + \
                str(size) + ' -format msh22'
            tmp = shlex.split(command_line)
            fileout = open('gmsh.log', 'w')
            p = subprocess.Popen(tmp, stdin=subprocess.PIPE, stdout=fileout,
                                 stderr=fileout, env=os.environ, shell=False, close_fds=True)
            retcode = p.wait()
        comm.Barrier()
        msh = gmsh.MeshLoader(partitioned_mesh_name,
                              work_dir).execute(myrank=rank)
    else:
        msh = gmsh.MeshLoader(filename_output + '.msh', work_dir).execute()

    return msh


def replace_strings(file_in, file_out, old_strings, new_strings):
    with open(file_in, 'r') as file:
        filedata = file.read()

    for i in range(0, len(old_strings)):
        filedata = filedata.replace(old_strings[i], new_strings[i])

    with open(file_out, 'w') as file:
        file.write(filedata)


def print_nodes(msh, tag_names, rank):

    for tag_name in tag_names:
        print('Rank ' + str(rank) + ' Name tag: ' + tag_name +
              ' Elements: ' + str(len(msh.ntags[tag_name].elems)))
        for el in msh.ntags[tag_name].elems:
            print(el)
        print('Nodes: ')
        print('{0:2s}   {1:3s}'.format('no', 'row'))
        for el in msh.ntags[tag_name].elems:
            for n1 in el.nodes:
                print('{0:2d}   {1:3d}'.format(n1.no, n1.row))


def extract(msh, tag_name, DOFperNode, which_pos, which_dof, ensemble_size, name_sol="x_mm.txt", lot_of_data=False):
    import numpy as np

    nodeslist = []
    #tag_name = "Center of cross section"
    for el in msh.ntags[tag_name].elems:
        for n1 in el.nodes:
            nodeslist.append(n1.no)

    nodes = list(set(nodeslist))

    x = np.empty(np.shape(nodes))
    x_sorted = np.empty(np.shape(nodes))
    nodes_sorted = np.empty(np.shape(nodes), dtype=int)
    sol = np.empty([len(nodes), ensemble_size])
    for i in range(0, len(x)):
        pos = msh.nodes[int(nodes[i])-1].pos
        x[i] = pos.x[which_pos]

    #print x

    indices = np.argsort(x)
    #print x

    for i in range(0, len(x)):
        x_sorted[i] = x[indices[i]]
        nodes_sorted[i] = nodes[indices[i]]
    #print x_sorted
    #print nodes_sorted

    x = x_sorted
    nodes = nodes_sorted

    if lot_of_data:
        sol_all = m.eigenvalues.read_mm(name_sol, is_sparse=False)
        indices = int(DOFperNode)*(nodes-1)+int(which_dof)
        if ensemble_size == 1:
            sol = sol_all[indices]
        else:
            sol = sol_all[indices, :]

    else:
        with open(name_sol) as fp:
            for i, line in enumerate(fp):
                for j in range(0, len(x)):
                    if i == DOFperNode*(nodes[j]-1)+which_dof+2:
                        sol[j, :] = np.array(line.split())

    return x, sol


def extract_2D(msh, tag_name, DOFperNode, which_dof, ensemble_size, name_sol):
    # work only with triangel 2D now
    comm, rank, size = import_MPI()

    from PyTrilinos import Teuchos
    import numpy as np

    nodeslist = []
    #tag_name = "Center of cross section"
    for el in msh.ntags[tag_name].elems:
        for n1 in el.nodes:
            nodeslist.append(n1.no-1)
    nodes = np.array(nodeslist)
    nodes = np.sort(nodes)

    i = 1
    while i < len(nodes):
        if nodes[i] == nodes[i-1]:
            nodes = np.delete(nodes, i)
        else:
            i = i + 1

    nnodes = len(msh.nodes)
    global_to_local = -1*np.ones((nnodes, ), dtype='int32')

    for i in range(0, len(nodes)):
        global_to_local[nodes[i]] = i

    ntri = len(msh.ntags[tag_name].elems)
    tri = np.zeros((ntri, 3), dtype='uint32')

    for i in range(0, ntri):
        tri[i, 0] = global_to_local[msh.ntags[tag_name].elems[i].nodes[0].no-1]
        tri[i, 1] = global_to_local[msh.ntags[tag_name].elems[i].nodes[1].no-1]
        tri[i, 2] = global_to_local[msh.ntags[tag_name].elems[i].nodes[2].no-1]

    x = np.empty(np.shape(nodes))
    y = np.empty(np.shape(nodes))
    z = np.empty(np.shape(nodes))

    sol = np.empty(np.shape(nodes))

    x_sorted = np.empty(np.shape(nodes))
    nodes_sorted = np.empty(np.shape(nodes))

    #sol = np.empty([len(nodes), ensemble_size])

    for i in range(0, len(x)):
        pos = msh.nodes[nodes[i]].pos
        x[i] = pos.x[0]
        y[i] = pos.x[1]
        z[i] = pos.x[2]

    from scipy import io

    #sol = x
    if ensemble_size == 1:
        sol_full = io.mmread(name_sol)
    for j in range(0, len(x)):
        sol[j] = sol_full[DOFperNode*(nodes[j])+which_dof]

    return x, y, z, tri, sol


def extract_2D_all(msh, tag_name, DOFperNode, ensemble_size, name_sol):
    # work only with triangel 2D now
    comm, rank, size = import_MPI()

    from PyTrilinos import Teuchos
    import numpy as np

    nodeslist = []
    for el in msh.ntags[tag_name].elems:
        for n1 in el.nodes:
            nodeslist.append(n1.no-1)
    nodes = np.array(nodeslist)
    nodes = np.sort(nodes)

    ntri = len(msh.ntags[tag_name].elems)
    tri = np.zeros((ntri, 3), dtype='uint32')

    for i in range(0, ntri):
        tri[i, 0] = msh.ntags[tag_name].elems[i].nodes[0].no-1
        tri[i, 1] = msh.ntags[tag_name].elems[i].nodes[1].no-1
        tri[i, 2] = msh.ntags[tag_name].elems[i].nodes[2].no-1

    i = 0
    while i < len(nodes):
        if nodes[i] == nodes[i-1]:
            nodes = np.delete(nodes, i)
        else:
            i = i + 1

    x = np.empty(np.shape(nodes))
    y = np.empty(np.shape(nodes))
    z = np.empty(np.shape(nodes))

    x_sorted = np.empty(np.shape(nodes))
    nodes_sorted = np.empty(np.shape(nodes))

    sol = np.empty([len(nodes), DOFperNode])

    for i in range(0, len(x)):
        pos = msh.nodes[nodes[i]].pos
        x[i] = pos.x[0]
        y[i] = pos.x[1]
        z[i] = pos.x[2]

    from scipy import io

    if ensemble_size == 1:
        sol_full = io.mmread(name_sol)
    for j in range(0, len(x)):
        for k in range(0, DOFperNode):
            sol[j, k] = sol_full[DOFperNode*(nodes[j])+k]

    return x, y, z, tri, sol


def compute_cross_and_theta(vec1, vec2):
    from numpy import linalg as LA

    norm1 = LA.norm(vec1)
    norm2 = LA.norm(vec2)
    cross = np.cross(vec1, vec2)
    dot = np.dot(vec1, vec2)
    theta = np.arccos(old_div(dot, (norm1*norm2)))

    return cross, theta


def rotation_matrix(axis, theta):
    """
    Return the rotation matrix associated with counterclockwise rotation about
    the given axis by theta radians.
    """
    axis = np.asarray(axis)
    axis = old_div(axis, np.sqrt(np.dot(axis, axis)))
    a = np.cos(old_div(theta, 2.0))
    b, c, d = -axis * np.sin(old_div(theta, 2.0))
    aa, bb, cc, dd = a * a, b * b, c * c, d * d
    bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
    return np.array([[aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)],
                     [2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)],
                     [2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc]])


def extract_2D_all_rotated(msh, tag_name, DOFperNode, ensemble_size, name_sol, axis_nodes_id):

    x, y, z, tri, sol = extract_2D_all(
        msh, tag_name, DOFperNode, ensemble_size, name_sol)

    x_r = np.empty(np.shape(x))
    y_r = np.empty(np.shape(x))
    z_r = np.empty(np.shape(x))
    sol_r = np.empty(np.shape(sol))

    pos_OX_1 = msh.nodes[axis_nodes_id[0]].pos
    pos_OX_2 = msh.nodes[axis_nodes_id[1]].pos

    vec_OX = np.array([pos_OX_2.x[0]-pos_OX_1.x[0], pos_OX_2.x[1] -
                       pos_OX_1.x[1], pos_OX_2.x[2]-pos_OX_1.x[2]])

    OX_ref = np.array([1, 0, 0])
    OY_ref = np.array([0, 1, 0])

    cross_prod_OX, theta_X = compute_cross_and_theta(vec_OX, OX_ref)
    rot_X = rotation_matrix(cross_prod_OX, theta_X)

    for i in range(0, len(x)):
        tmp = np.dot(rot_X, [x[i], y[i], z[i]])
        x_r[i] = tmp[0]
        y_r[i] = tmp[1]
        z_r[i] = tmp[2]

        tmp = np.dot(rot_X, [sol[i, 0], sol[i, 1], sol[i, 2]])

        sol_r[i, 0] = tmp[0]
        sol_r[i, 1] = tmp[1]
        sol_r[i, 2] = tmp[2]

    sol_r[:, 3] = sol[:, 3]

    if cross_prod_OX[1] != 0:
        '''
        vec_OY = np.dot(rot_X, vec_OY)
        print vec_OY
        vec_OY[0] = 0
        print vec_OY
        cross_prod_OY, theta_Y = compute_cross_and_theta(vec_OY, OY_ref)
        '''
        cross_prod_OY = np.array([1, 0, 0])
        max_i_z = np.argmax(z_r)
        min_i_z = np.argmin(z_r)

        dz = z_r[max_i_z] - z_r[min_i_z]
        dy = y_r[max_i_z] - y_r[min_i_z]

        theta_Y = np.arctan(old_div(-dz, dy))

        rot_Y = rotation_matrix(cross_prod_OY, theta_Y)

        for i in range(0, len(x)):
            tmp = np.dot(rot_Y, [x_r[i], y_r[i], z_r[i]])
            x_r[i] = tmp[0]
            y_r[i] = tmp[1]
            z_r[i] = tmp[2]

            tmp = np.dot(rot_Y, [sol_r[i, 0], sol_r[i, 1], sol_r[i, 2]])

            sol_r[i, 0] = tmp[0]
            sol_r[i, 1] = tmp[1]
            sol_r[i, 2] = tmp[2]

    return x_r, y_r, z_r, tri, sol_r


def compute_Zernike(x, y, z, tri, dz, n_Zernike, center_pos):
    from numpy import linalg as LA

    n_nodes = len(x)
    n_tri = np.size(tri, 0)

    Zernike_coeff = np.zeros((n_Zernike, ))

    Zernike_functions = np.zeros((n_nodes, n_Zernike))

    rho = np.empty(np.shape(x))
    theta = np.empty(np.shape(x))

    for i in range(0, len(x)):
        rho[i] = LA.norm(np.array([x[i], y[i]])-center_pos)
        if x[i] >= center_pos[0] and y[i] >= center_pos[1]:
            theta[i] = np.arctan(
                old_div((y[i]-center_pos[1]), (x[i]-center_pos[0])))
        elif y[i] >= center_pos[1] or x[i] < center_pos[0]:
            theta[i] = np.arctan(
                old_div((y[i]-center_pos[1]), (x[i]-center_pos[0]))) + np.pi
        else:
            theta[i] = np.arctan(
                old_div((y[i]-center_pos[1]), (x[i]-center_pos[0]))) + 2*np.pi

    max_rho = max(rho)
    rho = old_div(rho, max_rho)

    if n_Zernike >= 1:
        Zernike_functions[:, 0] = 1.
    if n_Zernike >= 2:
        v_max = np.amax(2*rho*np.cos(theta))
        Zernike_functions[:, 1] = 2*rho*np.cos(theta)/v_max
    if n_Zernike >= 3:
        v_max = np.amax(2*rho*np.sin(theta))
        Zernike_functions[:, 2] = 2*rho*np.sin(theta)/v_max
    if n_Zernike >= 4:
        v_max = np.amax(np.sqrt(3)*(2*np.power(rho, 2)-1))
        Zernike_functions[:, 3] = np.sqrt(3)*(2*np.power(rho, 2)-1)/v_max
    if n_Zernike >= 5:
        Zernike_functions[:, 4] = np.sqrt(6)*np.power(rho, 2)*np.sin(2*theta)
    if n_Zernike >= 6:
        Zernike_functions[:, 5] = np.sqrt(6)*np.power(rho, 2)*np.cos(2*theta)
    if n_Zernike >= 7:
        Zernike_functions[:, 6] = np.sqrt(
            8)*(3*np.power(rho, 3)-2*rho)*np.sin(theta)
    if n_Zernike >= 8:
        Zernike_functions[:, 7] = np.sqrt(
            8)*(3*np.power(rho, 3)-2*rho)*np.cos(theta)
    if n_Zernike >= 9:
        Zernike_functions[:, 8] = np.sqrt(8)*np.power(rho, 3)*np.sin(3*theta)
    if n_Zernike >= 10:
        Zernike_functions[:, 9] = np.sqrt(8)*np.power(rho, 3)*np.cos(3*theta)

    W = np.zeros((n_nodes, n_nodes))

    for i in range(0, n_tri):
        cross = np.cross(np.array([x[tri[i, 1]] - x[tri[i, 0]], y[tri[i, 1]] - y[tri[i, 0]], z[tri[i, 1]] - z[tri[i, 0]]]),
                         np.array([x[tri[i, 2]] - x[tri[i, 0]], y[tri[i, 2]] - y[tri[i, 0]], z[tri[i, 2]] - z[tri[i, 0]]]))
        area = old_div(LA.norm(cross), 6)  # area per node of the triangle

        for j in range(0, 3):
            W[tri[i, j], tri[i, j]] = W[tri[i, j], tri[i, j]] + area

    TMP_Z = np.dot(np.transpose(Zernike_functions), W)
    rhs_Z = np.dot(TMP_Z, dz)

    A_Z = np.dot(TMP_Z, Zernike_functions)

    Zernike_coeff = np.linalg.solve(A_Z, rhs_Z)

    return Zernike_coeff, Zernike_functions, theta, rho


def compute_RBM_Curvature_Irreg(x, y, z, tri, dz, unit, center_pos):
    from numpy import linalg as LA

    n_nodes = len(x)
    n_tri = np.size(tri, 0)

    RBM_Curvature_Irreg_coeff = np.zeros((4, ))

    RBM_Curvature_Irreg_functions = np.zeros((n_nodes, 4))

    if unit == "m":
        x = x*1000
        y = y*1000
        z = z*1000
        dz = dz*1000
        center_pos = center_pos*1000

    Zernike_coeff, Zernike_functions, theta, rho = compute_Zernike(
        x, y, z, tri, dz, 4, center_pos)
    dz_residual = dz - Zernike_coeff[0]*Zernike_functions[:, 0] - Zernike_coeff[1]*Zernike_functions[:,
                                                                                                     1] - Zernike_coeff[2]*Zernike_functions[:, 2] - Zernike_coeff[3]*Zernike_functions[:, 3]

    optical_coefficients = np.zeros((5, 1)).flatten()
    # piston  [mm]
    optical_coefficients[0] = Zernike_coeff[0]
    index = np.argmax(Zernike_functions[:, 1])
    optical_coefficients[1] = np.arctan(-Zernike_coeff[1]*(old_div(
        Zernike_functions[index, 1], (x[index]-center_pos[0]))))     # rot y   [rad]
    index = np.argmax(Zernike_functions[:, 2])
    optical_coefficients[2] = np.arctan(-Zernike_coeff[2]*(old_div(
        Zernike_functions[index, 2], (y[index]-center_pos[1]))))     # rot x   [rad]
    optical_coefficients[3] = Zernike_coeff[3] * \
        (np.amax(Zernike_functions[:, 3])-np.amin(Zernike_functions[:, 3]))
    optical_coefficients[4] = np.amax(dz_residual) - np.amin(dz_residual)

    return optical_coefficients, dz_residual


def compute_RBM_Curvature_Irreg_Sequential(x, y, z, tri, dz, unit, center_pos):
    from numpy import linalg as LA

    n_nodes = len(x)
    n_tri = np.size(tri, 0)

    RBM_Curvature_Irreg_coeff = np.zeros((4, ))

    RBM_Curvature_Irreg_functions = np.zeros((n_nodes, 4))

    if unit == "m":
        x = x*1000
        y = y*1000
        z = z*1000
        dz = dz*1000
        center_pos = center_pos*1000

    rho = np.empty(np.shape(x))

    for i in range(0, len(x)):
        rho[i] = LA.norm(np.array([x[i], y[i]])-center_pos)

    RBM_Curvature_Irreg_functions[:, 0] = 1.
    RBM_Curvature_Irreg_functions[:, 1] = y-center_pos[1]
    RBM_Curvature_Irreg_functions[:, 2] = x-center_pos[0]
    RBM_Curvature_Irreg_functions[:, 3] = np.power(rho, 2)

    W = np.zeros((n_nodes, n_nodes))

    print("mean dz")
    print(np.mean(dz))

    for i in range(0, n_tri):
        cross = np.cross(np.array([x[tri[i, 1]] - x[tri[i, 0]], y[tri[i, 1]] - y[tri[i, 0]], z[tri[i, 1]] - z[tri[i, 0]]]),
                         np.array([x[tri[i, 2]] - x[tri[i, 0]], y[tri[i, 2]] - y[tri[i, 0]], z[tri[i, 2]] - z[tri[i, 0]]]))
        area = old_div(LA.norm(cross), 6)  # area per node of the triangle

        for j in range(0, 3):
            W[tri[i, j], tri[i, j]] = W[tri[i, j], tri[i, j]] + area

    for i in range(0, 4):
        TMP_Z = np.dot(np.transpose(RBM_Curvature_Irreg_functions[:, i]), W)
        rhs_Z = np.dot(TMP_Z, dz)

        A_Z = np.dot(TMP_Z, RBM_Curvature_Irreg_functions[:, i])

        RBM_Curvature_Irreg_coeff[i] = old_div(rhs_Z, A_Z)
        dz = dz - RBM_Curvature_Irreg_coeff[i] * \
            RBM_Curvature_Irreg_functions[:, i]

    optical_coefficients = np.zeros((5, ))
    optical_coefficients[0] = RBM_Curvature_Irreg_coeff[0]  # piston
    optical_coefficients[1] = np.arctan(RBM_Curvature_Irreg_coeff[1])
    optical_coefficients[2] = np.arctan(RBM_Curvature_Irreg_coeff[2])
    optical_coefficients[3] = RBM_Curvature_Irreg_coeff[3] * \
        np.amax(RBM_Curvature_Irreg_functions[:, 3])
    optical_coefficients[4] = np.amax(dz) - np.amin(dz)

    print("optical_coefficients")
    print(optical_coefficients)

    return RBM_Curvature_Irreg_coeff, RBM_Curvature_Irreg_functions, dz


def check_optical_tolerance(optical_tolerance, unit):
    passed = 1
    max_angle = 0.00175
    if unit == "mm":
        max_fringe = 0.0316
        max_translation = 2
    elif unit == "m":
        max_fringe = 0.0000316
        max_translation = 0.002

    if np.absolute(optical_tolerance[0]) >= max_translation:
        print("translation")
        passed = 0

    if np.absolute(optical_tolerance[1]) >= max_angle:
        print("rotation x")
        passed = 0

    if np.absolute(optical_tolerance[2]) >= max_angle:
        print("rotation y")
        passed = 0

    if optical_tolerance[3] >= max_fringe:
        print("curvature")
        passed = 0

    if optical_tolerance[4] >= max_fringe:
        print("irregularity")
        passed = 0

    return passed


def check_optical_tolerance_Zernike(Zernike_coeff, Zernike_functions, dz, unit):

    passed = 1
    max_angle = 0.00175
    if unit == "mm":
        max_fringe = 0.0316
        max_translation = 2
    elif unit == "m":
        max_fringe = 0.0000316
        max_translation = 0.002

    if np.absolute(Zernike_coeff[0]) >= max_translation:
        print("translation")
        passed = 0

    if np.absolute(Zernike_coeff[1]) >= max_angle:
        print("rotation x")
        passed = 0

    if np.absolute(Zernike_coeff[2]) >= max_angle:
        print("rotation y")
        passed = 0

    tmp = np.dot(Zernike_coeff[3], Zernike_functions[:, 3])

    if np.absolute(max(tmp)-min(tmp)) >= max_fringe:
        print("curvature")
        passed = 0

    tmp = dz - np.dot(Zernike_coeff[0], Zernike_functions[:, 0]) - np.dot(Zernike_coeff[1], Zernike_functions[:, 1]) - \
        np.dot(Zernike_coeff[2], Zernike_functions[:, 2]) - \
        np.dot(Zernike_coeff[3], Zernike_functions[:, 3])

    if np.absolute(max(tmp)-min(tmp)) >= max_fringe:
        print(max(tmp))
        print(min(tmp))
        print("irregularity")
        passed = 0

    return passed
