/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wElement.h"
#include "wNode.h"
#include "wTag.h"
#include "wFct0.h"
#include "wMem.h"
#include "wCache.h"
#include "std_extra.h"

using namespace tbox;

namespace tbox
{

TBOX_API std::ostream &
operator<<(std::ostream &out, ELTYPE const &obj)
{
    switch (obj)
    {
    case ELTYPE::LINE2:
        out << "LINE2";
        break;
    case ELTYPE::TRI3:
        out << "TRI3";
        break;
    case ELTYPE::QUAD4:
        out << "QUAD4";
        break;
    case ELTYPE::TETRA4:
        out << "TETRA4";
        break;
    case ELTYPE::HEX8:
        out << "HEX8";
        break;
    case ELTYPE::POINT1:
        out << "POINT1";
        break;
    default:
        out << "UNKNOWN-" << static_cast<int>(obj);
        break;
    }
    return out;
}

TBOX_API std::ostream &
operator<<(std::ostream &out, MATTYPE const &obj)
{
    switch (obj)
    {
    case MATTYPE::MAT_K:
        out << "MAT_K";
        break;
    case MATTYPE::MAT_M:
        out << "MAT_M";
        break;
    case MATTYPE::MAT_S:
        out << "MAT_S";
        break;
    default:
        out << "UNKNOWN-" << static_cast<int>(obj);
        break;
    }
    return out;
}

} // namespace tbox

// ------------------------------------------------------------------------------------

Element::Element(int n, Tag *_ptag, Tag *_etag,
                 std::vector<Tag *> &_parts, std::vector<Node *> &nods)
    : wObject(), no(n), ptag(_ptag), etag(_etag), parts(_parts), nodes(nods)
{
    if (ptag)
        ptag->elems.push_back(this);
    if (etag)
        etag->elems.push_back(this);

    for (auto p : parts)
        p->elems.push_back(this);
}

void Element::write(std::ostream &out) const
{
    out << type() << " #" << no << ":";

    out << " nodes=[ ";
    for (auto n : nodes)
        out << "#" << n->no << ' ';
    out << "]";

    if (ptag)
        out << " ptag #" << ptag->no;
    if (etag)
        out << " etag #" << etag->no;
    if (!parts.empty())
    {
        out << " parts=[ ";
        for (auto p : parts)
            out << "#" << p->no << ' ';
        out << "]";
    }
}

/**
 * @brief Test fct - calculate K and print it
 */
void Element::buildK()
{
    getVMem().update(true);
    Eigen::MatrixXd K(nodes.size(), nodes.size());
    std::vector<double> u;
    buildK(K, u, Fct0C(1.0));
    std::cout << "K=" << K << '\n';
}

/**
 * @brief Test fct - calculate M and print it
 */
void Element::buildM()
{
    getVMem().update(true);
    Eigen::MatrixXd M(nodes.size(), nodes.size());
    buildM(M);
    std::cout << "M=" << M << '\n';
}

/**
 * @brief Test fct - calculate S and print it
 */
void Element::buildS()
{
    getVMem().update(true);
    Eigen::MatrixXd S;
    buildS(S);
    std::cout << "S=" << S << '\n';
}

/**
 * @brief Test fct - calculate s and print it
 */
void Element::builds()
{
    getVMem().update(true);
    Eigen::VectorXd s;
    std::vector<double> u;
    builds(s, u, Fct0C(1.0));
    std::cout << "s=" << s << '\n';
}

/**
 * @brief Return the values extrapolated at all the element nodes
 * based on the values at Gauss points
 * @authors Kim Liegeois
 */

void Element::gp2node(Eigen::MatrixXd &values_at_gp, Eigen::MatrixXd &values_at_node)
{
    Cache &cache = getVCache();
    Gauss &gauss = cache.getVGauss();

    values_at_node = Eigen::MatrixXd::Zero(nodes.size(), 18); // todo 18 comes from Kim, is it general?
    Eigen::MatrixXd A(gauss.p.size(), nodes.size());
    for (size_t i = 0; i < gauss.p.size(); ++i)
        for (size_t j = 0; j < nodes.size(); ++j)
            A(i, j) = cache.ff[i][j];

    values_at_node = (A.transpose() * A).inverse() * A.transpose() * values_at_gp;
}

double Element::buildJ(size_t k, Eigen::MatrixXd &J) const
{
    throw std::runtime_error("Element::buildJ not implemented\n");
}
double Element::buildJ(size_t k) const
{
    throw std::runtime_error("Element::buildJ not implemented\n");
}
void Element::buildK(Eigen::MatrixXd &K, std::vector<double> const &u, Fct0 const &fct)
{
    throw std::runtime_error("Element::buildK not implemented\n");
}
void Element::buildK(Eigen::MatrixXd &K, std::vector<double> const &u, Fct2 const &fct, bool fake)
{
    throw std::runtime_error("Element::buildK not implemented\n");
}
void Element::buildDK(Eigen::MatrixXd &K, std::vector<double> const &u, Fct0 const &fct)
{
    throw std::runtime_error("Element::buildNK not implemented\n");
}
//---------------------
void Element::buildK_linElastic(Eigen::MatrixXd &K, Eigen::MatrixXd const &H)
{
    throw std::runtime_error("Element::buildK_linElastic not implemented\n");
}

void Element::buildK_mechanic(Eigen::MatrixXd &K_e, Eigen::MatrixXd const &H)
{
    throw std::runtime_error("Element::buildK_mechanic not implemented\n");
}

void Element::buildS_thermomechanic(Eigen::MatrixXd &S_e, Eigen::MatrixXd const &D)
{
    throw std::runtime_error("Element::buildS_thermomechanic not implemented\n");
}

void Element::buildL_thermic(Eigen::MatrixXd &L_e, Eigen::MatrixXd const &K_conductivity)
{
    throw std::runtime_error("Element::buildL_thermic not implemented\n");
}

void Element::strain_stress(Eigen::MatrixXd &Strain, Eigen::MatrixXd &Stress, Eigen::MatrixXd const &H, std::vector<double> const &u)
{
    throw std::runtime_error("Element::strain_stress not implemented\n");
}

void Element::iso_coordinates(Pt x, double eta, double dzeta)
{
    throw std::runtime_error("Element::iso_coordinates not implemented\n");
}

Pt Element::normal()
{
    throw std::runtime_error("Element::normal not implemented\n");
}

//---------------------
void Element::buildM(Eigen::MatrixXd &M)
{
    throw std::runtime_error("Element::buildM not implemented\n");
}
void Element::buildS(Eigen::MatrixXd &S)
{
    throw std::runtime_error("Element::buildS not implemented\n");
}
void Element::builds(Eigen::VectorXd &s, std::vector<double> const &u, Fct0 const &f)
{
    throw std::runtime_error("Element::builds not implemented\n");
}
void Element::builds(Eigen::VectorXd &s, std::vector<double> const &u, Fct1 const &f)
{
    throw std::runtime_error("Element::builds not implemented\n");
}
void Element::buildDs(Eigen::VectorXd &s, std::vector<double> const &u, Fct0 const &f)
{
    throw std::runtime_error("Element::buildNs not implemented\n");
}
void Element::build(MATTYPE type, Eigen::MatrixXd &A)
{
    throw std::runtime_error("Element::build not implemented\n");
}
void Element::computeGrad(std::vector<double> const &u, Eigen::VectorXd &grad, size_t k)
{
    throw std::runtime_error("Element::computeGrad not implemented\n");
}
double Element::computeGrad2(std::vector<double> const &u, size_t k)
{
    throw std::runtime_error("Element::computeGrad2 not implemented\n");
}
void Element::computeqV(std::vector<double> const &u, Eigen::VectorXd &qV, Fct0 const &fct)
{
    throw std::runtime_error("Element::computeqV not implemented\n");
}
void Element::computeqV(std::vector<double> const &u, Eigen::VectorXd &qV, Fct2 const &fct)
{
    throw std::runtime_error("Element::computeqV not implemented\n");
}
void Element::computeqint(std::vector<double> const &u, Eigen::VectorXd &qV, Fct2 const &fct)
{
    throw std::runtime_error("Element::computeqint not implemented\n");
}
double Element::computeV(std::vector<double> const &u, Fct0 const &fct)
{
    throw std::runtime_error("Element::computeV not implemented\n");
}
void Element::computeV(std::vector<double> const &u, Eigen::MatrixXd &out, Fct2 const &fct)
{
    throw std::runtime_error("Element::computeV not implemented\n");
}

double Element::computeV()
{
    double vol = computeV(std::vector<double>(), Fct0C(1.0));
    return vol;
}

Mem &Element::getVMem()
{
    throw std::runtime_error("Element::getVMem not implemented\n");
}
