/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wFct2.h"
#include "wFct0.h"
#include "wPt.h"
#include "wElement.h"
#include "wNode.h"
#include "wCache.h"
#include <Eigen/Dense>

using namespace tbox;

Fct2C::Fct2C(double _v11, double _v22, double _v12) : Fct2()
{
    v(0, 0) = _v11;
    v(1, 1) = _v22;
    v(1, 0) = _v12;
    v(0, 1) = _v12;

    double det = v.determinant();
    if (det <= 0.0)
        throw std::runtime_error("det(Fct2C)<0!");
}

void Fct2C::eval(Element &e,
                 std::vector<double> const &u,
                 size_t k, Eigen::MatrixXd &out, bool fake) const
{
    out = v; // Eigen resizes automatically
}

void Fct2C::write(std::ostream &out) const
{
    out << "Fct2C=" << v;
}

//-------------------------------------------------------------------------

void Fct2XYZ::eval(Element &e,
                   std::vector<double> const &u,
                   size_t k, Eigen::MatrixXd &out, bool fake) const
{
    Cache &cache = e.getVCache();
    std::vector<std::vector<double>> const &ff = cache.ff;

    // position pg en x,y
    Pt xpg;
    for (size_t i = 0; i < e.nodes.size(); ++i)
        xpg += ff[k][i] * e.nodes[i]->pos;

    this->eval(xpg, out, fake);
}

void Fct2XYZ::write(std::ostream &out) const
{
    out << "Fct2XYZ";
}

//-------------------------------------------------------------------------

void Fct2U::eval(Element &e,
                 std::vector<double> const &u,
                 size_t k, Eigen::MatrixXd &out, bool fake) const
{
    Cache &cache = e.getVCache();
    std::vector<std::vector<double>> const &ff = cache.ff;

    // u au pg
    double upg = 0.0;
    for (size_t i = 0; i < e.nodes.size(); ++i)
        upg += ff[k][i] * u[e.nodes[i]->row];

    this->eval(upg, out, fake);
}

void Fct2U::write(std::ostream &out) const
{
    out << "Fct2U";
}

//-------------------------------------------------------------------------

void Fct2UdU::eval(Element &e,
                   std::vector<double> const &u,
                   size_t k, Eigen::MatrixXd &out, bool fake) const
{
    Cache &cache = e.getVCache();
    std::vector<std::vector<double>> const &ff = cache.ff;

    // u au pg => upg
    double upg = 0.0;
    for (size_t i = 0; i < e.nodes.size(); ++i)
        upg += ff[k][i] * u[e.nodes[i]->row];

    //gradU au pg => gradupg
    Eigen::VectorXd gradupg;
    e.computeGrad(u, gradupg, k);

    this->eval(e, k, upg, gradupg, out, fake);
}

void Fct2UdU::write(std::ostream &out) const
{
    out << "Fct2UdU";
}

//-------------------------------------------------------------------------

void Fct2PwLf::eval(double u, Eigen::MatrixXd &v, bool fake) const
{
    v.resize(2, 2);
    v(0, 0) = v11.eval(u);
    v(1, 1) = v22.eval(u);
    v(1, 0) = v(0, 1) = v12.eval(u);
}

void Fct2PwLf::write(std::ostream &out) const
{
    out << "Fct2PwLf";
}
