/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wHex8.h"
#include "wSfHex8.h"
#include "wGaussHex8.h"
#include "wCacheHex8.h"
#include "wMemHex8.h"
#include "wNode.h"
#include "wLazy.h"
#include "wFct0.h"

using namespace tbox;

Hex8::Hex8(int n, Tag *_ptag, Tag *_etag,
           std::vector<Tag *> &_parts, std::vector<Node *> &nods) : Element(n, _ptag, _etag, _parts, nods)
{
    pmem = new MemHex8(*this, static_cast<int>(getCache().gauss.p.size()));
}
Hex8::~Hex8()
{
    if (pmem)
        delete pmem;
}

/*
 * @brief Private memory handling Jacobian for Hex8 Element 
 * @authors Romain Boman, Adrien Crovato
 */
Mem &Hex8::getVMem()
{
    return getMem();
}
MemHex8 &Hex8::getMem()
{
    return *pmem;
}

/*
 * @brief Private cache handling shape functions and Gauss points for Hex8 
 * @authors Romain Boman, Adrien Crovato
 */
CacheHex8 &Hex8::getCache()
{
    static Lazy<CacheHex8> cache(2);
    return cache.get();
}
Cache &Hex8::getVCache()
{
    return getCache();
}

/**
 * @brief Compute Jacobian Matrix at Gauss point k
 * @authors Romain Boman
 */
double Hex8::buildJ(size_t k, Eigen::MatrixXd &J) const
{
    CacheHex8 &cache = getCache();

    J = Eigen::Matrix3d::Zero();
    size_t i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
    {
        std::vector<double> &dff = cache.dff[k][i];
        J += Eigen::Vector3d(dff.data()) * Eigen::RowVector3d((*it)->pos.x.data());
    }
    return J.determinant();
}

void Hex8::buildK(Eigen::MatrixXd &K, std::vector<double> const &u, Fct0 const &fct)
{
    CacheHex8 &cache = getCache();
    GaussHex8 &gauss = cache.gauss;
    MemHex8 &mem = getMem();

    K = Eigen::MatrixXd::Zero(8, 8);
    Eigen::Vector3d factor1;
    Eigen::Vector3d factor2;
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // function and Jacobian
        double fk = fct.eval(*this, u, k);
        double detJ = mem.getDetJ(k);
        Eigen::Matrix3d const &J = mem.getJinv(k);

        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[k][i];
            factor1 = J * Eigen::Vector3d(dffi.data());

            for (size_t j = 0; j < nodes.size(); ++j)
            {
                std::vector<double> &dffj = cache.dff[k][j];
                factor2 = J * Eigen::Vector3d(dffj.data());
                K(i, j) += factor1.dot(factor2) * gauss.w[k] * detJ * fk;
            }
        }
    }
}

void Hex8::buildK_mechanic(Eigen::MatrixXd &K_e, Eigen::MatrixXd const &H)
{
    CacheHex8 &cache = getCache();
    GaussHex8 &gauss = cache.gauss;
    MemHex8 &mem = getMem();

    K_e = Eigen::MatrixXd::Zero(24, 24);
    Eigen::Vector3d factor1;
    Eigen::Vector3d factor2;
    Eigen::Vector3d factor3;
    Eigen::Matrix3d TMP1;
    for (size_t a = 0; a < gauss.p.size(); ++a)
    {
        // calcul de la matrice jacobienne (au point de Gauss "ksi")
        double detJ = mem.getDetJ(a);
        Eigen::Matrix3d const &J = mem.getJinv(a);

        // calcul de K (au point de Gauss "ksi")
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[a][i];
            factor1 = J * Eigen::Vector3d(dffi.data());

            for (size_t j = i; j < nodes.size(); ++j)
            {
                std::vector<double> &dffj = cache.dff[a][j];
                factor2 = J * Eigen::Vector3d(dffj.data());

                for (size_t k = 0; k < 3; ++k)
                    for (size_t m = 0; m < 3; ++m)
                    {
                        for (size_t l = 0; l < 3; ++l)
                            for (size_t n = 0; n < 3; ++n)
                                TMP1(l, n) = H(k * 3 + l, m * 3 + n);
                        factor3 = TMP1 * factor2;
                        K_e(i * 3 + k, j * 3 + m) += factor1.dot(factor3) * gauss.w[a] * detJ;
                    }
            }
        }
    }
    for (size_t i = 0; i < 3 * nodes.size(); ++i)
        for (size_t j = i; j < 3 * nodes.size(); ++j)
            K_e(j, i) = K_e(i, j);
}

void Hex8::buildS_thermomechanic(Eigen::MatrixXd &S_e, Eigen::MatrixXd const &D)
{
    CacheHex8 &cache = getCache();
    GaussHex8 &gauss = cache.gauss;
    MemHex8 &mem = getMem();

    S_e = Eigen::MatrixXd::Zero(24, 8);
    Eigen::Vector3d factor1;
    Eigen::Vector3d factor2;
    for (size_t a = 0; a < gauss.p.size(); ++a)
    {
        // calcul de la matrice jacobienne (au point de Gauss "ksi")
        double detJ = mem.getDetJ(a);
        Eigen::Matrix3d const &J = mem.getJinv(a);
        std::vector<double> &ff = cache.ff[a];

        // calcul de K (au point de Gauss "ksi")
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[a][i];
            factor1 = J * Eigen::Vector3d(dffi.data());
            factor2 = D * factor1;
            for (size_t j = 0; j < nodes.size(); ++j)
                for (size_t k = 0; k < 3; ++k)
                    S_e(i * 3 + k, j) += factor2[k] * ff[j] * gauss.w[a] * detJ;
        }
    }
}

void Hex8::buildL_thermic(Eigen::MatrixXd &L_e, Eigen::MatrixXd const &K_conductivity)
{
    CacheHex8 &cache = getCache();
    GaussHex8 &gauss = cache.gauss;
    MemHex8 &mem = getMem();

    L_e = Eigen::MatrixXd::Zero(8, 8);
    Eigen::Vector3d factor1;
    Eigen::Vector3d factor2;
    Eigen::Vector3d factor3;
    for (size_t a = 0; a < gauss.p.size(); ++a)
    {
        // calcul de la matrice jacobienne (au point de Gauss "ksi")
        double detJ = mem.getDetJ(a);
        Eigen::Matrix3d const &J = mem.getJinv(a);

        // calcul de K (au point de Gauss "ksi")
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[a][i];
            factor1 = J * Eigen::Vector3d(dffi.data());
            factor2 = K_conductivity * factor1;
            for (size_t j = i; j < nodes.size(); ++j)
            {
                std::vector<double> &dffj = cache.dff[a][j];
                factor3 = J * Eigen::Vector3d(dffj.data());
                L_e(i, j) += factor2.dot(factor3) * gauss.w[a] * detJ;
            }
        }
    }
    for (size_t i = 0; i < nodes.size(); ++i)
        for (size_t j = i; j < nodes.size(); ++j)
            L_e(j, i) = L_e(i, j);
}

void Hex8::strain_stress(Eigen::MatrixXd &Strain, Eigen::MatrixXd &Stress, Eigen::MatrixXd const &H, std::vector<double> const &u)
{
    CacheHex8 &cache = getCache();
    GaussHex8 &gauss = cache.gauss;
    MemHex8 &mem = getMem();

    Strain = Stress = Eigen::Matrix3d::Zero();
    Eigen::Matrix3d Tmp = Eigen::Matrix3d::Zero();
    Eigen::MatrixXd u_e(3, nodes.size());
    for (size_t i = 0; i < nodes.size(); ++i)
    {
        u_e(0, i) = u[3 * (nodes[i]->row)];
        u_e(1, i) = u[3 * (nodes[i]->row) + 1];
        u_e(2, i) = u[3 * (nodes[i]->row) + 2];
    }
    for (size_t a = 0; a < gauss.p.size(); ++a)
    {
        // calcul de la matrice jacobienne (au point de Gauss "ksi")
        Eigen::Matrix3d const &J = mem.getJinv(a);
        Eigen::Matrix3d du_d_Phi = Eigen::Matrix3d::Zero();
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[a][i];
            for (auto k = 0; k < 3; ++k)
                for (auto j = 0; j < 3; ++j)
                    du_d_Phi(j, k) += dffi[j] * u_e(k, i);
        }
        for (size_t k = 0; k < 3; ++k)
            for (size_t m = 0; m < 3; ++m)
                for (size_t i = 0; i < 3; ++i)
                    Tmp(k, m) += J(k, i) * du_d_Phi(i, m) / (nodes.size());
    }
    for (size_t i = 0; i < 3; ++i)
        for (size_t j = 0; j < 3; ++j)
            Strain(i, j) = (Tmp(i, j) + Tmp(j, i)) / 2;

    for (size_t i = 0; i < 3; ++i)
        for (size_t j = 0; j < 3; ++j)
            for (size_t k = 0; k < 3; ++k)
                for (size_t l = 0; l < 3; ++l)
                    Stress(i, j) += H(i * 3 + j, k * 3 + l) * Strain(k, l);
}

void Hex8::buildM(Eigen::MatrixXd &M)
{
    CacheHex8 &cache = getCache();
    GaussHex8 &gauss = cache.gauss;
    MemHex8 &mem = getMem();

    M = Eigen::MatrixXd::Zero(8, 8);
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // calcul de la matrice jacobienne (au point de Gauss "ksi")
        double detJ = mem.getDetJ(k);
        std::vector<double> &ff = cache.ff[k];
        M += Eigen::Matrix<double, 8, 1>(ff.data()) * Eigen::Matrix<double, 1, 8>(ff.data()) * gauss.w[k] * detJ; // M = M + {ff}^T [ff].wG.detJ
    }
}

void Hex8::build(MATTYPE type, Eigen::MatrixXd &A, std::vector<double> const &u, Fct0 const &fct)
{
    switch (type)
    {
    case MATTYPE::MAT_K:
        buildK(A, u, fct);
        break;
    case MATTYPE::MAT_M:
        buildM(A);
        break;
    default:
        std::stringstream out;
        out << "Hex8 cannot build " << type;
        throw std::runtime_error(out.str());
    }
}

/**
 * @brief Compute intergal of function
 * I = int fct dV
 * @authors Adrien Crovato
 */
double Hex8::computeV(std::vector<double> const &u, Fct0 const &fct)
{
    CacheHex8 &cache = getCache();
    GaussHex8 &gauss = cache.gauss;
    MemHex8 &mem = getMem();

    double V = 0.0;
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // function evaluation
        double fk = fct.eval(*this, u, k);
        double detJ = mem.getDetJ(k);
        V += fk * gauss.w[k] * detJ;
    }
    return V;
}

/**
 * @brief Return the element volume
 * @authors Adrien Crovato
 */
double Hex8::computeV()
{
    MemHex8 &mem = getMem();
    return mem.getVol();
}
