/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wLine2.h"
#include "wSfLine2.h"
#include "wGaussLine2.h"
#include "wCacheLine2.h"
#include "wMemLine2.h"
#include "wNode.h"
#include "wMath.h"
#include "wLazy.h"
#include "wFct0.h"
#include "wFct1.h"

using namespace tbox;

Line2::Line2(int n, Tag *_ptag, Tag *_etag,
             std::vector<Tag *> &_parts, std::vector<Node *> &nods) : Element(n, _ptag, _etag, _parts, nods)
{
    pmem = new MemLine2(*this, static_cast<int>(getCache().gauss.p.size()));
}
Line2::~Line2()
{
    if (pmem)
        delete pmem;
}

/*
 * @brief Private memory handling Jacobian for Line2 Element 
 * @authors Romain Boman, Adrien Crovato
 */
Mem &Line2::getVMem()
{
    return getMem();
}
MemLine2 &Line2::getMem()
{
    return *pmem;
}

/*
 * @brief Private cache handling shape functions and Gauss points for Line2 
 * @authors Romain Boman, Adrien Crovato
 */
CacheLine2 &Line2::getCache()
{
    static Lazy<CacheLine2> cache(2);
    return cache.get();
}
Cache &Line2::getVCache()
{
    return getCache();
}

Pt Line2::normal()
{
    Pt dx = (nodes[1]->pos - nodes[0]->pos).normalized();
    return Pt(-dx.x[1], dx.x[0], 0.);
}

/**
 * @brief Compute Jacombian determinant at Gauss point k
 * @authors Romain Boman
 */
double Line2::buildJ(size_t k) const
{
    CacheLine2 &cache = getCache();

    Eigen::Vector3d t0 = Eigen::Vector3d::Zero();
    size_t i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
    {
        std::vector<double> &dff = cache.dff[k][i];
        t0 += Eigen::Vector3d((*it)->pos.x.data()) * dff[0];
    }
    return t0.norm();
}

/**
 * @brief Compute flux vector
 * 
 * s_i(2) = int f * N_i dV 
 * @authors Romain Boman
 */
void Line2::builds(Eigen::VectorXd &s, std::vector<double> const &u, Fct0 const &f)
{
    CacheLine2 &cache = getCache();
    GaussLine2 &gauss = cache.gauss;
    MemLine2 &mem = getMem();

    s = Eigen::Vector2d::Zero();
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        double sk = f.eval(*this, u, k);
        double detJ = mem.getDetJ(k);

        s += Eigen::Vector2d(cache.ff[k].data()) * sk * gauss.w[k] * detJ; // s = s + sk.wG.detJ
    }
}

/**
 * @brief Compute flux vector
 * 
 * s_i(2) = int f_j*n_j * N_i dV 
 * @authors Adrien Crovato
 */
void Line2::builds(Eigen::VectorXd &s, std::vector<double> const &u, Fct1 const &f)
{
    CacheLine2 &cache = getCache();
    GaussLine2 &gauss = cache.gauss;
    MemLine2 &mem = getMem();

    s = Eigen::Vector2d::Zero();
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        double sk = f.eval(*this, u, k) * normal();
        double detJ = mem.getDetJ(k);

        s += Eigen::Vector2d(cache.ff[k].data()) * sk * gauss.w[k] * detJ; // s = s + sk.wG.detJ
    }
}

/**
 * @brief Compute square of gradient along element
 * @authors Adrien Crovato
 */
double Line2::computeGrad2(std::vector<double> const &u, size_t k)
{
    double deltaU = u[nodes[1]->row] - u[nodes[0]->row];
    double deltaS = computeV();
    return deltaU / deltaS * deltaU / deltaS;
}

/**
 * @brief Compute intergal of function
 * 
 * I = int fct dV
 * @authors Adrien Crovato
 */
double Line2::computeV(std::vector<double> const &u, Fct0 const &fct)
{
    CacheLine2 &cache = getCache();
    GaussLine2 &gauss = cache.gauss;
    MemLine2 &mem = getMem();

    double V = 0.0;
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // evaluation de la fct a intégrer
        double fk = fct.eval(*this, u, k);
        double detJ = mem.getDetJ(k);
        V += fk * gauss.w[k] * detJ;
    }
    return V;
}

/**
 * @brief Return the element volume
 * @authors Adrien Crovato
 */
double Line2::computeV()
{
    MemLine2 &mem = getMem();
    return mem.getVol();
}
