/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wLinesearch.h"
#include <sstream>
#include <iomanip>
#include <cmath>
#include <stdexcept>

using namespace tbox;

LSFunction::LSFunction() : wObject()
{
}

double LSFunction::eval(double alpha)
{
    return 0.0;
}

RevFunction::RevFunction(LSFunction &_f) : LSFunction(), f(_f)
{
}

double RevFunction::eval(double alpha)
{
    return f.eval(-alpha);
}

// ---

Linesearch::Linesearch(LSFunction &_fct) : fct(_fct)
{
    verbose = false;
    maxLsIt = 1000;
    tol = 1e-10;
    fevalIt = 0;
}

/**
 * @brief Set control parameters
 * @authors Adrien Crovato
 */
void Linesearch::set(int _maxLSIt, double _lSTol, bool _verbose)
{
    verbose = _verbose;
    maxLsIt = _maxLSIt;
    tol = _lSTol;
}

double Linesearch::run()
{
    throw std::runtime_error("Linesearch::run() not implemented!\n");
}

// ---

BacktrackLS::BacktrackLS(LSFunction &_fct) : Linesearch(_fct)
{
}

/**
 * @brief Backtracking linesearch
 * @authors Adrien Crovato
 */
double BacktrackLS::run()
{
    double a = 1;
    double phi0 = fct.eval(0.);
    fevalIt = 1;
    do
    {
        double phi = fct.eval(a);
        fevalIt++;
        if (phi > phi0)
            a /= 2;
        else
            break;
    } while (fevalIt < maxLsIt);
    return a;
}

// ---

FleuryLS::FleuryLS(LSFunction &_fct, double _h0, double _hL, bool _revAllowed) : Linesearch(_fct), h0(_h0), hL(_hL), revAllowed(_revAllowed)
{
    hU = 10.0;
    hS = h0 / 10.0;
    brckIt = 0;
    lsIt = 0;
}

/**
 * @brief Execute 3-points quadratic linesearch
 * @authors Romain Boman, Adrien Crovato
 */
double FleuryLS::run()
{
    double h = h0;
    double phi1 = fct.eval(0);
    double phit = fct.eval(h);
    fevalIt = 2;

    if (verbose)
    {
        std::cout << "Starting Backeting \n";
        std::cout << "\t" << std::setprecision(16) << "phi1(" << 0 << ") = " << phi1 << '\n';
        std::cout << "\t" << std::setprecision(16) << "phit(" << h << ") = " << phit << '\n';
    }

    // Following is valid if Hessian matrix is positive definite
    // -> derivative is negative at 0.
    double phi2 = 0.0, phi3;
    size_t nB;
    if (phit < phi1)
    {
        for (nB = 0;; ++nB)
        {
            if (h > hU)
            {
                if (verbose)
                {
                    std::cout << "Positive slope detected (in ascending bracketing #1)!\n";
                }

                if (revAllowed)
                {
                    if (verbose)
                    {
                        std::cout << "Reverse Direction of Line Search (in ascending bracketing #1)!"
                                  << "\n";
                    }
                    RevFunction revfct(fct);
                    FleuryLS revsearch(revfct, h0, hL, false);
                    return revsearch.run();
                }
                else
                {
                    if (verbose)
                    {
                        std::cout << "Error in ascending bracketing #1:\n"
                                  << std::setprecision(16) << "a1=" << 0 << " a2=" << h << " a3 =" << 2.0 * h << '\n'
                                  << std::setprecision(16) << "phi1=" << phi1 << " phi2=" << phi2 << " phit=" << phit << '\n'
                                  << "Using default step:\n"
                                  << std::setprecision(16) << "a=" << hS << "\n";
                    }
                    fct.eval(hS);
                    fevalIt++;
                    brckIt = nB + 1;
                    return hS;
                }
            }
            phi2 = phit;
            phit = fct.eval(2 * h);
            fevalIt++;

            if (verbose)
            {
                std::cout << "Ascending bracketing iteration # " << nB << ": \n";
                std::cout << "\t"
                          << "a1=" << 0 << " a2=" << h << " a3 =" << 2.0 * h << '\n';
                std::cout << "\t"
                          << "phi1=" << phi1 << " phi2=" << phi2 << " phit=" << phit << '\n';
            }

            if (phit > phi2)
                break;
            h = 2 * h;
        }
        phi3 = phit;
    }
    else
    {
        for (nB = 0;; ++nB)
        {
            if (h < hL)
            {
                if (verbose)
                {
                    std::cout << "Positive slope detected (in descending bracketing #2)!\n";
                }

                if (revAllowed)
                {
                    if (verbose)
                    {
                        std::cout << "Reverse Direction of Line Search (in descending bracketing #2) !"
                                  << "\n";
                    }
                    RevFunction revfct(fct);
                    FleuryLS revsearch(revfct, h0, hL, false);
                    return revsearch.run();
                }
                else
                {
                    if (verbose)
                    {
                        std::cout << "Error in descending bracketing #2:\n"
                                  << std::setprecision(16) << "a1=" << 0 << " a2=" << h << " a3 =" << 2.0 * h << '\n'
                                  << std::setprecision(16) << "phi1=" << phi1 << " phi2=" << phi2 << " phit=" << phit << '\n'
                                  << "Using default step:\n"
                                  << std::setprecision(16) << "a=" << hS << "\n";
                    }
                    fct.eval(hS);
                    fevalIt++;
                    brckIt = nB + 1;
                    return hS;
                }
            }
            phi3 = phit;
            phit = fct.eval(h / 2);
            fevalIt++;
            if (verbose)
            {
                std::cout << "Descending bracketing iteration # " << nB << ": \n";
                std::cout << "\t"
                          << "a1=" << 0 << " a3=" << h << " a2 =" << h / 2.0 << '\n';
                std::cout << "\t"
                          << "phi1=" << phi1 << " phi3=" << phi3 << " phit=" << phit << '\n';
            }
            if (phit < phi1)
                break;
            h = h / 2;
        }
        phi2 = phit;
        h = h / 2;
    }

    double a1 = 0;
    double a2 = h;
    double a3 = 2 * h;
    double a4 = h * (4 * phi2 - 3 * phi1 - phi3) / (4 * phi2 - 2 * phi1 - 2 * phi3);

    if (verbose)
    {
        std::cout << "Backeting succeeds in " << nB << " iterations : \n";
        std::cout << "\t" << std::setprecision(16) << "phi1(" << a1 << ") = " << phi1 << '\n';
        std::cout << "\t" << std::setprecision(16) << "phi2(" << a2 << ") = " << phi2 << '\n';
        std::cout << "\t" << std::setprecision(16) << "phi3(" << a3 << ") = " << phi3 << '\n';
    }

    double phi4;
    double aa = 0.0;
    size_t nLS = 0;
    double A = 0.;
    double errorPhi;
    double errorAlpha;

    for (nLS = 0;; ++nLS)
    {
        phi4 = fct.eval(a4);
        fevalIt++;

        errorAlpha = fabs(a4 - a2);
        errorPhi = fabs(phi4 - phi2);

        if (nLS == maxLsIt)
        {
            brckIt = nB + 1;
            lsIt = nLS;
            std::stringstream msg;
            msg << "Line-search doesn't converge within prescribed tolerance :\n";
            msg << "a1=" << std::setprecision(16) << a1 << " a2=" << a2 << " a3=" << a3 << " a4=" << a4 << '\n';
            msg << "phi1=" << std::setprecision(16) << phi1 << " phi2=" << phi2 << " phi3=" << phi3 << " phi4=" << phi4 << '\n';
            msg << "|A| = " << std::setprecision(16) << fabs(A) << ", |phi2-phi4| = " << errorPhi << " and |a2-a4| = " << errorAlpha << '\n';
            throw std::runtime_error(msg.str());
        }

        if (verbose)
        {
            std::cout << "Line search iteration # " << nLS << ": \n";
            std::cout << "\t" << std::setprecision(16) << "phi1(" << a1 << ") = " << phi1 << '\n';
            std::cout << "\t" << std::setprecision(16) << "phi2(" << a2 << ") = " << phi2 << '\n';
            std::cout << "\t" << std::setprecision(16) << "phi3(" << a3 << ") = " << phi3 << '\n';
            std::cout << "\t" << std::setprecision(16) << "phi4(" << a4 << ") = " << phi4 << '\n';
            std::cout << "\t" << std::setprecision(16) << "|phi2-phi4| = " << errorPhi << " and |a2-a4| = " << errorAlpha << '\n';
        }

        if (a1 <= a4 && a4 <= a2)
        {
            if (phi1 >= phi4 && phi4 <= phi2)
            {
                a3 = a2;
                phi3 = phi2;
                a2 = a4;
                phi2 = phi4;
            }
            else
            {
                a1 = a4;
                phi1 = phi4;
            }
        }
        else
        {
            if (phi2 >= phi4 && phi4 <= phi3)
            {
                a1 = a2;
                phi1 = phi2;
                a2 = a4;
                phi2 = phi4;
            }
            else
            {
                a3 = a4;
                phi3 = phi4;
            }
        }

        A = (a2 - a3) * phi1 + (a3 - a1) * phi2 + (a1 - a2) * phi3;

        if (verbose)
        {
            std::cout << "\t" << std::setprecision(16) << "|A| = " << fabs(A) << "><" << tol << '\n';
        }
        // phi2 ~ phi4 and a2 ~ a4 -> algorithm converges
        if (fabs(A) <= tol)
        {
            if (phi4 < phi2)
                aa = a4;
            else
                aa = a2;
            break;
        }

        a4 = 0.5 * (((a2 * a2) - (a3 * a3)) * phi1 + ((a3 * a3) - (a1 * a1)) * phi2 + ((a1 * a1) - (a2 * a2)) * phi3) /
             (A);
    }

    if (verbose)
    {
        std::cout << "Line search succeeds in " << nLS << " iterations : \n";
        std::cout << "\t" << std::setprecision(16) << "|A| = " << fabs(A) << '\n';
        std::cout << "\t" << std::setprecision(16) << "|phi4-phi2| = " << errorPhi << '\n';
        std::cout << "\t" << std::setprecision(16) << "|a4-a2| = " << errorAlpha << '\n';
        std::cout << "\t" << std::setprecision(16) << "phi1(" << a1 << ") = " << phi1 << '\n';
        std::cout << "\t" << std::setprecision(16) << "phi2(" << a2 << ") = " << phi2 << '\n';
        std::cout << "\t" << std::setprecision(16) << "phi3(" << a3 << ") = " << phi3 << '\n';
        std::cout << "\t" << std::setprecision(16) << "phi4(" << a4 << ") = " << phi4 << '\n';
    }
    brckIt = nB + 1;
    lsIt = nLS;
    return aa;
}

// ---

BankRoseLS::BankRoseLS(LSFunction &_fct) : Linesearch(_fct)
{
    K = 0.;
}

double BankRoseLS::run()
{
    double phi0 = fct.eval(0);
    fevalIt = 1;
    while (1)
    {
        double t = 1 / (1 + K * phi0);
        double phi = fct.eval(t);
        fevalIt++;
        if (1 / t * (1 - phi / phi0) > tol)
        {
            K /= 10;
            return t;
        }
        else if (fevalIt == maxLsIt)
        {
            std::stringstream msg;
            msg << "Line-search doesn't converge within prescribed tolerance:\n";
            msg << "iter = " << std::setprecision(16) << fevalIt << '\n';
            msg << "K = " << std::setprecision(16) << K << ", t = " << t << '\n';
            msg << "phi/phi0 = " << std::setprecision(16) << phi / phi0 << std::endl;
            throw std::runtime_error(msg.str());
        }
        else
        {
            if (K == 0.)
                K = 1;
            else
                K *= 10;
        }
    }
}