/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wMem.h"
#include "wElement.h"

using namespace tbox;

Mem::Mem(int _ngp) : ngp(_ngp), detJs(_ngp), Js(_ngp), Jinvs(_ngp)
{
    valid = false;
    vol = 0;
    cg = Pt();
}

/*
 * @brief Return the number of Gauss points
 * @authors Adrien Crovato
 */
double Mem::getGPn()
{
    return ngp;
}

/*
 * @brief Check that determinant of Jacobian is positive
 * @authors Adrien Crovato
 */
void Mem::checkJac(Element const *e, double detJ) const
{
    if (detJ <= 0)
    {
        std::stringstream err;
        err << "Negative jacobian detected for element: " << *e << "\n";
        throw std::runtime_error(err.str());
    }
}

/*
 * @brief Return the determinant of the Jacobian at Gauss point k
 * @authors Adrien Crovato
 */
double Mem::getDetJ(size_t k)
{
    if (valid)
        return detJs[k];
    else
        throw std::runtime_error("Mem::getDetJ is invalid because element memory is not up-to-date!\n");
}

/*
 * @brief Return the the Jacobian matrix at Gauss point k
 * @authors Adrien Crovato
 */
Eigen::MatrixXd const &Mem::getJ(size_t k)
{
    if (valid)
        return Js[k];
    else
        throw std::runtime_error("Mem::getJ is invalid because element memory is not up-to-date!\n");
}

/*
 * @brief Return the inverse of the Jacobian matrix at Gauss point k
 * @authors Adrien Crovato
 */
Eigen::MatrixXd const &Mem::getJinv(size_t k)
{
    if (valid)
        return Jinvs[k];
    else
        throw std::runtime_error("Mem::getJinv is invalid because element memory is not up-to-date!\n");
}

/*
 * @brief Return the volume of the element
 * @authors Adrien Crovato
 */
double Mem::getVol()
{
    if (valid)
        return vol;
    else
        throw std::runtime_error("Mem::getVol is invalid because element memory is not up-to-date!\n");
}

/*
 * @brief Return the center of gravity of the element
 * @authors Adrien Crovato
 */
Pt Mem::getCg()
{
    if (valid)
        return cg;
    else
        throw std::runtime_error("Mem::getCg is invalid because element memory is not up-to-date!\n");
}

void Mem::update(bool sameDim)
{
    throw std::runtime_error("Mem::update() not implemented!\n");
}
