/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WMEM_H
#define WMEM_H

#include "tbox.h"
#include "wPt.h"
#include <vector>
#include <Eigen/Dense>

#ifndef SWIG

namespace tbox
{

/**
 * @brief Base class for element private memory
 * @authors Romain Boman, Adrien Crovato
 * @todo Remove capability to getJ and getJinv when not volume element
 */
class TBOX_API Mem
{
protected:
    // Element value
    int ngp;    ///< number of Gauss points
    double vol; ///< element volume
    Pt cg;      ///< element center of gravity

    // GP values
    std::vector<double> detJs;          ///< determinant of Jacobian matrix
    std::vector<Eigen::MatrixXd> Js;    ///< Jacobian matrix
    std::vector<Eigen::MatrixXd> Jinvs; ///< inverse of Jacobian matrix

    // Functions
    void checkJac(Element const *e, double detJ) const;

public:
    explicit Mem(int _npg);
    virtual ~Mem() {}

    double getDetJ(size_t k);
    Eigen::MatrixXd const &getJ(size_t k);
    Eigen::MatrixXd const &getJinv(size_t k);
    double getGPn();
    double getVol();
    Pt getCg();

    bool valid; ///< validity flag (true if memory has been updated once)
    virtual void update(bool sameDim);
};

} // namespace tbox

#endif

#endif //WMEM_H
