/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wMumpsInterface.h"
#include <sstream>
using namespace tbox;

// macros so that indices match documentation
#define ICNTL(I) icntl[(I)-1]
#define INFO(I) info[(I)-1]

MumpsInterface::MumpsInterface(std::map<std::string, int> const &opts)
{
    id.comm_fortran = -987654; // USE_COMM_WORLD;
    id.par = 1;                // host involved in factorization phase
    id.sym = 0;                // 0=unsymmetric, 1=symdef+, 2=symmetric
    id.job = -1;               // initialization phase
    dmumps_c(&id);
    setOptions(opts); // options must be set after init (id.job != -1)
}

/**
 * @brief Setup the system of equations
 * @authors Adrien Crovato
 */
void MumpsInterface::setSoE(Eigen::SparseMatrix<double> const &A, Eigen::Map<Eigen::VectorXd> const &b)
{
    // Check that matrix A is square
    if (A.rows() != A.cols())
        throw std::runtime_error("tbox::MumpsInterface::setMat: Matrix A is not square!\n");
    // Build data structure
    // the matrix size, structure and number of nonzero entries can change at each iteration!
    size_t nz = A.nonZeros();
    rows.clear();
    cols.clear();
    matx.clear();
    vect.setZero();
    rows.reserve(nz);
    cols.reserve(nz);
    matx.reserve(nz);
    vect = b;
    // Get the nonzero indices and values
    for (size_t k = 0; k < A.outerSize(); ++k)
        for (Eigen::SparseMatrix<double>::InnerIterator it(A, k); it; ++it)
        {
            rows.push_back(static_cast<MUMPS_INT>(it.row() + 1));
            cols.push_back(static_cast<MUMPS_INT>(it.col() + 1));
            matx.push_back(it.value());
        }
    // Set sizes
    id.n = int(A.rows());
    id.nz = int(nz);
    // Set matrices
    id.a = &matx[0];
    id.rhs = vect.data();
    // Set id
    id.irn = &rows[0];
    id.jcn = &cols[0];
}

/**
 * @brief Analyze the structure of the matrix
 * @authors Adrien Crovato, Romain Boman
 */
void MumpsInterface::analyze()
{
    id.job = 1;
    dmumps_c(&id);
}
/**
 * @brief Farctorize the matrix
 * @authors Adrien Crovato, Romain Boman
 */
void MumpsInterface::factorize()
{
    id.job = 2;
    dmumps_c(&id);
}
/**
 * @brief Solve the SoE
 * @authors Adrien Crovato, Romain Boman
 */
void MumpsInterface::solve(Eigen::Map<Eigen::VectorXd> &x)
{
    id.job = 3;
    dmumps_c(&id);
    check();
    x = vect;
}
/**
 * @brief Analyze the structure of the matrix, factorize it and solve the SoE
 * @authors Adrien Crovato, Romain Boman
 */
void MumpsInterface::compute(Eigen::Map<Eigen::VectorXd> &x)
{
    id.job = 6;
    dmumps_c(&id);
    check();
    x = vect;
}

/**
 * @brief Terminate mumps
 * @authors Adrien Crovato, Romain Boman
 */
void MumpsInterface::finalize()
{
    id.job = -2;
    dmumps_c(&id);
}

/**
 * @brief Set the options for MUMPS
 * @authors Adrien Crovato, Romain Boman
 */
void MumpsInterface::setOptions(std::map<std::string, int> const &opts)
{
    // Verbosity level
    try
    {
        if (opts.at("verbosity") == 0)
        {
            id.ICNTL(1) = -1; // stream for error messages [def=6]
            id.ICNTL(2) = -1; // stream for diag printing, statistics, warnings [def=0]
            id.ICNTL(3) = -1; // stream for global information [def=6]
        }
        id.ICNTL(4) = opts.at("verbosity"); // level of printing [def=2]
    }
    catch (const std::out_of_range&)
    {
        id.ICNTL(1) = -1;
        id.ICNTL(2) = -1;
        id.ICNTL(3) = -1;
        id.ICNTL(4) = 0;
    }
    // Matrix re-ordering
    //0 : amd, 2 : amf, 3 : scotch, 4 : pord, 5 : metis, 6 :qamd, 7 : auto [def=7]
    try
    {
        id.ICNTL(7) = opts.at("ordering");
        if (opts.at("ordering") == 5)
        {
            id.ICNTL(28) = 2; // parallel ordering
            id.ICNTL(29) = 2; // parmetis
        }
        else
            id.ICNTL(28) = 1; // sequential ordering
    }
    catch (const std::out_of_range&)
    {
        id.ICNTL(7) = 3; // defaults to scotch because mumps comes witch scotch if install using apt
    }
    // Error analysis
    try
    {
        id.ICNTL(11) = opts.at("error_analysis"); // enable error analysis [def=0]
    }
    catch (const std::out_of_range&)
    {
        id.ICNTL(11) = 0; // disabled
    }
    // Increase memory
    id.ICNTL(14) = 40;
    // Determinant
    try
    {
        id.ICNTL(33) = opts.at("compute_determinant"); // compute determinant of A
    }
    catch (const std::out_of_range&)
    {
        id.ICNTL(33) = 0;
    }
    // Not adressed
    // id.ICNTL(5)   // matrix input format (0=A,irn,jcn; 1=elemental)
    // id.ICNTL(6)   // permutation/scaling
    // id.ICNTL(8)   // scaling strategy [def=auto]
    // id.ICNTL(9)   // use A or A^T [def=A]
    // id.ICNTL(10)  // iterative refinement [def=0=disabled]
    // id.ICNTL(11)  // compute statistics on error [def=0=disabled]
    // id.ICNTL(12)  // ordering strategy for sym matrices [def=0=auto]
    // id.ICNTL(13)  // parallelism of root node (scalapack) [def=0=parallel with scalapack]
    // id.ICNTL(14)  // % incr of working space [def=20=20%]
    // id.ICNTL(15-17)  // NOT USED
    // id.ICNTL(18)  // distributed input matrix [def=0=centralized]
    // id.ICNTL(19)  // Schur complement [def=0=no schur cplt]
    // id.ICNTL(20)  // format of rhs [def=0=dense]
    // id.ICNTL(21)  // distribution of solution vectors [def=0=centralized]
    // id.ICNTL(22)  // out-of-core [def=0=in-core]
    // id.ICNTL(23)  // max memory [def=0=estimate]
    // id.ICNTL(24)  // null pivot detection [def=0=disabled]
    // id.ICNTL(25)  // solution for deficiant matrix [def=0=1 sol is returned]
    // id.ICNTL(26)  // [see schur cplt]
    // id.ICNTL(27)  // blocking size for multiple rhs [def=-32=auto]
    // id.ICNTL(30)  // compute some A^-1 entries
    // id.ICNTL(31)  // keep factors [def=0=yes]
    // id.ICNTL(32)  // forward elimination during factorization [def=0=disabled]
    // id.ICNTL(34)  // NOT USED
    // id.ICNTL(35)  // BLR factorization (def=0=disabled)
}

/**
 * @brief Check if an error occured during solve
 * @authors Adrien Crovato
 */
void MumpsInterface::check()
{
    if (id.INFO(1) < 0)
    {
        std::stringstream err;
        err << "tbox::MumpsInterface: MUMPS failed with error " << id.INFO(1);
        switch (id.INFO(1))
        {
        case -2:
            err << ": NZ = " << id.INFO(2) << " is out of range!\n";
            throw std::runtime_error(err.str());
        case -6:
        case -10:
            err << ": matrix is singular!\n";
            throw std::runtime_error(err.str());
        case -9:
            err << ": increase ICNTL(14)!\n";
            throw std::runtime_error(err.str());
        case -13:
            err << ": not enough memory!\n";
            throw std::runtime_error(err.str());
        default:
            err << "!\n";
            throw std::runtime_error(err.str());
        }
    }
}

void MumpsInterface::write(std::ostream &out) const
{
    out << "MumpsInterface with SoE of size " << matx.size() << std::endl;
}