/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wQuad4.h"
#include "wSfQuad4.h"
#include "wGaussQuad4.h"
#include "wCacheQuad4.h"
#include "wMemQuad4.h"
#include "wNode.h"
#include "wMath.h"
#include "wLazy.h"
#include "wFct0.h"

using namespace tbox;

Quad4::Quad4(int n, Tag *_ptag, Tag *_etag,
             std::vector<Tag *> &_parts, std::vector<Node *> &nods) : Element(n, _ptag, _etag, _parts, nods)
{
    pmem = new MemQuad4(*this, static_cast<int>(getCache().gauss.p.size()));
}
Quad4::~Quad4()
{
    if (pmem)
        delete pmem;
}

/*
 * @brief Private memory handling Jacobian for Quad4 Element 
 * @authors Romain Boman, Adrien Crovato
 */
Mem &Quad4::getVMem()
{
    return getMem();
}
MemQuad4 &Quad4::getMem()
{
    return *pmem;
}

/*
 * @brief Private cache handling shape functions and Gauss points for Quad4 
 * @authors Romain Boman, Adrien Crovato
 */
CacheQuad4 &Quad4::getCache()
{
    static Lazy<CacheQuad4> cache(2);
    return cache.get();
}
Cache &Quad4::getVCache()
{
    return getCache();
}

Pt Quad4::normal()
{
    Pt x1 = nodes[0]->pos;
    Pt x2 = nodes[1]->pos;
    Pt x3 = nodes[2]->pos;
    Pt x4 = nodes[3]->pos;

    return ((x3 - x1) ^ (x4 - x1)).normalized();
}

/**
 * @brief Compute Jacobian determinant at Gauss point k
 * @authors Romain Boman
 */
double Quad4::buildJ(size_t k) const
{
    CacheQuad4 &cache = getCache();

    Eigen::Vector3d t0 = Eigen::Vector3d::Zero();
    Eigen::Vector3d t1 = Eigen::Vector3d::Zero();
    size_t i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
    {
        std::vector<double> &dff = cache.dff[k][i];
        t0 += Eigen::Vector3d((*it)->pos.x.data()) * dff[0];
        t1 += Eigen::Vector3d((*it)->pos.x.data()) * dff[1];
    }
    return t0.cross(t1).norm();
}
/**
 * @brief Compute Jacobian matrix at Gauss point k
 * @authors Romain Boman
 */
double Quad4::buildJ(size_t k, Eigen::MatrixXd &J) const
{
    CacheQuad4 &cache = getCache();

    J = Eigen::Matrix2d::Zero();
    size_t i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
    {
        std::vector<double> &dff = cache.dff[k][i];
        J += Eigen::Vector2d(dff.data()) * Eigen::RowVector3d((*it)->pos.x.data()).block<1, 2>(0, 0);
    }
    return J.determinant();
}

void Quad4::buildS(Eigen::MatrixXd &S)
{
    CacheQuad4 &cache = getCache();
    GaussQuad4 &gauss = cache.gauss;
    MemQuad4 &mem = getMem();

    S = Eigen::Matrix4d::Zero();
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // Jacobian
        double detJ = mem.getDetJ(k);

        std::vector<double> &ff = cache.ff[k];
        S += Eigen::Vector4d(ff.data()) * Eigen::RowVector4d(ff.data()) * gauss.w[k] * detJ; // S = S + {ff}^T [ff].wG.detJ
    }
}

void Quad4::build(MATTYPE type, Eigen::MatrixXd &A)
{
    switch (type)
    {
    case MATTYPE::MAT_S:
        buildS(A);
        break;
    default:
        std::stringstream out;
        out << "Quad4 cannot build " << type;
        throw std::runtime_error(out.str());
    }
}

/**
 * @brief Compute intergal of function
 * 
 * I = int fct dV
 * @authors Romain Boman
 */
double Quad4::computeV(std::vector<double> const &u, Fct0 const &fct)
{
    CacheQuad4 &cache = getCache();
    GaussQuad4 &gauss = cache.gauss;
    MemQuad4 &mem = getMem();

    double V = 0.0;
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // Function evaluation
        double fk = fct.eval(*this, u, k);
        double detJ = mem.getDetJ(k);
        V += fk * gauss.w[k] * detJ;
    }
    return V;
}

/**
 * @brief Return the element volume
 * @authors Adrien Crovato
 */
double Quad4::computeV()
{
    MemQuad4 &mem = getMem();
    return mem.getVol();
}