/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wTetra4.h"
#include "wSfTetra4.h"
#include "wGaussTetra4.h"
#include "wCacheTetra4.h"
#include "wMemTetra4.h"
#include "wNode.h"
#include "wLazy.h"
#include "wFct0.h"

using namespace tbox;

Tetra4::Tetra4(int n, Tag *_ptag, Tag *_etag,
               std::vector<Tag *> &_parts, std::vector<Node *> &nods) : Element(n, _ptag, _etag, _parts, nods)
{
    pmem = new MemTetra4(*this, static_cast<int>(getCache().gauss.p.size()));
}
Tetra4::~Tetra4()
{
    if (pmem)
        delete pmem;
}

/*
 * @brief Private memory handling Jacobian for Tetra4 Element 
 * @authors Romain Boman, Adrien Crovato
 */
Mem &Tetra4::getVMem()
{
    return getMem();
}
MemTetra4 &Tetra4::getMem()
{
    return *pmem;
}

/*
 * @brief Private cache handling shape functions and Gauss points for Tetra4 
 * @authors Romain Boman, Adrien Crovato
 */
CacheTetra4 &Tetra4::getCache()
{
    static Lazy<CacheTetra4> cache(2);
    return cache.get();
}
Cache &Tetra4::getVCache()
{
    return getCache();
}

/**
 * @brief Compute Jacobian Matrix at Gauss point k
 * @authors Romain Boman
 */
double Tetra4::buildJ(size_t k, Eigen::MatrixXd &J) const
{
    CacheTetra4 &cache = getCache();

    J = Eigen::Matrix3d::Zero();
    size_t i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
    {
        std::vector<double> &dff = cache.dff[k][i];
        J += Eigen::Vector3d(dff.data()) * Eigen::RowVector3d((*it)->pos.x.data());
    }
    return J.determinant();
}

/**
 * @brief Compute mass matrix
 * @authors Romain Boman
 */
void Tetra4::buildM(Eigen::MatrixXd &M)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    M = Eigen::Matrix4d::Zero();
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // Jacobian
        double detJ = mem.getDetJ(k);

        std::vector<double> &ff = cache.ff[k];
        M += Eigen::Vector4d(ff.data()) * Eigen::RowVector4d(ff.data()) * gauss.w[k] * detJ; // M = M + {ff}^T [ff].wG.detJ
    }
}

/**
 * @brief Compute stiffness matrix for scalar field
 * 
 * K_ij(4,4) = int fct*dN_j*dN_i dV
 * @authors Romain Boman, Adrien Crovato
 */
void Tetra4::buildK(Eigen::MatrixXd &K, std::vector<double> const &u, Fct0 const &fct)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    K = Eigen::Matrix4d::Zero();
    Eigen::Vector3d factor1;
    Eigen::Vector3d factor2;
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // Function and Jacobian evaluation
        double fk = fct.eval(*this, u, k);
        double detJ = mem.getDetJ(k);
        Eigen::Matrix3d const &J = mem.getJinv(k);

        // compute stiffness matrix for each GP
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[k][i];
            factor1 = J * Eigen::Vector3d(dffi.data());

            for (size_t j = 0; j < nodes.size(); ++j)
            {
                std::vector<double> &dffj = cache.dff[k][j];
                factor2 = J * Eigen::Vector3d(dffj.data());
                K(i, j) += factor1.dot(factor2) * gauss.w[k] * detJ * fk;
            }
        }
    }
}

/**
 * @brief Compute (first part of) derivative stiffness matrix for scalar field (Newton method)
 * 
 * K_ij(4,4) = int (df du*dN_j du*dN_i) dV
 * @authors Adrien Crovato
 */
void Tetra4::buildDK(Eigen::MatrixXd &K, std::vector<double> const &u, Fct0 const &fct)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    K = Eigen::Matrix4d::Zero();
    Eigen::Vector3d factor1;
    Eigen::Vector3d factor2;
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // "derivative" of function, Jacobian and gradient evaluation
        double dfk = fct.evalD(*this, u, k);
        double detJ = mem.getDetJ(k);
        Eigen::Matrix3d const &J = mem.getJinv(k);
        Eigen::VectorXd gradu;
        computeGrad(u, gradu, k);

        // computation of K
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[k][i];
            factor1 = J * Eigen::Vector3d(dffi.data());
            for (size_t j = 0; j < nodes.size(); ++j)
            {
                std::vector<double> &dffj = cache.dff[k][j];
                factor2 = J * Eigen::Vector3d(dffj.data());
                K(i, j) += gradu.dot(factor1) * gradu.dot(factor2) * gauss.w[k] * detJ * dfk;
            }
        }
    }
}

/**
 * @brief Compute derivative of flux vector for a scalar field (Newton method)
 * 
 * s_i(4) = int f*du*dN_i dV
 * @authors Adrien Crovato
 */
void Tetra4::buildDs(Eigen::VectorXd &s, std::vector<double> const &u, Fct0 const &f)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    s = Eigen::Vector4d::Zero();
    Eigen::Vector3d factor1;
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // Gradient, function and Jacobian
        Eigen::VectorXd gradu;
        this->computeGrad(u, gradu, k);
        double fk = f.eval(*this, u, k);
        double detJ = mem.getDetJ(k);
        Eigen::Matrix3d const &J = mem.getJinv(k);

        // s(i) = f * gradu * gradN(i)
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[k][i];
            factor1 = J * Eigen::Vector3d(dffi.data());
            s(i) += gradu.dot(factor1) * gauss.w[k] * detJ * fk;
        }
    }
}

/**
 * @brief Compute linear elasticity stiffness matrix
 * @authors Adrien Crovato
 * @todo this is a clean duplicate of buildK_mechnical, should be merged
 */
void Tetra4::buildK_linElastic(Eigen::MatrixXd &K, Eigen::MatrixXd const &H)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    K = Eigen::MatrixXd::Zero(12, 12);
    Eigen::Vector3d factori;
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // Jacobian
        double detJ = mem.getDetJ(k);
        Eigen::Matrix3d const &invJ = mem.getJinv(k);

        // Fill B matrix
        Eigen::MatrixXd B = Eigen::MatrixXd::Zero(6, 3 * nodes.size());
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            // Compute [Ni,x Ni,y, Ni_z]
            std::vector<double> &dffi = cache.dff[k][i];
            factori = invJ * Eigen::Vector3d(dffi.data());
            B(0, 3 * i) = factori(0);
            B(1, 3 * i + 1) = factori(1);
            B(2, 3 * i + 2) = factori(2);
            B(3, 3 * i + 1) = factori(2);
            B(3, 3 * i + 2) = factori(1);
            B(4, 3 * i) = factori(2);
            B(4, 3 * i + 2) = factori(0);
            B(5, 3 * i) = factori(1);
            B(5, 3 * i + 1) = factori(0);
        }

        // Compute stiffness matrix
        K += B.transpose() * H * B * detJ * gauss.w[k];
    }
}

/* -------------------------------------------------------------------------- */
/**
 * @brief Compute mechanical stiffness matrix
 * @authors Kim Liegeois
 */
void Tetra4::buildK_mechanic(Eigen::MatrixXd &K_e, Eigen::MatrixXd const &H)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    K_e = Eigen::MatrixXd::Zero(12, 12);
    Eigen::Vector3d factor1;
    Eigen::Vector3d factor2;
    Eigen::Vector3d factor3;
    Eigen::Matrix3d TMP1;
    for (size_t a = 0; a < gauss.p.size(); ++a)
    {
        // calcul de la matrice jacobienne (au point de Gauss "ksi")
        double detJ = mem.getDetJ(a);
        Eigen::Matrix3d const &J = mem.getJinv(a);

        // calcul de K (au point de Gauss "ksi")
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[a][i];
            factor1 = J * Eigen::Vector3d(dffi.data());
            for (size_t j = i; j < nodes.size(); ++j)
            {
                std::vector<double> &dffj = cache.dff[a][j];
                factor2 = J * Eigen::Vector3d(dffj.data());
                for (size_t k = 0; k < 3; ++k)
                    for (size_t m = 0; m < 3; ++m)
                    {
                        for (size_t l = 0; l < 3; ++l)
                            for (size_t n = 0; n < 3; ++n)
                                TMP1(l, n) = H(k * 3 + l, m * 3 + n);
                        factor3 = TMP1 * factor2;
                        K_e(i * 3 + k, j * 3 + m) += factor1.dot(factor3) * gauss.w[a] * detJ;
                    }
            }
        }
    }
    for (size_t i = 0; i < 3 * nodes.size(); ++i)
        for (size_t j = i; j < 3 * nodes.size(); ++j)
            K_e(j, i) = K_e(i, j);
}

/**
 * @brief Compute thermomechanical mass matrix
 * @authors Kim Liegeois
 */
void Tetra4::buildS_thermomechanic(Eigen::MatrixXd &S_e, Eigen::MatrixXd const &D)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    S_e = Eigen::MatrixXd::Zero(12, 4);
    Eigen::Vector3d factor1;
    Eigen::Vector3d factor2;
    for (size_t a = 0; a < gauss.p.size(); ++a)
    {
        // calcul de la matrice jacobienne (au point de Gauss "ksi")
        double detJ = mem.getDetJ(a);
        Eigen::Matrix3d const &J = mem.getJinv(a);
        std::vector<double> &ff = cache.ff[a];

        // calcul de K (au point de Gauss "ksi")
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[a][i];
            factor1 = J * Eigen::Vector3d(dffi.data());
            factor2 = D * factor1;
            for (size_t j = 0; j < nodes.size(); ++j)
                for (size_t k = 0; k < 3; ++k)
                    S_e(i * 3 + k, j) += factor2[k] * ff[j] * gauss.w[a] * detJ;
        }
    }
}

/**
 * @brief Compute thermic matrix
 * @authors Kim Liegeois
 */
void Tetra4::buildL_thermic(Eigen::MatrixXd &L_e, Eigen::MatrixXd const &K_conductivity)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    L_e = Eigen::Matrix4d::Zero();
    Eigen::Vector3d factor1;
    Eigen::Vector3d factor2;
    Eigen::Vector3d factor3;
    for (size_t a = 0; a < gauss.p.size(); ++a)
    {
        // calcul de la matrice jacobienne (au point de Gauss "ksi")
        double detJ = mem.getDetJ(a);
        Eigen::Matrix3d const &J = mem.getJinv(a);

        // calcul de K (au point de Gauss "ksi")
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[a][i];
            factor1 = J * Eigen::Vector3d(dffi.data());
            factor2 = K_conductivity * factor1;
            for (size_t j = i; j < nodes.size(); ++j)
            {
                std::vector<double> &dffj = cache.dff[a][j];
                factor3 = J * Eigen::Vector3d(dffj.data());
                L_e(i, j) += factor2.dot(factor3) * gauss.w[a] * detJ;
            }
        }
    }
    for (size_t i = 0; i < nodes.size(); ++i)
        for (size_t j = i; j < nodes.size(); ++j)
            L_e(j, i) = L_e(i, j);
}

/**
 * @brief Compute stress and strain
 * @authors Kim Liegeois
 */
void Tetra4::strain_stress(Eigen::MatrixXd &Strain, Eigen::MatrixXd &Stress, Eigen::MatrixXd const &H, std::vector<double> const &u)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    Strain = Stress = Eigen::Matrix3d::Zero();
    Eigen::Matrix3d Tmp = Eigen::Matrix3d::Zero();
    Eigen::MatrixXd u_e(3, nodes.size());
    for (size_t i = 0; i < nodes.size(); ++i)
    {
        u_e(0, i) = u[3 * (nodes[i]->row)];
        u_e(1, i) = u[3 * (nodes[i]->row) + 1];
        u_e(2, i) = u[3 * (nodes[i]->row) + 2];
    }
    for (size_t a = 0; a < gauss.p.size(); ++a)
    {
        // calcul de la matrice jacobienne (au point de Gauss "ksi")
        Eigen::Matrix3d const &J = mem.getJinv(a);
        Eigen::Matrix3d du_d_Phi = Eigen::Matrix3d::Zero();

        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[a][i];
            for (auto k = 0; k < 3; ++k)
                for (auto j = 0; j < 3; ++j)
                    du_d_Phi(j, k) += dffi[j] * u_e(k, i);
        }

        for (size_t k = 0; k < 3; ++k)
            for (size_t m = 0; m < 3; ++m)
                for (size_t i = 0; i < 3; ++i)
                    Tmp(k, m) += J(k, i) * du_d_Phi(i, m) / (nodes.size());
    }
    for (size_t i = 0; i < 3; ++i)
        for (size_t j = 0; j < 3; ++j)
            Strain(i, j) = (Tmp(i, j) + Tmp(j, i)) / 2;

    for (size_t i = 0; i < 3; ++i)
        for (size_t j = 0; j < 3; ++j)
            for (size_t k = 0; k < 3; ++k)
                for (size_t l = 0; l < 3; ++l)
                    Stress(i, j) += H(i * 3 + j, k * 3 + l) * Strain(k, l);
}

/* -------------------------------------------------------------------------- */

/**
 * @brief Compute gradient of a scalar field at Gauss point k
 * @authors Romain Boman, Adrien Crovato
 */
void Tetra4::computeGrad(std::vector<double> const &u, Eigen::VectorXd &grad, size_t k)
{
    CacheTetra4 &cache = getCache();
    MemTetra4 &mem = getMem();

    // local matrix of sf derivatives, solution vector and Jacobian
    Eigen::MatrixXd gradN(3, nodes.size());
    Eigen::Vector4d ue;
    for (size_t i = 0; i < nodes.size(); ++i)
    {
        gradN.col(i) = Eigen::Vector3d(cache.dff[k][i].data());
        ue(i) = u[nodes[i]->row];
    }
    Eigen::Matrix3d const &Jinv = mem.getJinv(k);
    // gradient in global axis: inv(J)_ij * d_jN_i * ue_i
    grad = Jinv * gradN * ue;
}

/**
 * @brief Compute square of gradient of a scalar field at Gauss point k
 * @authors Adrien Crovato
 */
double Tetra4::computeGrad2(std::vector<double> const &u, size_t k)
{
    Eigen::VectorXd grad;
    computeGrad(u, grad, k);
    return grad.squaredNorm();
}

/**
 * @brief Compute intergal of function
 * 
 * I = int fct dV
 * @authors Adrien Crovato
 */
double Tetra4::computeV(std::vector<double> const &u, Fct0 const &fct)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    double V = 0.0;
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // function evaluation
        double fk = fct.eval(*this, u, k);
        double detJ = mem.getDetJ(k);
        V += fk * gauss.w[k] * detJ;
    }
    return V;
}

/**
 * @brief Return the element volume
 * @authors Adrien Crovato
 */
double Tetra4::computeV()
{
    MemTetra4 &mem = getMem();
    return mem.getVol();
}
