/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WTETRA4_H
#define WTETRA4_H

#include "tbox.h"
#include "wElement.h"

namespace tbox
{

/**
 * @brief Linear tetrahedron finite element
 * @authors Kim Liegeois, Romain Boman, Adrien Crovato
 */
class TBOX_API Tetra4 : public Element
{
#ifndef SWIG
    friend class MemTetra4;
    MemTetra4 *pmem; ///< private memory of precalculated values
    static CacheTetra4 &getCache();
    MemTetra4 &getMem();
    virtual double buildJ(size_t k, Eigen::MatrixXd &J) const override;

    virtual void buildM(Eigen::MatrixXd &M) override;
    virtual void buildK(Eigen::MatrixXd &K, std::vector<double> const &u, Fct0 const &fct) override;
    virtual void buildDK(Eigen::MatrixXd &K, std::vector<double> const &u, Fct0 const &fct) override;
    virtual void buildDs(Eigen::VectorXd &s, std::vector<double> const &u, Fct0 const &f) override;
    virtual void buildK_linElastic(Eigen::MatrixXd &K, Eigen::MatrixXd const &H) override;

    /* @todo Kim's work, will have to be cleaned at some point */
    virtual void buildK_mechanic(Eigen::MatrixXd &K_e, Eigen::MatrixXd const &H) override;
    virtual void buildS_thermomechanic(Eigen::MatrixXd &S_e, Eigen::MatrixXd const &D) override;
    virtual void buildL_thermic(Eigen::MatrixXd &L_e, Eigen::MatrixXd const &K_conductivity) override;
    virtual void strain_stress(Eigen::MatrixXd &Strain, Eigen::MatrixXd &Stress, Eigen::MatrixXd const &H, std::vector<double> const &u) override;
    /* --- */

    virtual void computeGrad(std::vector<double> const &u, Eigen::VectorXd &grad, size_t k) override;
    virtual double computeGrad2(std::vector<double> const &u, size_t k) override;
    virtual double computeV(std::vector<double> const &u, Fct0 const &fct) override;
    virtual double computeV() override;
#endif

public:
    Tetra4(int n, Tag *_ptag, Tag *_etag, std::vector<Tag *> &_parts, std::vector<Node *> &nods);
    virtual ~Tetra4();
    virtual ELTYPE type() const override { return ELTYPE::TETRA4; }

#ifndef SWIG
    virtual Cache &getVCache() override;
    virtual Mem &getVMem() override;
#endif
};

} // namespace tbox

#endif //WTETRA4_H
