/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wTri3.h"
#include "wSfTri3.h"
#include "wGaussTri3.h"
#include "wCacheTri3.h"
#include "wMemTri3.h"
#include "wNode.h"
#include "wMath.h"
#include "wLazy.h"
#include "wFct0.h"
#include "wFct1.h"
#include "wFct2.h"

using namespace tbox;

Tri3::Tri3(int n, Tag *_ptag, Tag *_etag,
           std::vector<Tag *> &_parts, std::vector<Node *> &nods) : Element(n, _ptag, _etag, _parts, nods)
{
    pmem = new MemTri3(*this, static_cast<int>(getCache().gauss.p.size()));
}
Tri3::~Tri3()
{
    if (pmem)
        delete pmem;
}

/*
 * @brief Private memory handling Jacobian for Tri3 Element 
 * @authors Romain Boman, Adrien Crovato
 */
Mem &Tri3::getVMem()
{
    return getMem();
}
MemTri3 &Tri3::getMem()
{
    return *pmem;
}

/*
 * @brief Private cache handling shape functions and Gauss points for Tri3 
 * @authors Romain Boman, Adrien Crovato
 */
CacheTri3 &Tri3::getCache()
{
    static Lazy<CacheTri3> cache(2);
    return cache.get();
}
Cache &Tri3::getVCache()
{
    return getCache();
}

Pt Tri3::normal()
{
    Pt x1 = nodes[0]->pos;
    Pt x2 = nodes[1]->pos;
    Pt x3 = nodes[2]->pos;

    return ((x2 - x1) ^ (x3 - x1)).normalized();
}

/**
 * @brief Compute Jacobian determinant at Gauss point k
 * @authors Romain Boman
 */
double Tri3::buildJ(size_t k) const
{
    CacheTri3 &cache = getCache();

    Eigen::Vector3d t0 = Eigen::Vector3d::Zero();
    Eigen::Vector3d t1 = Eigen::Vector3d::Zero();
    size_t i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
    {
        std::vector<double> &dff = cache.dff[k][i];
        t0 += Eigen::Vector3d((*it)->pos.x.data()) * dff[0];
        t1 += Eigen::Vector3d((*it)->pos.x.data()) * dff[1];
    }
    return t0.cross(t1).norm();
}
/**
 * @brief Compute Jacobian matrix at Gauss point k
 * @authors Romain Boman
 */
double Tri3::buildJ(size_t k, Eigen::MatrixXd &J) const
{
    CacheTri3 &cache = getCache();

    J = Eigen::Matrix2d::Zero();
    size_t i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
    {
        std::vector<double> &dff = cache.dff[k][i];
        J += Eigen::Vector2d(dff.data()) * Eigen::RowVector3d((*it)->pos.x.data()).block<1, 2>(0, 0);
    }
    return J.determinant();
}

/**
 * @brief Compute stiffness matrix for scalar field
 * 
 * K_ij(3,3) = int fct*dN_j*dN_i dV
 * @authors Romain Boman
 */
void Tri3::buildK(Eigen::MatrixXd &K, std::vector<double> const &u, Fct0 const &fct)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    K = Eigen::Matrix3d::Zero();
    Eigen::Vector2d factor1;
    Eigen::Vector2d factor2;
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // Function and Jacobian inverse
        double fk = fct.eval(*this, u, k);
        double detJ = mem.getDetJ(k);
        Eigen::Matrix2d const &J = mem.getJinv(k);

        // Elementary stiffness matrix
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[k][i];
            factor1 = J * Eigen::Vector2d(dffi.data());

            for (size_t j = 0; j < nodes.size(); ++j)
            {
                std::vector<double> &dffj = cache.dff[k][j];
                factor2 = J * Eigen::Vector2d(dffj.data());

                K(i, j) += factor1.dot(factor2) * gauss.w[k] * detJ * fk;
            }
        }
        // other way, for the future
        //Eigen::Matrix<double, 2, 3> dN;
        //for (size_t i = 0; i < nodes.size(); ++i)
        //    dN.col(i) = Eigen::Vector2d(cache.dff[k][i].data());
        //K = (J * dN).transpose() * J * dN * gauss.w[k] * detJ * fk;
    }
}

void Tri3::buildK(Eigen::MatrixXd &K, std::vector<double> const &u, Fct2 const &fct, bool fake)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;

    if (fake) // fake run - fill MPI job list
    {
        for (size_t k = 0; k < gauss.p.size(); ++k)
        {
            Eigen::MatrixXd fk;
            fct.eval(*this, u, k, fk, fake);
        }
    }
    else // true run - use MPI cached results
    {
        MemTri3 &mem = getMem();
        K = Eigen::Matrix3d::Zero();
        Eigen::Vector2d factor0;
        Eigen::Vector2d factor1;
        Eigen::Vector2d factor2;
        for (size_t k = 0; k < gauss.p.size(); ++k)
        {
            // Function and Jacobian inverse
            Eigen::MatrixXd fk;
            fct.eval(*this, u, k, fk, fake);
            double detJ = mem.getDetJ(k);
            Eigen::Matrix2d const &J = mem.getJinv(k);

            // Elementary stiffness matrix
            for (size_t i = 0; i < nodes.size(); ++i)
            {
                std::vector<double> &dffi = cache.dff[k][i];
                factor0 = J * Eigen::Vector2d(dffi.data());
                factor1 = fk * factor0;

                for (size_t j = 0; j < nodes.size(); ++j)
                {
                    std::vector<double> &dffj = cache.dff[k][j];
                    factor2 = J * Eigen::Vector2d(dffj.data());
                    K(i, j) += factor1.dot(factor2) * gauss.w[k] * detJ;
                }
            }
        }
    }
}

/**
 * @brief Compute mechanical stiffness matrix
 * 
 * K_ij(6,6) = int eps_k H_ijkl eps_l dV = B(6,3)H(3,3)B(3,6)
 * @authors Adrien Crovato
 */
void Tri3::buildK_linElastic(Eigen::MatrixXd &K, Eigen::MatrixXd const &H)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    K = Eigen::MatrixXd::Zero(6, 6);
    Eigen::Vector2d factori;
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // Jacobian inverse
        double detJ = mem.getDetJ(k);
        Eigen::Matrix2d const &invJ = mem.getJinv(k);

        // Fill B matrix
        Eigen::MatrixXd B = Eigen::MatrixXd::Zero(3, 6);
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            // Compute [Ni,x Ni,y]
            std::vector<double> &dffi = cache.dff[k][i];
            factori = invJ * Eigen::Vector2d(dffi.data());
            B(0, 2 * i) = factori(0);
            B(1, 2 * i + 1) = factori(1);
            B(2, 2 * i + 1) = factori(0);
            B(2, 2 * i) = factori(1);
        }

        // Compute stiffness matrix
        K += B.transpose() * H * B * detJ * gauss.w[k];
    }
}

/**
 * @brief Compute (first part of) derivative stiffness matrix for scalar field (Newton method)
 * 
 * K_ij(3,3) = int (df du*dN_j du*dN_i) dV
 * @authors Adrien Crovato
 */
void Tri3::buildDK(Eigen::MatrixXd &K, std::vector<double> const &u, Fct0 const &fct)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    K = Eigen::Matrix3d::Zero();
    Eigen::Vector2d factor1;
    Eigen::Vector2d factor2;
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // "derivative" of f, gradient and inverse Jacobian
        double dfk = fct.evalD(*this, u, k);
        Eigen::VectorXd gradu;
        this->computeGrad(u, gradu, k);
        double detJ = mem.getDetJ(k);
        Eigen::Matrix2d const &J = mem.getJinv(k);

        // computation of K
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[k][i];
            factor1 = J * Eigen::Vector2d(dffi.data());
            for (size_t j = 0; j < nodes.size(); ++j)
            {
                std::vector<double> &dffj = cache.dff[k][j];
                factor2 = J * Eigen::Vector2d(dffj.data());
                K(i, j) += gradu.dot(factor1) * gradu.dot(factor2) * gauss.w[k] * detJ * dfk;
            }
        }
    }
}

/**
 * @brief Compute flux vector for a scalar field
 * 
 * s_i(3) = int f*N_i dV
 * @authors Romain Boman
 */
void Tri3::builds(Eigen::VectorXd &s, std::vector<double> const &u, Fct0 const &f)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    s = Eigen::Vector3d::Zero();
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        double sk = f.eval(*this, u, k);
        double detJ = mem.getDetJ(k);
        s += Eigen::Vector3d(cache.ff[k].data()) * sk * gauss.w[k] * detJ; // s = s + sk.wG.detJ
    }
}

/**
 * @brief Compute flux vector for scalar field
 * 
 * s_i(3) = int f_j*n_j * N_i dV 
 * @authors Adrien Crovato
 */
void Tri3::builds(Eigen::VectorXd &s, std::vector<double> const &u, Fct1 const &f)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    s = Eigen::Vector3d::Zero();
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        double sk = f.eval(*this, u, k) * normal();
        double detJ = mem.getDetJ(k);
        s += Eigen::Vector3d(cache.ff[k].data()) * sk * gauss.w[k] * detJ; // s = s + sk.wG.detJ
    }
}

/**
 * @brief Compute derivative of flux vector for a scalar field (Newton method)
 * 
 * s_i(3) = int f*du*dN_i dV
 * @authors Adrien Crovato
 */
void Tri3::buildDs(Eigen::VectorXd &s, std::vector<double> const &u, Fct0 const &f)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    s = Eigen::Vector3d::Zero();
    Eigen::Vector2d factor1;
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        Eigen::VectorXd gradu;
        this->computeGrad(u, gradu, k);
        double fk = f.eval(*this, u, k);
        double detJ = mem.getDetJ(k);
        Eigen::Matrix2d const &J = mem.getJinv(k);

        // s(i) = f * gradu * gradN(i)
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[k][i];
            factor1 = J * Eigen::Vector2d(dffi.data());
            s(i) += gradu.dot(factor1) * gauss.w[k] * detJ * fk;
        }
    }
}

/**
 * @brief Compute something
 * 
 * S_ij(3,3) = int N_i*N_j dV (from Quad4)
 * @authors Romain Boman
 */
void Tri3::buildS(Eigen::MatrixXd &S)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    S = Eigen::Matrix3d::Zero();
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // Jacobian and shape functions
        double detJ = mem.getDetJ(k);
        std::vector<double> &ff = cache.ff[k];

        S += Eigen::Vector3d(ff.data()) * Eigen::RowVector3d(ff.data()) * gauss.w[k] * detJ; // S = S + {ff}^T [ff] * wG * detJ
    }
}

void Tri3::build(MATTYPE type, Eigen::MatrixXd &A, std::vector<double> const &u, Fct0 const &fct)
{
    switch (type)
    {
    case MATTYPE::MAT_S:
        buildS(A);
        break;
    case MATTYPE::MAT_K:
        buildK(A, u, fct);
        break;
    default:
        std::stringstream out;
        out << "Tri3 cannot build " << type;
        throw std::runtime_error(out.str());
    }
}

/**
 * @brief Compute gradient of a scalar field at Gauss point k
 * @authors Romain Boman, Adrien Crovato
 */
void Tri3::computeGrad(std::vector<double> const &u, Eigen::VectorXd &grad, size_t k)
{
    CacheTri3 &cache = getCache();
    MemTri3 &mem = getMem();

    // local matrix of sf derivatives, solution vector and Jacobian
    Eigen::MatrixXd gradN(2, nodes.size());
    Eigen::Vector3d ue;
    for (size_t i = 0; i < nodes.size(); ++i)
    {
        gradN.col(i) = Eigen::Vector2d(cache.dff[k][i].data());
        ue(i) = u[nodes[i]->row];
    }
    Eigen::Matrix2d const &Jinv = mem.getJinv(k);
    // Gradient in global axis: inv(J)_ij * d_jN_i * ue_i
    grad = Jinv * gradN * ue;
}

/**
 * @brief Compute square of gradient of a scalar field at Gauss point k
 * @authors Adrien Crovato
 */
double Tri3::computeGrad2(std::vector<double> const &u, size_t k)
{
    Eigen::VectorXd grad;
    computeGrad(u, grad, k);
    return grad.squaredNorm();
}

/**
 *  calculation of int_V -gradu.fct dV
 */
void Tri3::computeqV(std::vector<double> const &u, Eigen::VectorXd &qV, Fct0 const &fct)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    qV = Eigen::Vector2d::Zero();
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // Gradient function and Jacobian evaluation
        Eigen::VectorXd gradk;
        computeGrad(u, gradk, k);
        double fk = fct.eval(*this, u, k);
        double detJ = mem.getDetJ(k);

        qV -= gradk * fk * gauss.w[k] * detJ; // qV = qV + fk.wG.detJ
    }
}

/**
 *  calculation of int_V flux grad N_j dV
 */
void Tri3::computeqint(std::vector<double> const &u, Eigen::VectorXd &qV, Fct2 const &fct)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    qV = Eigen::Vector3d::Zero();
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // Gradient and flux evaluation
        Eigen::VectorXd gradk;
        computeGrad(u, gradk, k);
        Eigen::MatrixXd fk(2, 2);
        fct.eval(*this, u, k, fk, false);
        double detJ = mem.getDetJ(k);
        Eigen::Vector2d flux;
        flux = fk * gradk;

        // calcul du gradient des fcts de forme et produit scalaire
        Eigen::Vector2d factor1;
        Eigen::Matrix2d const &J = mem.getJinv(k);
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[k][i];
            factor1 = J * Eigen::Vector2d(dffi.data());
            qV(i) += factor1.dot(flux) * gauss.w[k] * detJ;
        }
    }
}

/**
 *  calculation of int_V -[fct2]gradu dV
 */
void Tri3::computeqV(std::vector<double> const &u, Eigen::VectorXd &qV, Fct2 const &fct)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    qV = Eigen::Vector2d::Zero();
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // Gradient and matrix function evaluation
        Eigen::VectorXd gradk;
        computeGrad(u, gradk, k);
        Eigen::MatrixXd fk(2, 2);
        fct.eval(*this, u, k, fk, false);
        double detJ = mem.getDetJ(k);

        qV -= fk * gradk * gauss.w[k] * detJ; // qV = qV + fk.wG.detJ
    }
}

/**
 * @brief Compute intergal of function
 * 
 * I = int fct dV
 * @authors Romain Boman
 */
double Tri3::computeV(std::vector<double> const &u, Fct0 const &fct)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    double V = 0.0;
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // Function evaluation
        double fk = fct.eval(*this, u, k);
        double detJ = mem.getDetJ(k);
        V += fk * gauss.w[k] * detJ;
    }
    return V;
}
void Tri3::computeV(std::vector<double> const &u, Eigen::MatrixXd &out, Fct2 const &fct)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    out = Eigen::MatrixXd::Zero(2, 2); // todo out should be resized to match fk size
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // Function evaluation
        Eigen::MatrixXd fk;
        fct.eval(*this, u, k, fk, false);
        double detJ = mem.getDetJ(k);

        out += fk * gauss.w[k] * detJ; // [out] = [out] + [fk].wG.detJ
    }
}

void Tri3::iso_coordinates(Pt x, double eta, double dzeta)
{
    Eigen::Vector2d iso, b;
    Eigen::Matrix2d A;

    double detA;

    //std::vector<int> index_1 = {1,1,2}; // pas supporté en VC 2012!
    //std::vector<int> index_2 = {2,3,3}; // pas supporté en VC 2012!
    std::vector<int> index_1(3);
    index_1[0] = 1;
    index_1[1] = 1;
    index_1[2] = 2;
    std::vector<int> index_2(3);
    index_2[0] = 2;
    index_2[1] = 3;
    index_2[2] = 3;

    auto it0 = nodes.begin();
    auto it1 = it0 + 1;
    auto it2 = it0 + 2;

    Pt P1 = ((*it1)->pos - (*it0)->pos);
    Pt P2 = ((*it2)->pos - (*it0)->pos);
    Pt P3 = (x - (*it0)->pos);

    for (size_t i = 0; i < 3; ++i)
    {
        A(0, 0) = P1.x[index_1[i]];
        A(1, 0) = P1.x[index_2[i]];

        A(0, 1) = P2.x[index_1[i]];
        A(1, 1) = P2.x[index_2[i]];

        detA = A.determinant();

        if (detA == 0)
            continue;
        else
        {
            b(0) = P3.x[index_1[i]];
            b(1) = P3.x[index_2[i]];
            break;
        }
    }

    iso = A.partialPivLu().solve(b);

    if (iso(0) < 0 || iso(1) > 1 || iso(0) + iso(1) > 1)
        eta = -2;
    else
    {
        eta = iso(0);
        dzeta = iso(1);
    }
}

/**
 * @brief Return the element volume
 * @authors Romain Boman
 */
double Tri3::computeV()
{
    MemTri3 &mem = getMem();
    return mem.getVol();
}
