/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WTRI3_H
#define WTRI3_H

#include "tbox.h"
#include "wElement.h"

namespace tbox
{

/**
 * @brief Linear triangle finite element
 * @authors Romain Boman, Adrien Crovato
 */
class TBOX_API Tri3 : public Element
{
#ifndef SWIG
    friend class MemTri3;
    MemTri3 *pmem; ///< private memory of precalculated values
    MemTri3 &getMem();
    static CacheTri3 &getCache();
    virtual double buildJ(size_t k, Eigen::MatrixXd &J) const override;
    virtual double buildJ(size_t k) const override;

    virtual void build(MATTYPE type, Eigen::MatrixXd &A, std::vector<double> const &u, Fct0 const &fct);
    virtual void buildK(Eigen::MatrixXd &K, std::vector<double> const &u, Fct0 const &fct) override;
    virtual void buildK(Eigen::MatrixXd &K, std::vector<double> const &u, Fct2 const &fct, bool fake) override;
    virtual void buildK_linElastic(Eigen::MatrixXd &K, Eigen::MatrixXd const &H) override;
    virtual void builds(Eigen::VectorXd &s, std::vector<double> const &u, Fct0 const &f) override;
    virtual void builds(Eigen::VectorXd &s, std::vector<double> const &u, Fct1 const &f) override;
    virtual void buildDK(Eigen::MatrixXd &K, std::vector<double> const &u, Fct0 const &fct) override;
    virtual void buildDs(Eigen::VectorXd &s, std::vector<double> const &u, Fct0 const &f) override;
    virtual void buildS(Eigen::MatrixXd &S) override;

    virtual void computeGrad(std::vector<double> const &u, Eigen::VectorXd &grad, size_t k) override;
    virtual double computeGrad2(std::vector<double> const &u, size_t k) override;
    virtual void computeqV(std::vector<double> const &u, Eigen::VectorXd &qV, Fct0 const &fct) override;
    virtual void computeqV(std::vector<double> const &u, Eigen::VectorXd &qV, Fct2 const &fct) override;
    virtual void computeqint(std::vector<double> const &u, Eigen::VectorXd &qV, Fct2 const &fct) override;
    virtual double computeV(std::vector<double> const &u, Fct0 const &fct) override;
    virtual void computeV(std::vector<double> const &u, Eigen::MatrixXd &out, Fct2 const &fct) override;
    virtual double computeV() override;

    virtual void iso_coordinates(Pt x, double eta, double dzeta) override;
#endif

public:
    Tri3(int n, Tag *_ptag, Tag *_etag, std::vector<Tag *> &_parts, std::vector<Node *> &nods);
    virtual ~Tri3();
    virtual ELTYPE type() const override { return ELTYPE::TRI3; }

#ifndef SWIG
    virtual Mem &getVMem() override;
    virtual Cache &getVCache() override;
    virtual Pt normal() override;
#endif
};

} // namespace tbox

#endif //WTRI3_H
